/*
** FileName:        rfc1577.h
** Purpose:
**  Contains the macros, data structures and external API's 
**  provided by rfc1577 module.  
**  It also contains the macros, data structures and 
**  internal API's required for implementing Rfc1577
** Notes:
** Edit History:
**  June 8th 2000   -    New File
**  June 27th 2000  -   Editing the  file for cleaning/integration
*/
#ifndef _RFC1577_H_
#define _RFC1577_H_

#include "Gp.h"

/**********************************************************************
** External API's, macros
**********************************************************************/

#define IP_CLASS_A_PATTERN    0x00000000 /* 0000 */
#define IP_CLASS_B_PATTERN    0x80000000 /* 1000 */
#define IP_CLASS_C_PATTERN    0xC0000000 /* 1100 */
#define IP_CLASS_D_PATTERN    0xE0000000 /* 1110 */


#define IP_NETID_MASK_A       0xFF000000
#define IP_NETID_MASK_B       0xFFFF0000
#define IP_NETID_MASK_C       0xFFFFFF00
#define IP_NETID_MASK_D       0xFFFFFFFF

/*
** Return Values of 1577 function API's
*/
#define RFC1577_FAILURE     NDIS_STATUS_FAILURE
#define RFC1577_SUCCESS     NDIS_STATUS_SUCCESS
#define RFC1577_PENDING     NDIS_STATUS_PENDING

/*
** Timer Ticks
*/
#define TIMER_PERIODIC_TICKS        1 /* Every sec */

typedef ATM_UNI_HANDLE VCCHandle_t;

/**********************************************************************
**  API Declaractions:
***********************************************************************/
/*
** External API's  interface provided by Rfc1577
*/
BOOLEAN Rfc1577Init (PATM_USB_INSTANCE NdisContext, Rfc1577Config_t *pConfig);

int Rfc1577Close (void *Rfc1577Context);

int Rfc1577ConfigIntf (void *Rfc1577Context, 
                       unsigned int IPAddress,      /* IPAddress    */
                       unsigned int IPMask,         /* IP Mask      */
                       int MTU);                    /* Max Tx Unit  */

int Rfc1577EnableIntf (void *Rfc1577Context, 
                       unsigned int IPAddress);     /* Intf IPaddr  */

int Rfc1577DisableIntf (void *Rfc1577Context, 
                       unsigned int IPAddress);     /* Intf IPAddr  */

int Rfc1577DeleteIntf (void *Rfc1577Context,         
                       unsigned int IPAddress);     /* Intf IPAddr  */

int Rfc1577CreatePVC (void *Rfc1577Context,
                      unsigned int IPAddress,       /* Intf IPAddr  */          
                      unsigned char VPI,             /* VPI          */
                      unsigned short VCI,             /* VCI          */
                      RFC1483_MODE EncapType,       /* Encapsulation*/
                      unsigned int NextHop);        /* Optional     */

int Rfc1577DeletePVC (void *Rfc1577Context,
                      unsigned int IPAddress,       /* Intf IPAddr  */
                      unsigned int VPI,             /* VPI          */
                      unsigned int VCI);            /* VCI          */

NDIS_STATUS NdisRfc1577Send (PATM_USB_INSTANCE pInstance, 
                             PNDIS_PACKET pPacket, 
                             unsigned int Flag,
							 gsi_lan_t    *plan);

void Rfc1483Rfc1577SendComplete (PATM_USB_INSTANCE pInstance,
                                VCCHandle_t VCCHandle,
                                PNDIS_PACKET pPacket,
                                BOOLEAN ArpFlag);

void Rfc1483Rfc1577Receive (void * pRfc1577Context,
                            VCCHandle_t VCCHandle,
                            PXFER_DATA_BUFFER pXData,
                            unsigned char *pRfc1483Header,
                            unsigned long Rfc1483HeaderSize,
                            BOOLEAN ArpFlag);


int Rfc1577ResolveARP (void *Rfc1577Context,
                       unsigned int IPAddr,         /* Dest Addr        */
                       unsigned char *pMacAddr);             /* Returned MacAddr */
                      
int 
Rfc1577GetArpEntries(void *pRfc1577Context , 
                         PQUERY_ARP_BUFFER pBuffer
                         );

void 
Rfc1577DeleteArpEntry(void *pRfc1577Context , 
                      Rfc1577Status_t *pArpBuffer
                      );

/***************************************************************************
** Constants which can be tweaked 
***************************************************************************/
/*
** Maximum Interfaces,  PVCs,
*/
#define MAX_INTF    2       /* Maximum Interfaces supported         */
#define MAX_PVC     2       /* Maximum PVC per interface            */
#define MAX_TRIES   0xffff  /* Maximum ARP Tries, before giving up  */

/*
** Timer Values
*/
#define ARP_TIMEOUT         15*60       /* Aging  time out               */
#define ARP_REQ_TIMEOUT     1*10        /* Expected  In/ARP reply timeout*/


/*************************************************************************
**
** INTERNAL API/MACROS and data structure
**
*************************************************************************/

/*
 * IP related macros
 */
#define IPADDR_LEN          0x4
#define IP_PROTOCOL         0x0800

/*
** ATM address related macros and structure
*/
#define E164_ADDRLEN       20           /* E.164 Address len */
#define NSAP_ADDRLEN       20           /* NSAP- address len */

struct ATMAddr{
    unsigned char AddrType;                 /* Type of Address          */
    unsigned char AddrLen;                  /* Length of Address        */
    unsigned char Addr[MAX_ATMADDR_LEN];    /* ATM Addr                 */
    unsigned char SubAddrType;              /* Type Of sub-address      */
    unsigned char SubAddrLen;               /* Length of sub-address    */
    unsigned char SubAddr[MAX_ATMADDR_LEN]; /* SubAddr                  */
};


/*
** The VPI, VCI entry table
*/
typedef struct ARPEntry {
    int IntfIndex;              /* Associated intf index                */
    unsigned char VPI;                    /* least significant 12 bits            */
    unsigned short VCI;                    /* least significant 16 bits            */
    VCCHandle_t Handle;         /* Handle given by 1483                 */
    RFC1483_MODE Encap;         /* Encapsulation type                   */
    void *TimerHandle;          /* TimerHandle, returned by Timer module*/
    struct ATMAddr  TgtAddr;    /* Target ATM address,                  */
    unsigned int TgtIPAddr;     /* Target IP Address                    */
    unsigned int State;         /* VALID, TOBEvalid, Invalid            */
    int  NumOfTries;            /* Number of ARP tries                  */
    unsigned char EthMacAddr[ETH_LENGTH_OF_ADDRESS]; 
                                /* Resolved Ether Mac address           */
}ARPEntry_t;

/*
** ARP State
*/
#define ARP_INVALID     0x0         /* Entry is Invalid             */
#define ARP_DOWN        0x1         /* Configured but not opened    */
#define ARP_SENT        0x2         /* Sent ARP waiting for reply   */
#define ARP_MAXTRY      0x4         /* Timeout after Max tries      */
#define ARP_VALID       0x8         /* Got reply                    */

/*
** E164 flag used in the ARP Packet encoding
*/
#define TL_E164         0x40        /* Or'ing flag for E.164 Addr in ARP */

/*
** Interface is maintained as a arraylist in the Context structure
*/
struct Intf {
    int Index;                          /* Interface index              */
    unsigned int IntfAddr;              /* IP Address of this interface */
    unsigned int IntfMask;              /* IPMask                       */
    int MTU;                            /* Maximum Transmission Unit    */
    ARPEntry_t ArpTable[MAX_PVC];       /* ATM-ARP table                */
    struct ATMAddr ATMAddress;          /* ATM-address of this interface*/
    int State;                          /* state, UP, Down, Invalid     */
};

/*
** Interface State
*/
#define IF_INVALID      0x0         /* Interface not yet configured */
#define IF_DOWN         0x1         /* Interface is marked down     */
#define IF_UP           0x2         /* Interface is configured up   */

/*
** ARP Packet
*/
struct CIPArpHdr {
    unsigned short ArpHrd;     /* Hardware type                         */
    unsigned short ArpPro;     /* Protocol type                         */
    unsigned char  ArpShtl;    /* Type & length of source ATM number    */
    unsigned char  ArpSstl;    /* Type & length of source ATM subaddress*/
    unsigned short ArpOp;      /* Operation code                        */
    unsigned char  ArpSpln;    /* Length of source protocol address  (4)*/
    unsigned char  ArpThtl;    /* Type & length of target ATM number    */
    unsigned char  ArpTstl;    /* Type & length of target ATM subaddress*/
    unsigned char  ArpTpln;    /* Length of target protocol address (4) */
    unsigned char  Data[1];    /* addresses                             */
};

#define MAX_ARP_PKTLEN (sizeof(struct CIPArpHdr)+ 2*(E164_ADDRLEN + NSAP_ADDRLEN + sizeof(unsigned int))-1)

/*
** ATMForum Address Family
*/
#define ATM_ADDR_FAMILY     0x0013

/*
** Operation codes
*/
#define ARP_REQUEST     0x0001      /* ARP Request          */
#define ARP_REPLY       0x0002      /* ARP Reply            */
#define INARP_REQUEST   0x0008      /* Inverse ARP Request  */
#define INARP_REPLY     0x0009      /* Inverse ARP Reply    */
#define ARP_NAK         0x000a      /* ARP NAK              */

/**********************************************************************
** Timer Data structure and function prototypes
***********************************************************************/

/* 
** Call back function type definitions 
*/
typedef int (*DrvCallBack_t)(void *Context, void *ArpEntry);

/* 
** Timeout callback function structure 
*/
typedef struct _DrvCbfnInfo_t
{
    DrvCallBack_t    Cbfn;  /* function pointer to callback function     */
    void *Data;             /* data pointer as 1st argument for function */
    void *Arg1;             /* Second argument for the callback function */
} DrvCbfnInfo_t;

typedef struct _DrvTimerNode_t
{   
    DrvCbfnInfo_t   CbfnInfo;        /* function callback after timeout   */
    unsigned int    ExpiryCounter;   /* expiration counter for this timer */
    struct _DrvTimerNode_t   *Next;  /* next timer in linked list         */
} DrvTimerNode_t;

typedef struct _DrvTimerList_t
{
    DrvTimerNode_t *DrvTimerLeastTimer;/* pointer to the least timer node   */
    unsigned long CurrentTicks;        /* Gbl counter, keeps track of ticks */
} DrvTimerList_t;

/*
** Context Structure, Passed as input parameter in all API's.
*/
struct Rfc1577Context {
    struct Intf IntfList[MAX_INTF]; /* Interfaces Configured            */
    struct ATMAddr  ATMAddress;     /* If no ATM address is configured per
                                       interface, This value is taken   */
    unsigned int TotalIntf;         /* Total # Interfaces Configured    */
    DrvTimerList_t  DrvTimerList;   /* 1577 Timer list                  */
    PATM_USB_INSTANCE NdisHandle;   /* NDIS Layer Handle                */
    RFC1577_HANDLE TimerHandle;     /* TimerHandle provided by Ndis     */
    RFC1577_HANDLE TimerSpinLock;   /* TimerSpinLock                    */
    /*
     * Call back Function Pointers passed during configuration
     */

    RFC1577RFC1483SEND  Rfc1577Rfc1483Send;
    RFC1577RFC1483RECEIVECOMPLETE Rfc1577Rfc1483ReceiveComplete;

    BOOLEAN Rfc1577Enable;
    /*
     * If Rfc1577 Enabled, the default VPI, VCI will be used
     * for sending data 
     */
    VCCHandle_t DefHandle;
    RFC1483_MODE DefEncap;
    unsigned char DefVPI;             /*Default  VPI          */
    unsigned short DefVCI;             /*Default  VCI          */
    unsigned long  MTUSize;


};

/************************************************************************
** Internal API's implemented
************************************************************************/
int BuildARPPkt (struct Rfc1577Context *Context,
                 int Type, 
                 unsigned char *Buffer,
                 int BuffLen,
                 struct ARPEntry *VCC);

int DecodeARPPkt (struct Rfc1577Context *Context, 
                  VCCHandle_t VCCHandle,
                  unsigned char *Buffer,
                  int BuffLen);

int LookupIntf (struct Rfc1577Context *Context, unsigned int IPAddress);

struct ARPEntry *LookForVCC (struct Rfc1577Context *Context, VCCHandle_t VCCHandle);

int Rfc1577TimerExpiry (void *Context, void *VCC);

struct ARPEntry *GetArpEntry (struct Rfc1577Context *pContext, 
                             unsigned int NextHopIPAddr);
struct ARPEntry *GetArpEntryFromMac (struct Rfc1577Context *pContext, 
                                     unsigned char *pMacAddr);


/*****************************************************************************
 * Timer Module API's
 ****************************************************************************/
int DeleteTimer (void *pHandle, void * pTHandle);

void *AddTimer( void *pHandle, void *pAEntry, 
                int TimeOut, 
                int (*CallBack)(void *, void *));

void Rfc1577TimerInit (struct Rfc1577Context *pRfc1577Context);

void Rfc1577TimerClose (struct Rfc1577Context *pRfc1577Context);

void Rfc1577InitTimerCallback (DrvTimerNode_t *Timer, 
                               DrvCallBack_t Fn, 
                               void *Data, 
                               void *Arg1);

void _Rfc1577SetTimer (struct Rfc1577Context *pRfc1577Context, 
                       DrvTimerNode_t *Timer, 
                       unsigned int msec);

void _Rfc1577CancelTimer (struct Rfc1577Context *pRfc1577Context, 
                          DrvTimerNode_t *Timer);

void Rfc1577ProcessTimeout (void *Notused1, void *pRfc1577Context, 
                            void *Notused2, 
                            void *Notused3);

unsigned int Rfc1577GetNetIDMask(unsigned int IpAddress );

BOOLEAN Rfc1577Initialize(PATM_USB_INSTANCE Instance);
BOOLEAN Rfc1577CancelAllTimers(PATM_USB_INSTANCE Instance);
BOOLEAN Rfc1577SetAllTimers(PATM_USB_INSTANCE Instance );


#ifdef DEBUG
void DbgPrintAllEntry(void *pRfc1577Context);
#else
#define DbgPrintAllEntry(NotUsed)
#endif
#endif 
