#ifdef GSI_LAN

#include "Gp.h"	
#include "GpSnull.h"
#include "api1577.h"

extern void HexDump(PUCHAR, ULONG);



////////////////////////////////////////////////////////////////////////////////
//
// NAME: Rfc1577ApiInitialize
//
// ABSTRACT:
//		Initializes the Rfc1577 Api, This Api's are used by Rfc1577 Layer( ATM ARP)
//		
//
// RETURNS:
//		Boolean
//
// ASSUMPTIONS:
//		Runs in the context of MiniportInitialize.
//
////////////////////////////////////////////////////////////////////////////////



BOOLEAN Api1577Initialize(PATM_USB_INSTANCE Instance)
{
    RFC1577_HANDLE i;
    PRFC1577 pRfc1577 = &Instance->Rfc1577;
    NDIS_STATUS  Status;

    DBGPRINT(
            DBG_PRINT_MODULE_RFC1577_ALL,
            DBG_PRINT_LEVEL_TRACE,
            ("Api1577Initialize\n"));

    ASSERT( Instance);

//init spin lock
    for (i = 0; i < RFC1577_SPIN_LOCK_COUNT; i++)
    {
        pRfc1577->Rfc1577SpinLock[i].busy = FALSE;
    }
//init Timers
    for (i = 0; i < RFC1577_TIMER_COUNT; i++)
    {
        pRfc1577->Rfc1577Timer[i].busy = FALSE;
    }

//init Packet pool & Buffer Pool
    NdisAllocatePacketPool (
                           &Status,
                           &Instance->Rfc1577.Rfc1577NdisPktPoolHandle,
                           RFC1577_MAX_PACKET_POOL,
                           PROTOCOL_RESERVED_LEN );


    if (Status != NDIS_STATUS_SUCCESS)
        return(FALSE);

    NdisAllocateBufferPool (
                           &Status,
                           &Instance->Rfc1577.Rfc1577NdisBuffPoolHandle,
                           RFC1577_MAX_PACKET_POOL);


    if (Status != NDIS_STATUS_SUCCESS)
        return(FALSE);

return TRUE;
}


////////////////////////////////////////////////////////////////////////////////
//
// NAME: Api1577AllocateSpinLock
//
// ABSTRACT:
//      Initializes SpinLock, This Api's are used by Rfc1577 Layer( ATM ARP)
//      This function has to be called just ones per eatch spin Lock.
//
// RETURNS:
//      Handle to spin Lock. This Handle will Be used by Rfc1577FreeSpinLock &
//      Rfc1577AcquireSpinLock to reffer to this spin Lock
//
// ASSUMPTIONS:
//      runing at PASSIVE_LEVEL during initialization
//
////////////////////////////////////////////////////////////////////////////////

RFC1577_HANDLE Api1577AllocateSpinLock( PATM_USB_INSTANCE Instance )
{

    RFC1577_HANDLE handle;
    PRFC1577_SPIN_LOCK pRfc1577SpinLick = Instance->Rfc1577.Rfc1577SpinLock;


    DBGPRINT(
            DBG_PRINT_MODULE_RFC1577_ALL,
            DBG_PRINT_LEVEL_TRACE,
            ("Api1577AllocateSpinLock\n"));



    ASSERT( Instance);

    handle = GetFreeSpinLockHandle( pRfc1577SpinLick );

    if ((handle == RFC1577_ERROR) ||  (pRfc1577SpinLick[handle].busy == TRUE))
    {
        DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_ERR,
                 ("ERROR handle = %ld , busy = %d\n",  handle , pRfc1577SpinLick[handle].busy));
        return (RFC1577_ERROR);

    }

    pRfc1577SpinLick[handle].busy = TRUE;
    NdisAllocateSpinLock( &pRfc1577SpinLick[handle].NdisSpinLock );

return handle;
}

////////////////////////////////////////////////////////////////////////////////
//
// NAME: Rfc1577FreeSpinLock
//
// ABSTRACT:
//      releases a spin lock initialized in a preceding call to Rfc1577AllocateSpinLock
//     This Api's are used by Rfc1577 Layer( ATM ARP)
//      This function has to be called just ones per eatch spin Lock
//
// RETURNS: BOOLEAN
//      TRUE -   spin Lock handle is valid
//      FALSE -  spin lock handle is not valid
//
// ASSUMPTIONS:
//      runing at PASSIVE_LEVEL
//
////////////////////////////////////////////////////////////////////////////////


BOOLEAN Api1577FreeSpinLock(PATM_USB_INSTANCE Instance , RFC1577_HANDLE handle)
{
    PRFC1577_SPIN_LOCK pRfc1577SpinLick = Instance->Rfc1577.Rfc1577SpinLock;


    DBGPRINT(
            DBG_PRINT_MODULE_RFC1577_ALL,
            DBG_PRINT_LEVEL_TRACE,
            ("Api1577FreeSpinLock\n"));

    ASSERT( Instance);

    if ((handle >= RFC1577_SPIN_LOCK_COUNT) ||  (pRfc1577SpinLick[handle].busy == FALSE))
    {
        DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_ERR,
                 ("ERROR handle = %ld , busy = %d\n",  handle , pRfc1577SpinLick[handle].busy));
        return (FALSE);

    }

    pRfc1577SpinLick[handle].busy = FALSE;
    NdisFreeSpinLock( &pRfc1577SpinLick[handle].NdisSpinLock);
return TRUE;
}

////////////////////////////////////////////////////////////////////////////////
//
// NAME: Rfc1577AcquireSpinLock
//
// ABSTRACT:
//     acquires a spin lock
//     This Api's are used by Rfc1577 Layer( ATM ARP)
//
//
// RETURNS: BOOLEAN
//      TRUE -   spin Lock handle is valid
//      FALSE -  spin lock handle is not valid
//
// ASSUMPTIONS:
//      runing at IRQL <= DISPATCH_LEVEL
//
////////////////////////////////////////////////////////////////////////////////



BOOLEAN Api1577AcquireSpinLock(PATM_USB_INSTANCE Instance , RFC1577_HANDLE handle)
{
    PRFC1577_SPIN_LOCK pRfc1577SpinLick = Instance->Rfc1577.Rfc1577SpinLock;

    DBGPRINT(
            DBG_PRINT_MODULE_RFC1577_ALL,
            DBG_PRINT_LEVEL_TRACE,
            ("Api1577AcquireSpinLock\n"));

    ASSERT( Instance);


    if ((handle >= RFC1577_SPIN_LOCK_COUNT) ||  (pRfc1577SpinLick[handle].busy == FALSE))
    {
        DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_ERR,
                 ("ERROR handle = %ld , busy = %d\n",  handle , pRfc1577SpinLick[handle].busy));
        return (FALSE);

    }
    NdisAcquireSpinLock( &pRfc1577SpinLick[handle].NdisSpinLock);
return TRUE;
}

////////////////////////////////////////////////////////////////////////////////
//
// NAME: Api1577ReleaseSpinLock
//
// ABSTRACT:
//     releases a spin lock that was acquired in a preceding call Rfc1577AcquireSpinLock
//     This Api's are used by Rfc1577 Layer( ATM ARP)
//
//
// RETURNS: BOOLEAN
//      TRUE -   spin Lock handle is valid
//      FALSE -  spin lock handle is not valid
//
// ASSUMPTIONS:
//      runing at IRQL <= DISPATCH_LEVEL
//
////////////////////////////////////////////////////////////////////////////////


BOOLEAN Api1577ReleaseSpinLock(PATM_USB_INSTANCE Instance , RFC1577_HANDLE handle)
{
    PRFC1577_SPIN_LOCK pRfc1577SpinLick = Instance->Rfc1577.Rfc1577SpinLock;

    DBGPRINT(
            DBG_PRINT_MODULE_RFC1577_ALL,
            DBG_PRINT_LEVEL_TRACE,
            ("Api1577ReleaseSpinLock\n"));

    ASSERT( Instance);


    if ((handle >= RFC1577_SPIN_LOCK_COUNT) ||  (pRfc1577SpinLick[handle].busy == FALSE))
    {
        DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_ERR,
                 ("ERROR handle = %ld , busy = %d\n",  handle , pRfc1577SpinLick[handle].busy));
        return (FALSE);

    }
    NdisReleaseSpinLock( &pRfc1577SpinLick[handle].NdisSpinLock);
return TRUE;
}

////////////////////////////////////////////////////////////////////////////////
//
// NAME: GetFreeSpinLockHandle
//
// ABSTRACT:
//
//     Internal Rfc1577 Api Function
//       It gets free Spin lock handle
//
// RETURNS: BOOLEAN
//      Handle - in case of seccess
//      RFC1577_ERROR - in case of ERROR
//
// ASSUMPTIONS:
//      Rfc1577ApiInit was called.
//
////////////////////////////////////////////////////////////////////////////////


RFC1577_HANDLE GetFreeSpinLockHandle(PRFC1577_SPIN_LOCK pRfc1577SpinLock)
{
    RFC1577_HANDLE i;

    DBGPRINT(
            DBG_PRINT_MODULE_RFC1577_ALL,
            DBG_PRINT_LEVEL_TRACE,
            ("GetFreeSpinLockHandle\n"));


    for (i = 0; i < RFC1577_SPIN_LOCK_COUNT; i++)
    {
        if (pRfc1577SpinLock[i].busy == FALSE)
            break;
    }
    return(i < RFC1577_SPIN_LOCK_COUNT) ? i : RFC1577_ERROR;
}


////////////////////////////////////////////////////////////////////////////////
//
// NAME: GetFreeTimerHandle
//
// ABSTRACT:
//
//     Internal Rfc1577 Api Function
//       It gets free Timer handle
//
// RETURNS: BOOLEAN
//      Handle - in case of seccess
//      RFC1577_ERROR - in case of ERROR
//
// ASSUMPTIONS:
//      Rfc1577ApiInit was called.
//
////////////////////////////////////////////////////////////////////////////////

RFC1577_HANDLE GetFreeTimerHandle( PRFC1577_TIMER pRfc1577Timer )
{
    RFC1577_HANDLE i;
    DBGPRINT(
            DBG_PRINT_MODULE_RFC1577_ALL,
            DBG_PRINT_LEVEL_TRACE,
            ("GetFreeTimerHandle\n"));


    for (i = 0; i < RFC1577_TIMER_COUNT; i++)
    {
        if (pRfc1577Timer[i].busy == FALSE)
            break;
    }
    return(i < RFC1577_TIMER_COUNT) ? i : RFC1577_ERROR;
}


////////////////////////////////////////////////////////////////////////////////
//
// NAME: Api1577InitializeTimer
//
// ABSTRACT:
//      Initializes Timer, This Api's are used by Rfc1577 Layer( ATM ARP)
//      This function has to be called just ones per eatch Timer.
//
// RETURNS:
//      Handle to Timer. This Handle will Be used by Rfc1577SetTimer &
//      Rfc1577CancelTimer to reffer to this Timer
//
// ASSUMPTIONS:
//      runing at PASSIVE_LEVEL during initialization
//
////////////////////////////////////////////////////////////////////////////////


RFC1577_HANDLE Api1577InitializeTimer(PATM_USB_INSTANCE Instance, CALLBACK_FUNC callbackfunc )
{
    RFC1577_HANDLE handle;
    PRFC1577_TIMER pRfc1577Timer = Instance->Rfc1577.Rfc1577Timer;

    DBGPRINT(
            DBG_PRINT_MODULE_RFC1577_ALL,
            DBG_PRINT_LEVEL_TRACE,
            ("Api1577InitializeTimer\n"));

    ASSERT( Instance);
    ASSERT( callbackfunc);
    ASSERT(Instance->Rfc1577Config.Rfc1577Context);

    handle = GetFreeTimerHandle(pRfc1577Timer);

    if ((handle == RFC1577_ERROR) ||  (pRfc1577Timer[handle].busy == TRUE))
    {
        DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_ERR,
                 ("ERROR handle = %ld , busy = %d\n",  handle , pRfc1577Timer[handle].busy));
        return (RFC1577_ERROR);

    }

    pRfc1577Timer[handle].busy = TRUE;
    pRfc1577Timer[handle].TimerFunction = callbackfunc;

    NdisMInitializeTimer( &pRfc1577Timer[handle].NdisTimer ,
                          Instance->MiniportAdapterHandle,
                          (PNDIS_TIMER_FUNCTION)callbackfunc,
                          Instance->Rfc1577Config.Rfc1577Context
                        );

    return (handle);
}

////////////////////////////////////////////////////////////////////////////////
//
// NAME: Api1577SetTimer
//
// ABSTRACT:
//      sets a timer to fire after a specified interval, This Api's are used by Rfc1577 Layer( ATM ARP)
//      Time in ms.
//
// RETURNS:
//      TRUE - in case of valid Timer handle
//      FALSE - in Case of invalid Handle
//
// ASSUMPTIONS:
//
//
////////////////////////////////////////////////////////////////////////////////


BOOLEAN Api1577SetTimer(PATM_USB_INSTANCE Instance, RFC1577_HANDLE handle, UINT MillisecondsToDelay )
{
    PRFC1577_TIMER pRfc1577Timer = Instance->Rfc1577.Rfc1577Timer;

    DBGPRINT(
            DBG_PRINT_MODULE_RFC1577_ALL,
            DBG_PRINT_LEVEL_TRACE,
            ("Api1577SetTimer\n"));

    ASSERT( Instance);

    if ((handle >= RFC1577_TIMER_COUNT) || (handle < 0) || (pRfc1577Timer[handle].busy == FALSE))
    {
        DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_ERR,
                 ("ERROR handle = %ld , busy = %d\n",  handle , pRfc1577Timer[handle].busy));
        return (FALSE);

    }

    NdisMSetTimer(
                 &pRfc1577Timer[handle].NdisTimer,
                 MillisecondsToDelay
                 );

return TRUE;
}


////////////////////////////////////////////////////////////////////////////////
//
// NAME: Api1577CancelTimer
//
// ABSTRACT:
//      Cansel Timer that was set by Rfc1577SetTimer. This Api's are used by Rfc1577 Layer( ATM ARP)
//      Time in ms.
//
// RETURNS:
//      TRUE - in case of valid Timer handle
//      FALSE - in Case of invalid Handle
//
// ASSUMPTIONS:
//
//
////////////////////////////////////////////////////////////////////////////////

BOOLEAN Api1577CancelTimer(PATM_USB_INSTANCE Instance, RFC1577_HANDLE handle, UINT MillisecondsToDelay )
{
PRFC1577_TIMER pRfc1577Timer = Instance->Rfc1577.Rfc1577Timer;
BOOLEAN  TimerCancelled;

    DBGPRINT(
        DBG_PRINT_MODULE_RFC1577_ALL,
        DBG_PRINT_LEVEL_TRACE,
        ("Api1577CancelTimer\n"));


    if((handle >= RFC1577_TIMER_COUNT) || (pRfc1577Timer[handle].busy == FALSE) )
    {
           DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_ERR,
                   ("ERROR handle = %d , busy = %d\n",  handle , pRfc1577Timer[handle].busy));
            return FALSE;    
    
    }

    SafeCancelTimer(
        &pRfc1577Timer[handle].NdisTimer,
        &pRfc1577Timer[handle].NdisTimerEvent,
        MillisecondsToDelay);

return TRUE;
}


#endif // GSI_LAN
