/*+M*************************************************************************
 * GlobeSpan ADSL LAN Interface
 *
 * Copyright (c) 2000 GlobeSpan Semiconductor Inc. -- http://www.globespan.net
 * All rights reserved.
 *
 *-M*************************************************************************/
#ifdef MODULE
#include <linux/module.h>
#endif /* MODULE */

#include <Gp.h>
#include <rfc1577.h>
#include <gsi_ioctl.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>

extern u8		PVCCount;

#define TX_TIMEOUT          ((500 * HZ) / 1000)

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,3,43))
#define netif_wake_queue(dev)   do { \
                                    clear_bit(0, &dev->tbusy); \
                                    mark_bh(NET_BH); \
                                } while(0)
#define netif_start_queue(dev)  do { \
                                    dev->tbusy = 0; \
                                    dev->interrupt = 0; \
                                    dev->start = 1; \
                                } while (0)
#define netif_stop_queue(dev)   set_bit(0, &dev->tbusy)
#define netif_running(dev)      dev->start
#define netdevice_start(dev)    dev->start = 1
#define netdevice_stop(dev)     dev->start = 0
#define netif_set_tx_timeout(dev, tf, tm)
#define dev_kfree_skb_irq(x)    dev_kfree_skb(x)
#else /* LINUX_VERSION_CODE */
#define netdevice_start(dev)
#define netdevice_stop(dev)
#define netif_set_tx_timeout(dev, tf, tm) \
                                do { \
                                    (dev)->tx_timeout = (tf); \
                                    (dev)->watchdog_timeo = (tm); \
                                } while(0)
#endif /* LINUX_VERSION_CODE */
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,3,47))
#define netif_device_attach(dev) netif_start_queue(dev)
#define netif_device_detach(dev) netif_stop_queue(dev)
#endif /* LINUX_VERSION_CODE */

DBG_STATIC int GpLanOpen(struct net_device *dev);
DBG_STATIC int GpLanClose(struct net_device *dev);
DBG_STATIC int GpLanIoctl(struct net_device *dev, struct ifreq *rq, int cmd);
DBG_STATIC int GpLanTx(struct sk_buff *skb, struct net_device *dev);

DBG_STATIC struct net_device_stats *GpLanStats(struct net_device *dev);
DBG_STATIC void GpLanMulticast(struct net_device *dev);

/*+F*************************************************************************
 * Function:
 *   gsi_lan_alloc
 *
 * Description:
 *   Allocate the resources necessary for the LAN interface.
 *-F*************************************************************************/
int
GpLanAlloc(
    adapter_t       *pAdapter,
    int		   		iPvcIndex
    )
{
    DBG_FUNCX("GpLanAlloc", DBG_LAN_SYS)
    struct net_device  *dev = NULL;
    gsi_lan_t      *pLan;
    int             status = 0;
    NDIS_STATUS     nStatus;
    int             x;

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pAdapter", ("(0x%p)", pAdapter));

    if ((dev = (struct net_device *) kmalloc(sizeof(*dev), GFP_KERNEL)) == NULL)
    {
        DBG_ERROR(DbgInfo, ("Unable to allocate network device data\n"));
        status = -ENOMEM;
        goto gsi_lan_alloc_exit;
    }
    memset(dev, 0, sizeof(*dev));
    //dev->priv = pAdapter; /*  Now we store lan pointer in the private member */

    // Setup the device defaults as an ethernet device
    ether_setup(dev);

    pLan = &(pAdapter->lan[iPvcIndex]);

    spin_lock_init(&(pLan->slock));

    pLan->txAvailable = GSI_LAN_NUM_NDIS_PACKETS;
    pLan->txHead = 0;
    for (x = 0; x < pLan->txAvailable; x++)
    {
        NdisAllocateMemoryWithTag((PVOID *) &(pLan->bufferPool[x]),
            GSI_LAN_NDIS_BUFFER_SIZE, GSI_LAN_BUFFER_TAG);
        if (pLan->bufferPool[x] == NULL)
        {
            DBG_ERROR(DbgInfo, ("NdisAllocateMemoryWithTag failed\n"));
            status = -ENOMEM;
            goto gsi_lan_alloc_exit;
        }
    }

    NdisAllocatePacketPool(&nStatus, &(pLan->NdisPacketPoolHandle),
        GSI_LAN_NUM_NDIS_PACKETS, PROTOCOL_RESERVED_SIZE_IN_PACKET);
    if ((nStatus != NDIS_STATUS_SUCCESS) || (pLan->NdisPacketPoolHandle == NULL))
    {
        DBG_ERROR(DbgInfo, ("NdisAllocatePacketPool failed!\n"));
        status = -ENOMEM;
        goto gsi_lan_alloc_exit;
    }

    NdisAllocateBufferPool(&nStatus, &(pLan->NdisBufferPoolHandle),
        GSI_LAN_NUM_NDIS_PACKETS);
    if ((nStatus != NDIS_STATUS_SUCCESS) || (pLan->NdisBufferPoolHandle == NULL))
    {
        DBG_ERROR(DbgInfo, ("NdisAllocateBufferPool failed!\n"));
        status = -ENOMEM;
        goto gsi_lan_alloc_exit;
    }

gsi_lan_alloc_exit:
    if (status != 0)
    {
      GpLanFree(pAdapter, iPvcIndex);
    }
    else
    {
	  /*
	  **	Lets now store lan device with in the private member of net device
	  **	and Adapter into the lan device.
	  */
	  dev->priv = pLan;
	  pLan->pAdapter = pAdapter;
      pAdapter->lan[iPvcIndex].dev = dev;
    }

    DBG_RETURNF(DbgInfo, ("status %d", status));
    return status;
}

/*+F*************************************************************************
 * Function:
 *   GpLanFree
 *
 * Description:
 *   Free the LAN interface resources.
 *-F*************************************************************************/
void
GpLanFree(
    adapter_t       *pAdapter,
    int		   		iPvcIndex
    )
{
    DBG_FUNCX("GpLanFree", DBG_LAN_SYS)
    struct net_device  *dev;
    gsi_lan_t      *pLan = &(pAdapter->lan[iPvcIndex]);
    unsigned long   flags;
    int             x;

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pAdapter", ("(0x%p)", pAdapter));

    if ((dev = pLan->dev) != NULL)
    {
        pLan->dev = NULL;
    }

    spin_lock_irqsave(&(pLan->slock), flags);

    for (x = 0; x < pLan->txAvailable; x++)
    {
        if (pLan->bufferPool[x] != NULL)
        {
            NdisFreeMemory(pLan->bufferPool[x], GSI_LAN_NDIS_BUFFER_SIZE, 0);
            pLan->bufferPool[x] = NULL;
        }
    }
    pLan->txAvailable = 0;

    spin_unlock_irqrestore(&(pLan->slock), flags);

    if (pLan->NdisBufferPoolHandle != NULL)
    {
        NdisFreeBufferPool(pLan->NdisBufferPoolHandle);
        pLan->NdisBufferPoolHandle = NULL;
    }

    if (pLan->NdisPacketPoolHandle != NULL)
    {
        NdisFreePacketPool(pLan->NdisPacketPoolHandle);
        pLan->NdisPacketPoolHandle = NULL;
    }

    if (dev != NULL)
    {
        kfree(dev);
    }

    DBG_LEAVE(DbgInfo);
}

/*+F*************************************************************************
 * Function:
 *   GpLanRegister
 *
 * Description:
 *   Register a LAN interface with the operating system. This will request
 *   that the operating system create an ethernet interface to handle the
 *   LAN traffic for the adapter.
 *-F*************************************************************************/
int
GpLanRegister(
    adapter_t       *pAdapter,
	int				ilanindex
    )
{
    DBG_FUNCX("GpLanRegister", DBG_LAN_SYS)
    struct net_device  *dev;
    int     status,x;

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, " Reg pAdapter", ("(0x%p)", pAdapter));

    if ((dev = pAdapter->lan[ilanindex].dev) != NULL)
    {

#ifdef	PCI_ADAPTER
        dev->irq        = pAdapter->irq;
        dev->mem_start  = pAdapter->physAddr;
        dev->mem_end    = pAdapter->physAddr + GSI_PCI_MEMORY_SIZE - 1;

#endif	/* PCI_ADAPTER */        

        dev->open               = &GpLanOpen;
        dev->stop               = &GpLanClose;
        dev->hard_start_xmit    = &GpLanTx;
        dev->get_stats          = &GpLanStats;
        dev->set_multicast_list = &GpLanMulticast;
        dev->do_ioctl           = &GpLanIoctl;
        for (x = 0; x < GSI_MAC_ADDRESS_LENGTH; x++)
        {
            dev->dev_addr[x] = pAdapter->macAddr[x];
        }
        status = register_netdev(dev);
    	//strcpy(pAdapter->lan[ilanindex].devName,dev->name);
    }
    else
    {
        DBG_WARNING(DbgInfo, ("No network device to register!\n"));
        status = -ENODEV;
    }

    DBG_RETURNF(DbgInfo, ("status %d", status));
    return status;
}

/*+F*************************************************************************
 * Function:
 *   GpLanUnregister
 *
 * Description:
 *   Unregister the ethernet (LAN) interface for the adapter.
 *-F*************************************************************************/
void
GpLanUnregister(
    adapter_t       *pAdapter,
	int				ilanindex
    )
{
    DBG_FUNCX("GpLanUnregister", DBG_LAN_SYS)
    struct net_device  *dev;

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pAdapter", ("(0x%p)", pAdapter));

    if ((dev = pAdapter->lan[ilanindex].dev) != NULL)
    {
        unregister_netdev(dev);
    }

    DBG_LEAVE(DbgInfo);
}

/*+F*************************************************************************
 * Function:
 *   GpLanRx
 *
 * Description:
 *   A new LAN receive packet is available. Create a SKB for the data, and
 *   give it to the operating system.
 *-F*************************************************************************/
void
GpLanRx(
    adapter_t      *pAdapter,
    void           *pHeader,
    u32             headerSize,
    void           *pData,
    u32             dataSize,
	gsi_lan_t		*pLan	/* This is the lan interface on which we have got some packet */
    )
{
    DBG_FUNCX("GpLanRx", DBG_LAN_SYS)
    u32     pktlen;

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pAdapter", ("(0x%p)", pAdapter));
    DBG_PARAMF(DbgInfo, "pHeader", ("(0x%p)", pHeader));
    DBG_PARAMF(DbgInfo, "headerSize", ("%d", headerSize));
    DBG_PARAMF(DbgInfo, "pData", ("(0x%p)", pData));
    DBG_PARAMF(DbgInfo, "dataSize", ("%d", dataSize));

    pktlen = (headerSize + dataSize);
    DBG_RX(DbgInfo, ("###### RxLan %d bytes\n", pktlen));

    if (pktlen != 0)
    {
        struct sk_buff *skb;

        if ((skb = dev_alloc_skb(pktlen + 2)) != NULL)
        {
            skb->dev = pLan->dev;
            skb_reserve(skb, 2); /* align IP on 16 bit boundary */

            memcpy(skb_put(skb, headerSize), pHeader, headerSize);
            memcpy(skb_put(skb, dataSize), pData, dataSize);

            skb->protocol = eth_type_trans(skb, skb->dev);
            netif_rx(skb);

            pLan->stats.rx_packets++;
            pLan->stats.rx_bytes += pktlen;
        }
        else
        {
            DBG_WARNING(DbgInfo, ("Unable to allocate Rx buffer.\n"));
            pLan->stats.rx_dropped++;
        }
    }

    DBG_LEAVE(DbgInfo);
}

/*+F*************************************************************************
 * Function:
 *   GpLanTxComplete
 *
 * Description:
 *   The driver/adapter have finished transmitting the packet. Update the
 *   LAN statistics, free the buffers, and tell the operating system that
 *   we can handle another buffer.
 *-F*************************************************************************/
void
GpLanTxComplete(
    adapter_t      *pAdapter,
    int             txStatus,
	gsi_lan_t		*pLan
    )
{
    DBG_FUNCX("GpLanTxComplete", DBG_LAN_SYS)
    unsigned long   flags;
    PNDIS_PACKET    pPacket;

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pAdapter", ("(0x%p)", pAdapter));
    DBG_PARAMF(DbgInfo, "txStatus", ("%d", txStatus));

	if(pLan == NULL)
	{
		return;
	}

    if ((pPacket = pLan->txPacket) != NULL)
    {

        PNDIS_BUFFER    pBuffer;
        PUCHAR          pMemory;
        UINT            length;

        pLan->txPacket = NULL;

        DBG_TX(DbgInfo, ("###### TxLan complete.\n"));

        NdisUnchainBufferAtFront(pPacket, &pBuffer);
        DBG_ASSERT(pBuffer);

        NdisQueryBuffer(pBuffer, (PVOID *) &pMemory, &length);
        DBG_ASSERT(pMemory);
        DBG_ASSERT(length);

        if (txStatus == 0)
        {
            pLan->stats.tx_packets++;
            pLan->stats.tx_bytes += length;
        }
        else
        {
            pLan->stats.tx_errors++;
        }

        NdisFreePacket(pPacket);
        NdisFreeBuffer(pBuffer);
        spin_lock_irqsave(&(pLan->slock), flags);
        pLan->txAvailable++;

        // Tell the upper layers that we are ready for another packet.
        if ((pLan->flags & _TX_FULL) && (pLan->txAvailable > 0))
        {
            pLan->flags &= ~_TX_FULL;
            spin_unlock_irqrestore(&(pLan->slock), flags);
            netif_wake_queue(pLan->dev);
        }
        else
        {
            spin_unlock_irqrestore(&(pLan->slock), flags);
        }
    }

    DBG_LEAVE(DbgInfo);
}

/*+F*************************************************************************
 * Function:
 *   GpLanMediaChange
 *
 * Description:
 *   The adapter is notifying the LAN subsystem that the link state has
 *   changed. What we care about here is if the link is now available, and
 *   there are buffers ready, we need to tell the operating system that we
 *   are ready to handle more transmit requests.
 *-F*************************************************************************/
void
GpLanMediaChange(
    adapter_t      *pAdapter
    )
{
    DBG_FUNCX("GpLanMediaChange", DBG_LAN_SYS)
    gsi_lan_t      *pLan = &(pAdapter->lan);

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pAdapter", ("(0x%p)", pAdapter));

    if (pLan != NULL)
    {
        if (((PUSB_PHY)(pAdapter->UsbPhyHandle))->dspState == DSP_PHY_CONNECTED)
        {
			if (pLan->flags & _NO_CARRIER) 
			{
	            DBG_TX(DbgInfo, ("###### TxLan wakeup...\n"));
    	        pLan->flags &= ~_NO_CARRIER;
        	    netif_wake_queue(pLan->dev);
        	}
		}
		else
		{
	        DBG_TX(DbgInfo, ("###### TxLan Stopped...\n"));
			pLan->flags |= _NO_CARRIER;
		}
    }

    DBG_LEAVE(DbgInfo);
}

/*+F*************************************************************************
 * Function:
 *   GpLanOpen
 *
 * Description:
 *   Called by the LAN subsystem when the ethernet interface is configured
 *   make ready to accept data (ifconfig up).
 *-F*************************************************************************/
DBG_STATIC int
GpLanOpen(
    struct net_device  *dev
    )
{
	PUSB_PHY	UsbPhy = NULL;
	adapter_t	*pAdapter = NULL;
	NTSTATUS	NtStatus;
	int     	status = 0;
	int			ilanindex = 0;
	gsi_lan_t	*plan = NULL;
	
 	DBG_FUNCX("GpLanOpen", DBG_LAN_SYS)

	DBG_ENTER(DbgInfo);
	DBG_PARAMF(DbgInfo, "dev", ("%s (0x%p)", dev->name, dev));

	//pAdapter = (adapter_t *)dev->priv;
	plan = (gsi_lan_t *)dev->priv;
	pAdapter = plan->pAdapter;

	UsbPhy = (PUSB_PHY)pAdapter->UsbPhyHandle;

	if( (pAdapter->flags & _WAN_OPEN) != _WAN_OPEN )
	{
		/* Submit the URBs */
		if (UsbPhy->BulkReceivePipeHandle)
		{
			//
			// Start the Bulk Receive pipe
			//
			NtStatus = UsbPhyStartBulkReceive( UsbPhy);

			if (!NT_SUCCESS(NtStatus))
			{
				DBGPRINT(
					DBG_PRINT_MODULE_USB_ALL,
					DBG_PRINT_LEVEL_ERR,
					("Start Bulk Receive failed\n"));
				status = -EIO;
				goto LAN_OPEN_EXIT;
			} // if
		}

		if (UsbPhy->IsoReceivePipeHandle)
		{
			//
			// Start the Iso Receive pipe
			//
			NtStatus = UsbPhyStartIsoReceive(UsbPhy);
			if (!NT_SUCCESS(NtStatus))
			{
				DBGPRINT(
					DBG_PRINT_MODULE_USB_ALL,
					DBG_PRINT_LEVEL_ERR,
					("Start Iso Receive failed\n"));
				
				if (UsbPhy->BulkReceivePipeHandle)
					UsbPhyStopBulkReceive(UsbPhy);
				status = -EIO;
				goto LAN_OPEN_EXIT;
			} // if
		}

		//
		// Start the Bulk Transmit pipe
		//
		NtStatus = UsbPhyStartBulkTransmit( UsbPhy);

		if (!NT_SUCCESS(NtStatus))
		{
			DBGPRINT(
				DBG_PRINT_MODULE_USB_ALL,
				DBG_PRINT_LEVEL_ERR,
				("Start Bulk Transmit failed\n"));
			
			if (UsbPhy->BulkReceivePipeHandle)
				UsbPhyStopBulkReceive(UsbPhy);
			if (UsbPhy->IsoReceivePipeHandle)
				UsbPhyStopIsoReceive(UsbPhy);
			status = -EIO;
			goto LAN_OPEN_EXIT;
		} // if
	}
	/* Start Tx queuing */
	netdevice_start(dev);
	netif_start_queue(dev);
	pAdapter->flags |= _LAN_OPEN;

	/*
	**	Lets increment the Link up count to indicate that atleast one of the interface is UP.
	*/
	pAdapter->LinkUpCount++;

	
    MOD_INC_USE_COUNT;
	
LAN_OPEN_EXIT:
	DBG_RETURNF(DbgInfo, ("status %d", status));
	return status;
}

/*+F*************************************************************************
 * Function:
 *   GpLanClose
 *
 * Description:
 *   Called by the LAN subsystem when the ethernet interface is disabled
 *   (ifconfig down).
 *-F*************************************************************************/
DBG_STATIC int
GpLanClose(struct net_device  *dev)
{
	adapter_t	*pAdapter = NULL;
	PUSB_PHY	UsbPhy = NULL;
	int			status = 0;
	int			ilanindex = 0;
	gsi_lan_t	*plan = NULL;

	DBG_FUNCX("GpLanClose", DBG_LAN_SYS)

	DBG_ENTER(DbgInfo);
	DBG_PARAMF(DbgInfo, "dev", ("%s (0x%p)", dev->name, dev));

	//pAdapter = dev->priv;
	plan = (gsi_lan_t *)dev->priv;
	pAdapter = plan->pAdapter;
	
	/*
	**	Lets increment the Link up count to indicate that atleast one of the interface is UP.
	*/
	if (pAdapter)
		pAdapter->LinkUpCount--;

	if ((pAdapter) && (!pAdapter->LinkUpCount))
	{
    	UsbPhy = (PUSB_PHY)pAdapter->UsbPhyHandle;
		// abort all the pipes to clear all URBs out of the queue.
		DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE, ("Aborting all pipes\n"));
		if( (pAdapter->flags & _WAN_OPEN ) != _WAN_OPEN )
		{
			UsbPhyStopBulkTransmit(UsbPhy);

			if (UsbPhy->BulkReceivePipeHandle)
				UsbPhyStopBulkReceive(UsbPhy);

			if (UsbPhy->IsoReceivePipeHandle)
				UsbPhyStopIsoReceive(UsbPhy);
		}
		set_current_state( TASK_INTERRUPTIBLE );
		if(!in_interrupt())
			schedule_timeout( HZ * 3 );
	}

	/* Stop Tx queuing */
	netif_stop_queue(dev);
	netdevice_stop(dev);

	DBG_PRINT(DbgInfo, ("%s: Shutting down adapter.\n", dev->name));

	pAdapter->flags &= ~_LAN_OPEN;
    MOD_DEC_USE_COUNT;
	DBG_RETURNF(DbgInfo, ("status %d", status));
	return status;
}

/*+F*************************************************************************
 * Function:
 *   GpLanIoctl
 *
 * Description:
 *   LAN interface IOCTL entry point.
 *-F*************************************************************************/
DBG_STATIC int
GpLanIoctl(
    struct net_device  *dev,
    struct ifreq       *rq,
    int                 cmd
    )
{
    DBG_FUNC("GpLanIoctl")
    adapter_t  *pAdapter;
	gsi_lan_t  *plan = NULL;
    int         status = 0;

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "dev", ("%s (0x%p)", dev->name, dev));
    DBG_PARAMF(DbgInfo, "rq", ("(0x%p)", rq));
    DBG_PARAMF(DbgInfo, "cmd", ("%s (0x%04x)", DbgIoctlCmd(cmd), cmd));

	plan = (gsi_lan_t *)dev->priv;
	pAdapter = plan->pAdapter;

    if (pAdapter != NULL)
    {
        struct gsi_lan_req *pLanReq = (struct gsi_lan_req *) rq;

        if (!GpatmisIoctl(pAdapter, cmd, &(pLanReq->request), &status))
        {
            switch (cmd)
            {
            default:
                DBG_WARNING(DbgInfo, ("Unhandled IOCTL command: %s (0x%x)\n",
                    DbgIoctlCmd(cmd), cmd));
                status = -EOPNOTSUPP;
                break;
            }
        }
    }
    else
    {
        DBG_WARNING(DbgInfo, ("No adapter for IOCTL command!\n"));
        status = -ENODEV;
    }

    DBG_RETURNF(DbgInfo, ("status %d", status));
    return status;
}

/*+F*************************************************************************
 * Function:
 *   GpLanTx
 *
 * Description:
 *   Handle a LAN interface buffer (SKB) transmit request. Allocate an NDIS
 *   buffer, copy the data into the NDIS packet, and give it to the RFC layer.
 *-F*************************************************************************/
DBG_STATIC int
GpLanTx(
    struct sk_buff     *skb,
    struct net_device  *dev
    )
{
    DBG_FUNCX("GpLanTx", DBG_LAN_SYS)

    adapter_t      *pAdapter = NULL;
    int             status = 0;
	int				ilanindex = 0;
	gsi_lan_t      *pLan = NULL ;
    unsigned long   flags;
    u8             *pData = NULL;
    PNDIS_BUFFER    pBuffer = NULL;
    PNDIS_PACKET    pPacket = NULL;
    NDIS_STATUS     nStatus;

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "skb", ("(0x%p)", skb));
    DBG_PARAMF(DbgInfo, "dev", ("%s (0x%p)", dev->name, dev));
	
	pLan = (gsi_lan_t *)dev->priv;
	pAdapter = pLan->pAdapter;

    if (pAdapter != NULL)
    {

#if DBG
        if (skb->next)
        {
            DBG_ERROR(DbgInfo, ("Transmit skb has next!\n"));
        }
#endif // DBG

        if (((PUSB_PHY)(pAdapter->UsbPhyHandle))->dspState == DSP_PHY_CONNECTED)
        {
            DBG_ASSERT(skb->len < GSI_LAN_NDIS_BUFFER_SIZE);

            spin_lock_irqsave(&(pLan->slock), flags);

            pLan->txAvailable--;
            if (pLan->txAvailable >= 0)
            {
                u8  head = pLan->txHead;

                pLan->txHead = (pLan->txHead + 1) % GSI_LAN_NUM_NDIS_PACKETS;
                pData = pLan->bufferPool[head];
                if (pLan->txAvailable == 0)
                {
            		DBG_ERROR(DbgInfo, ("TxBuff FULL\n"));
                    pLan->flags |= _TX_FULL;
                    netif_stop_queue(dev);
                }
                spin_unlock_irqrestore(&(pLan->slock), flags);
            }
            else
            {
                DBG_WARNING(DbgInfo, ("Allocate buffer failed\n"));
                pLan->txAvailable++;
                pLan->stats.tx_dropped++;
                status = -ENOMEM;
                netif_stop_queue(dev);
                spin_unlock_irqrestore(&(pLan->slock), flags);
                goto gsi_tx_exit;
            }

            NdisAllocateBuffer(&nStatus, &pBuffer, pLan->NdisBufferPoolHandle,
                pData, skb->len);
            if ((nStatus != NDIS_STATUS_SUCCESS) || (pBuffer == NULL))
            {
                DBG_WARNING(DbgInfo, ("NdisAllocateBuffer failed\n"));
                pLan->stats.tx_dropped++;
                status = -ENOMEM;
                goto gsi_tx_exit;
            }

            NdisAllocatePacket(&nStatus, &pPacket, pLan->NdisPacketPoolHandle);
            if ((nStatus != NDIS_STATUS_SUCCESS) || (pPacket == NULL))
            {
                DBG_WARNING(DbgInfo, ("NdisAllocatePacket failed\n"));
                pLan->stats.tx_dropped++;
                status = -ENOMEM;
                goto gsi_tx_exit;
            }

            NdisChainBufferAtFront(pPacket, pBuffer);
            *((PVOID *) pPacket->ProtocolReserved) = NULL;

            // Copy the SKB data into the NDIS packet.
            memcpy(pData, skb->data, skb->len);

            spin_lock_irqsave(&(pLan->slock), flags);
            DBG_TX(DbgInfo, ("###### TxLan %d bytes...\n", skb->len));
            pLan->txPacket = pPacket;
            dev->trans_start = jiffies;

            switch (NdisRfc1577Send(pAdapter, pPacket, 0, pLan))
            {
            case NDIS_STATUS_SUCCESS:
                DBG_TX(DbgInfo, ("###### TxLan complete.\n"));
				NdisMSendComplete(pAdapter, pPacket, NDIS_STATUS_SUCCESS,pLan);
                pPacket = NULL; pBuffer = NULL; pData = NULL;
                break;

            case NDIS_STATUS_PENDING:
                DBG_TX(DbgInfo, ("###### TxLan pending...\n"));
                break;

            default:
                DBG_TX(DbgInfo, ("###### TxLan failed!\n"));
                NdisMSendComplete(pAdapter, pPacket, NDIS_STATUS_FAILURE,pLan);
                pPacket = NULL; pBuffer = NULL; pData = NULL;
                status = -EIO;
                break;
            }
            spin_unlock_irqrestore(&(pLan->slock), flags);

        gsi_tx_exit:
            if (status != 0)
            {
                if (pPacket != NULL)
                {
                    NdisFreePacket(pPacket);
                }

                if (pBuffer != NULL)
                {
                    NdisFreeBuffer(pBuffer);
                }

                if (pData != NULL)
                {
                    spin_lock_irqsave(&(pLan->slock), flags);
                    pLan->txAvailable++;
                    spin_unlock_irqrestore(&(pLan->slock), flags);
                }
            }
        }
        else
        {
            DBG_TX(DbgInfo, ("###### TxLan carrier failed!\n"));
            pLan->stats.tx_carrier_errors++;
            pLan->flags |= _NO_CARRIER;

			/*
			**	Stop the network queue
			*/
            netif_stop_queue(dev);
            status = -ENOTCONN;
        }
    }  /* if ((pAdapter = (adapter_t *) dev->priv) != NULL) */
    else
    {
_TXERROR:
        DBG_WARNING(DbgInfo, ("No adapter for transmit!\n"));

		
		/*
		**	Stop the network queue
		*/
        netif_stop_queue(dev);
        status = -ENODEV;
    }

    if (0 == status)
    {
        dev_kfree_skb(skb);
    }

    DBG_RETURNF(DbgInfo, ("status %d", status));
    return status;
}

/*+F*************************************************************************
 * Function:
 *   GpLanMulticast
 *
 * Description:
 *   Handle the LAN subsystem multicast, broadcast, and promiscous updates.
 *   These values can be modified by using the ifconfig command or other
 *   applications like network sniffers.
 *-F*************************************************************************/
DBG_STATIC void
GpLanMulticast(
    struct net_device  *dev
    )
{
    DBG_FUNCX("GpLanMulticast", DBG_LAN_SYS)
    struct dev_mc_list *mclist;
    adapter_t  *pAdapter;
	gsi_lan_t	*plan = NULL;

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "dev", ("%s (0x%p)", dev->name, dev));

#if DBG
    if (_DBG_SYS(DbgInfo, DBG_PARAM_ON))
    {
        int     x;

        DBG_PRINT(DbgInfo, ("  flags: %s%s%s%s\n",
            (dev->flags & IFF_BROADCAST) ? "Broadcast " : "",
            (dev->flags & IFF_PROMISC) ? "Promiscous " : "",
            (dev->flags & IFF_MULTICAST) ? "Multicast " : "",
            (dev->flags & IFF_ALLMULTI) ? "All-Multicast" : ""));
        DBG_PRINT(DbgInfo, ("  mc_count: %d\n", dev->mc_count));
        for (x = 0, mclist = dev->mc_list; mclist && x < dev->mc_count;
             x++, mclist = mclist->next)
        {
            DBG_PRINT(DbgInfo, ("    %s (%d)\n",
                DbgHwAddr(mclist->dmi_addr), mclist->dmi_addrlen));
        }
    }
#endif // DBG

	plan = (gsi_lan_t *)dev->priv;
	pAdapter = plan->pAdapter;

    if (pAdapter != NULL)
    {
        pAdapter->Rfc1483PacketFilter |= NDIS_PACKET_TYPE_DIRECTED;

        if (dev->flags & IFF_BROADCAST)
        {
            pAdapter->Rfc1483PacketFilter |= NDIS_PACKET_TYPE_BROADCAST;
        }
        else
        {
            pAdapter->Rfc1483PacketFilter &= ~NDIS_PACKET_TYPE_BROADCAST;
        }

        if (dev->flags & IFF_PROMISC)
        {
            pAdapter->Rfc1483PacketFilter |= NDIS_PACKET_TYPE_PROMISCUOUS;
        }
        else
        {
            pAdapter->Rfc1483PacketFilter &= ~NDIS_PACKET_TYPE_PROMISCUOUS;
        }

        if (dev->flags & IFF_ALLMULTI)
        {
            pAdapter->Rfc1483PacketFilter |= NDIS_PACKET_TYPE_ALL_MULTICAST;
        }
        else
        {
            pAdapter->Rfc1483PacketFilter &= ~NDIS_PACKET_TYPE_ALL_MULTICAST;
        }

        if ((dev->flags & IFF_MULTICAST) && (dev->mc_count != 0))
        {
            int     mc_count = dev->mc_count;
            int     x;

            pAdapter->Rfc1483PacketFilter |= NDIS_PACKET_TYPE_MULTICAST;
            if (mc_count > RFC1483_MAX_MULTICAST_ADDRESSES)
            {
                pAdapter->Rfc1483PacketFilter |= NDIS_PACKET_TYPE_ALL_MULTICAST;
                mc_count = 0;
            }

            pAdapter->Rfc1483MulticastFilterSize = 0;
            for (x = 0, mclist = dev->mc_list;
                (x < mc_count) && (mclist != NULL);
                 x++, mclist = mclist->next)
            {
                memcpy(
                    &(pAdapter->Rfc1483MulticastFilter[x * ETH_LENGTH_OF_ADDRESS]),
                    mclist->dmi_addr, ETH_LENGTH_OF_ADDRESS);
                pAdapter->Rfc1483MulticastFilterSize += ETH_LENGTH_OF_ADDRESS;
            }
        }
        else
        {
            pAdapter->Rfc1483PacketFilter &= ~NDIS_PACKET_TYPE_MULTICAST;
        }
    }

    DBG_LEAVE(DbgInfo);
}

/*+F*************************************************************************
 * Function:
 *   GpLanStats
 *
 * Description:
 *   Return the current LAN device statistics.
 *-F*************************************************************************/
DBG_STATIC struct net_device_stats *
GpLanStats(
    struct net_device  *dev
    )
{
    DBG_FUNCX("GpLanStats", DBG_LAN_SYS)
    struct net_device_stats    *pStats = NULL;
	gsi_lan_t   *plan = NULL;
	int			ilanindex = 0;

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "dev", ("%s (0x%p)", dev->name, dev));
    
	plan = (gsi_lan_t *)dev->priv;
	pStats = &(plan->stats);

    DBG_RETURNF(DbgInfo, ("pStats (0x%p)", pStats));
    return pStats;
}

/*+F*************************************************************************
 * Function:
 *   DbgHwAddr
 *
 * Description:
 *   Convert a hardware ethernet address to a character string
 *-F*************************************************************************/
const char *
DbgHwAddr(
    u8     *hwAddr
    )
{
    static char     buffer[18];

    sprintf(buffer, "%02X:%02X:%02X:%02X:%02X:%02X",
        hwAddr[0], hwAddr[1], hwAddr[2], hwAddr[3], hwAddr[4], hwAddr[5]);

    return buffer;
}
