/*+M*************************************************************************
 * GlobeSpan ADSL WAN Interface
 *
 * This driver bypasses the flip buffer usage since this driver only accepts
 * synchronous PPP data. The flip buffer only holds 512 bytes at a time,
 * so the driver maintains it's own buffer that can hold a whole ethernet
 * frame.
 *
 * Copyright (c) 2000 GlobeSpan Semiconductor Inc. -- http://www.globespan.net
 * All rights reserved.
 *
 * Developed for GlobeSpan Semiconductor Inc. by TriplePoint, Inc.
 *   http://www.triplepoint.com
 *-M*************************************************************************/
#ifdef MODULE
#include <linux/module.h>
#endif /* MODULE */

#include <Gp.h>
#include <gsi_ioctl.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/serial_reg.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>

#ifdef	USB_STEP
extern	int	g_break;
#endif

#define GSI_WAN_MAX_ADAPTERS    8

#define GSI_TTY_MAJOR           43
#define GSI_TTYAUX_MAJOR        44

#define GSI_WAN_BH                  CYCLADES_BH
#define GSI_WAN_EVENT_WRITE_WAKEUP  0

#define GSI_WAN_DEFAULT_FLAGS   ASYNC_LOW_LATENCY

DBG_STATIC int GpWanOpen(struct tty_struct *tty, struct file *pFile);
DBG_STATIC void GpWanClose(struct tty_struct *tty, struct file *pFile);
DBG_STATIC int GpWanTx(struct tty_struct *tty, int fromUser,
    const unsigned char *buffer, int bufferLen);
DBG_STATIC int GpWanIoctl(struct tty_struct *tty, struct file *pFile,
    unsigned int cmd, unsigned long arg);
DBG_STATIC int GpWanWriteRoom(struct tty_struct *tty);
DBG_STATIC int GpWanCharsInBuffer(struct tty_struct *tty);
DBG_STATIC void GpWanHangup(struct tty_struct *tty);
DBG_STATIC void GpWanBh(void);
DBG_STATIC void GpWanSoftint(void *context);
DBG_STATIC void GpWanScheduleEvent(channel_t *pChan, unsigned int event);

#if DBG
const char *DbgIoctlCmd(unsigned int cmd);
#endif // DBG

#ifdef HEXDUMPWAN
void HexDumpWan(unsigned char *pBuffer, unsigned long nSize);
#endif // HEXDUMPWAN

DBG_STATIC DECLARE_TASK_QUEUE(tq_gsi_wan);


/*+F*************************************************************************
 * Function:
 *   GpWanAlloc
 *
 * Description:
 *   Allocate the WAN interface.
 *-F*************************************************************************/
int
GpWanAlloc(
    adapter_t      *pAdapter
    )
{
    DBG_FUNCX("GpWanAlloc", DBG_WAN_SYS)
    gsi_wan_t  *pWan;
    int         status = 0;

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pAdapter", ("(0x%p)", pAdapter));

    pWan = &(pAdapter->wan);
    memset(pWan, 0, sizeof(*pWan));

    DBG_RETURNF(DbgInfo, ("status %d", status));
    return status;
}

/*+F*************************************************************************
 * Function:
 *   GpWanFree
 *
 * Description:
 *   Free the WAN interface.
 *-F*************************************************************************/
void
GpWanFree(
    adapter_t      *pAdapter
    )
{
    DBG_FUNCX("GpWanFree", DBG_WAN_SYS)

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pAdapter", ("(0x%p)", pAdapter));

    // Nothing to free

    DBG_LEAVE(DbgInfo);
}

/*+F*************************************************************************
 * Function:
 *   GpWanRegister
 *
 * Description:
 *   Register the WAN interface of the adapter as a TTY device with the
 *   operating system.
 *-F*************************************************************************/
int
GpWanRegister(
    adapter_t      *pAdapter
    )
{
    DBG_FUNCX("GpWanRegister", DBG_WAN_SYS)
    gsi_wan_t  *pWan = &(pAdapter->wan);
    int     status = 0;
    int     x;

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pAdapter", ("(0x%p)", pAdapter));

	INIT_LIST_HEAD(&(tq_gsi_wan));
    init_bh(GSI_WAN_BH, GpWanBh);

    memset(&(pWan->tty), 0, sizeof(pWan->tty));
    pWan->tty.magic           = TTY_DRIVER_MAGIC;
    pWan->tty.driver_name     = "gsi_wan";
    pWan->tty.name            = GSI_WAN_TTY_NAME;
    pWan->tty.major           = GSI_TTY_MAJOR;
    pWan->tty.minor_start     = 0;
    pWan->tty.num             = GSI_WAN_MAX_CHANNELS;
    pWan->tty.type            = TTY_DRIVER_TYPE_SERIAL;
    pWan->tty.subtype         = SERIAL_TYPE_NORMAL;
    pWan->tty.init_termios    = tty_std_termios;
    pWan->tty.init_termios.c_cflag = B9600 | CS8 | CREAD | HUPCL | CLOCAL;
    pWan->tty.flags           = TTY_DRIVER_REAL_RAW;
    pWan->tty.refcount        = &(pWan->refcount);
    pWan->tty.table           = pWan->table;
    pWan->tty.termios         = pWan->termios;
    pWan->tty.termios_locked  = pWan->termios_locked;
    pWan->tty.driver_state    = pAdapter;

    pWan->tty.open            = GpWanOpen;
    pWan->tty.close           = GpWanClose;
    pWan->tty.write           = GpWanTx;
    pWan->tty.write_room      = GpWanWriteRoom;
    pWan->tty.chars_in_buffer = GpWanCharsInBuffer;
    pWan->tty.flush_chars     = NULL;
    pWan->tty.flush_buffer    = NULL;
    pWan->tty.ioctl           = GpWanIoctl;
    pWan->tty.set_termios     = NULL;
    pWan->tty.stop            = NULL;
    pWan->tty.start           = NULL;
    pWan->tty.throttle        = NULL;
    pWan->tty.unthrottle      = NULL;
    pWan->tty.hangup          = GpWanHangup;
    pWan->tty.break_ctl       = NULL;
    pWan->tty.send_xchar      = NULL;
    pWan->tty.wait_until_sent = NULL;
    pWan->tty.read_proc       = NULL;

    if (tty_register_driver(&(pWan->tty)))
    {
        DBG_WARNING(DbgInfo, ("Unabled to register TTY interface!\n"));
        status = -ENODEV;
        goto gsi_wan_register_exit;
    }
    pWan->flags |= _TTY_ALLOC;

    memset(&(pWan->cu), 0, sizeof(pWan->cu));
    pWan->cu                 = pWan->tty;
    pWan->cu.name            = GSI_WAN_CU_NAME;
    pWan->cu.major           = GSI_TTYAUX_MAJOR;
    pWan->cu.subtype         = SERIAL_TYPE_CALLOUT;
    pWan->cu.read_proc       = NULL;
    pWan->cu.proc_entry      = NULL;
    if (tty_register_driver(&(pWan->cu)))
    {
        DBG_WARNING(DbgInfo, ("Unabled to register CU interface!\n"));
        status = -ENODEV;
        goto gsi_wan_register_exit;
    }
    pWan->flags |= _CU_ALLOC;

    for (x = 0; x < GSI_WAN_MAX_CHANNELS; x++)
    {
        channel_t  *pChan;
	
		DBG_PRINT(DbgInfo, ("pAdapter->wan.channel[%d] in GpWanRegister\n",x));

        pChan = &(pAdapter->wan.channel[x]);

        memset(pChan, 0, sizeof(*pChan));

        pChan->pAdapter         = pAdapter;
        pChan->count            = 0;
        pChan->flags            = GSI_WAN_DEFAULT_FLAGS;
        pChan->tqueue.routine   = GpWanSoftint;
        pChan->tqueue.data      = pChan;
        init_waitqueue_head( &(pChan->HangupCheckQ) );
        spin_lock_init(&(pChan->slock));
    }

gsi_wan_register_exit:
    if (status != 0)
    {
        GpWanUnregister(pAdapter);
    }

    DBG_RETURNF(DbgInfo, ("status %d", status));
    return status;
}

/*+F*************************************************************************
 * Function:
 *   GpWanUnregister
 *
 * Description:
 *   Unregister the TTY interface of the adapter with the operating system.
 *-F*************************************************************************/
void
GpWanUnregister(
    adapter_t      *pAdapter
    )
{
    DBG_FUNCX("GpWanUnregister", DBG_WAN_SYS)
    gsi_wan_t  *pWan = &(pAdapter->wan);
	int	error = 0;

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pAdapter", ("(0x%p)", pAdapter));
    DBG_PARAMF(DbgInfo, "TTY RefCount = ", ("(0x%x)", pWan->refcount));
	
	/* Close all ppp sessions by indicating hangup */
	GpWanMediaChange(pAdapter);

    remove_bh(GSI_WAN_BH);
#ifdef	USB_STEP
	//g_break = 1;
#endif
    if (pWan->flags & _CU_ALLOC)
    {
        pWan->flags &= ~_CU_ALLOC;
        if((error = tty_unregister_driver(&(pWan->cu))))
		{
    		DBG_PRINT(DbgInfo, ("Unable to deregister CUG Err = %d\n", error));
    		printk("Unable to deregister CUG Err = %d\n", error);
		}
    }
    if (pWan->flags & _TTY_ALLOC)
    {
        pWan->flags &= ~_TTY_ALLOC;
        if((error = tty_unregister_driver(&(pWan->tty))))
		{
    		DBG_PRINT(DbgInfo, ("Unable to deregister TTY Err = %d\n", error));
    		printk("Unable to deregister TTY Err = %d\n", error);
		}
    }

    DBG_LEAVE(DbgInfo);
}

/*+F*************************************************************************
 * Function:
 *   GpWanTx
 *
 * Description:
 *   Send a buffer to the adapter. It is assumed that the data has already
 *   been formatted as a synchronous PPP packet. This can be assured by
 *   setting the sync flag to PPP. Since the adapter does not need the PPP
 *   Address/Control bytes, they are stripped before the packet is sent down
 *   to the adapter.
 *-F*************************************************************************/
DBG_STATIC int
GpWanTx(
    struct tty_struct      *tty,
    int                     fromUser,
    const unsigned char    *buffer,
    int                     bufferLen
    )
{
    DBG_FUNCX("GpWanTx", DBG_WAN_SYS)
    channel_t  *pChan;
    int         length = 0;

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "tty", ("(0x%p)", tty));
    DBG_PARAMF(DbgInfo, "fromUser", ("%d", fromUser));
    DBG_PARAMF(DbgInfo, "buffer", ("(0x%p)", buffer));
    DBG_PARAMF(DbgInfo, "bufferLen", ("%d", bufferLen));

#ifdef	USB_STEP
	if( g_break )
	{
    	DBG_PRINT(DbgInfo, ("WAN CHECK 1\n"));
		return 0;
	}
#endif

    if ((pChan = (channel_t *) tty->driver_data) != NULL)
    {
        adapter_t  *pAdapter;

        pAdapter = pChan->pAdapter;
        DBG_ASSERT(pAdapter != NULL);

        if (((PUSB_PHY)(pAdapter->UsbPhyHandle))->dspState == DSP_PHY_CONNECTED)
        {
            unsigned long   flags;
            u8             *pData = NULL;
            int             dataLen;
            PNDIS_BUFFER    pBuffer = NULL;
            PNDIS_PACKET    pPacket = NULL;
            NDIS_STATUS     nStatus;

            DBG_ASSERT(bufferLen < GSI_WAN_NDIS_BUFFER_SIZE);

            if (bufferLen == 0)
            {
                goto gsi_wan_write_exit;
            }

            pChan->txAvailable--;
            if (pChan->txAvailable >= 0)
            {
                u8  head = pChan->txHead;

                pChan->txHead = (pChan->txHead + 1) % GSI_WAN_NUM_NDIS_PACKETS;
                pData = pChan->bufferPool[head];
            }
            else
            {
                DBG_TX(DbgInfo, ("###### TxWan no buffer available...\n"));
                pChan->txAvailable++;
                goto gsi_wan_write_exit;
            }

            dataLen = bufferLen;
            NdisAllocateBuffer(&nStatus, &pBuffer, pChan->NdisBufferPoolHandle,
                pData, dataLen);
            if ((nStatus != NDIS_STATUS_SUCCESS) || (pBuffer == NULL))
            {
                DBG_ERROR(DbgInfo, ("NdisAllocateBuffer failed\n"));
                length = -ENOMEM;
                goto gsi_wan_write_exit;
            }

            NdisAllocatePacket(&nStatus, &pPacket, pChan->NdisPacketPoolHandle);
            if ((nStatus != NDIS_STATUS_SUCCESS) || (pPacket == NULL))
            {
                DBG_ERROR(DbgInfo, ("NdisAllocatePacket failed\n"));
                length = -ENOMEM;
                goto gsi_wan_write_exit;
            }

            NdisChainBufferAtFront(pPacket, pBuffer);
            *((PVOID *) pPacket->ProtocolReserved) = pChan;

            // Copy the data into the NDIS packet if it is coming from
            // the user address space.
#ifdef	HEXDUMPWAN
			HexDumpWan(buffer,bufferLen);
#endif
            if (fromUser)
            {
                if (copy_from_user(pData, buffer, bufferLen))
                {
                    DBG_ERROR(DbgInfo, ("copy_from_user failed\n"));
                    length = -EFAULT;
                    goto gsi_wan_write_exit;
                }
                buffer = pData;
            }

            // Strip of the Address/Control bytes from the sync PPP data.
            // Data Striping is done in rfc2364 code
            dataLen = bufferLen - 2;
            memcpy(pData, buffer + 2, dataLen);

            spin_lock_irqsave(&(pChan->slock), flags);
            DBG_TX(DbgInfo, ("###### TxWan %d bytes...\n", dataLen));
            pChan->txPacket = pPacket;
            pChan->txPacketLen = dataLen;
            pChan->txBytes += pChan->txPacketLen;
            switch (Rfc2364Send(pAdapter, pAdapter, pPacket))
            {
            case NDIS_STATUS_SUCCESS:
                DBG_TX(DbgInfo, ("###### TxWan complete.\n"));
                NdisMWanSendComplete(pAdapter, pPacket, NDIS_STATUS_SUCCESS, NULL);
                pPacket = NULL; pBuffer = NULL; pData = NULL;
                length = bufferLen;
                break;

            case NDIS_STATUS_PENDING:
                DBG_TX(DbgInfo, ("###### TxWan pending...\n"));
                length = bufferLen;
                break;

            default:
                DBG_TX(DbgInfo, ("###### TxWan failed!\n"));
                NdisMWanSendComplete(pAdapter, pPacket, NDIS_STATUS_FAILURE, NULL);
                pPacket = NULL; pBuffer = NULL; pData = NULL;
                length = -EIO;
                break;
            }
            spin_unlock_irqrestore(&(pChan->slock), flags);

        gsi_wan_write_exit:
            if (length < 0)
            {
                if (pPacket != NULL)
                {
                    NdisFreePacket(pPacket);
                }

                if (pBuffer != NULL)
                {
                    NdisFreeBuffer(pBuffer);
                }

                if (pData != NULL)
                {
                    pChan->txAvailable++;
                }
            }
        }
        else
        {
            DBG_TX(DbgInfo, ("###### TxWan carrier failed!\n"));
            length = -ENOTCONN;
        }
    }
    DBG_ASSERT(pChan != NULL);

    DBG_RETURNF(DbgInfo, ("length %d", length));
    return length;
}

/*+F*************************************************************************
 * Function:
 *   GpWanTxComplete
 *
 * Description:
 *   The previous transmit request has finished. Release the buffer(s) back
 *   to the free list and tell the TTY subsystem that we can accept more
 *   transmit data.
 *-F*************************************************************************/
void
GpWanTxComplete(
    adapter_t      *pAdapter,
    channel_t      *pChan,
    int             status
    )
{
    DBG_FUNCX("GpWanTxComplete", DBG_WAN_SYS)
    PNDIS_PACKET    pPacket;
    unsigned long   flags;

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pAdapter", ("(0x%p)", pAdapter));
    DBG_PARAMF(DbgInfo, "pChan", ("(0x%p)", pChan));
    DBG_PARAMF(DbgInfo, "status", ("%d", status));

#ifdef	USB_STEP
	if( g_break )
	{
    	DBG_PRINT(DbgInfo, ("WAN CHECK 2\n"));
		return ;
	}
#endif


    spin_lock_irqsave(&(pChan->slock), flags);

    if ((pPacket = pChan->txPacket) != NULL)
    {
        PNDIS_BUFFER    pBuffer;
        PUCHAR          pMemory;
        UINT            length;

        pChan->txPacket = NULL;

        DBG_TX(DbgInfo, ("###### TxWan complete.\n"));

        NdisUnchainBufferAtFront(pPacket, &pBuffer);
        DBG_ASSERT(pBuffer);

        NdisQueryBuffer(pBuffer, (PVOID *) &pMemory, &length);
        DBG_ASSERT(pMemory);
        DBG_ASSERT(length);

        NdisFreePacket(pPacket);
        NdisFreeBuffer(pBuffer);
        pChan->txAvailable++;
        pChan->txBytes -= pChan->txPacketLen;
        pChan->txPacketLen = 0;
        DBG_ASSERT(pChan->txBytes >= 0);

        spin_unlock_irqrestore(&(pChan->slock), flags);

        // Tell the upper layers that we are ready for another packet.
        GpWanScheduleEvent(pChan, GSI_WAN_EVENT_WRITE_WAKEUP);
    }
    else
    {
        spin_unlock_irqrestore(&(pChan->slock), flags);
    }

    DBG_LEAVE(DbgInfo);
}

/*+F*************************************************************************
 * Function:
 *   GpWanMediaChange
 *
 * Description:
 *   Used by the adapter to inform this interface layer that link state
 *   has changed.
 *-F*************************************************************************/
void
GpWanMediaChange(
    adapter_t      *pAdapter
    )
{
    DBG_FUNCX("GpWanMediaChange", DBG_WAN_SYS)
    gsi_wan_t      *pWan = NULL;
	int	ChannelCount,RefCountWait = 0;

#ifdef	USB_STEP
	if( g_break )
	{
    	DBG_PRINT(DbgInfo, ("WAN CHECK 3\n"));
		return ;
	}
#endif

    pWan = &(pAdapter->wan);

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pAdapter", ("(0x%p)", pAdapter));

	//Hangup each port on the adapter. 
    for (ChannelCount = 0; ChannelCount < GSI_WAN_MAX_CHANNELS; ChannelCount++)
    {
        channel_t  *pChan;
	
		DBG_PRINT(DbgInfo, ("pAdapter->wan.channel[%d] in GpWanRegister\n",
			ChannelCount));

        pChan = &(pAdapter->wan.channel[ChannelCount]);

		if( pChan && pChan->count && pChan->tty )
		{
			DBG_PRINT(DbgInfo, ("Hangup called for ch[%d]\n",ChannelCount));
			tty_hangup(pChan->tty);
			//Wait for the TTY to get disconnected
			interruptible_sleep_on(&(pChan->HangupCheckQ));
			//Wait for some arbitrary time to get refcount reach zero
			for(RefCountWait = 0; 
				(RefCountWait < 20) && (pWan->refcount != 0); 
				RefCountWait++)
			{
    			DBG_PARAMF(DbgInfo, "LOOP TTY RefCount = ", ("(0x%x)", pWan->refcount));
				set_current_state( TASK_INTERRUPTIBLE );
				if(!in_interrupt())
					schedule_timeout( HZ );
			}
		}
    }

    DBG_LEAVE(DbgInfo);
}

/*+F*************************************************************************
 * Function:
 *   GpWanRx
 *
 * Description:
 *   Handle newly received data coming from the adapter. As the data comes
 *   from the adapter, it will be missing the Address/Control bytes that are
 *   needed for the complete synchronous PPP data packet. Those bytes need
 *   to be re-added to the packet before it is give up to the application.
 *-F*************************************************************************/
void
GpWanRx(
    adapter_t      *pAdapter,
    channel_t      *pChan,
    void           *pPacket,
    u32             packetSize
    )
{
    DBG_FUNCX("GpWanRx", DBG_WAN_SYS)
    struct tty_struct  *tty;
    unsigned long       flags;

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pAdapter", ("(0x%p)", pAdapter));
    DBG_PARAMF(DbgInfo, "pChan", ("(0x%p)", pChan));
    DBG_PARAMF(DbgInfo, "pPacket", ("(0x%p)", pPacket));
    DBG_PARAMF(DbgInfo, "packetSize", ("%d", packetSize));

#ifdef	USB_STEP
	if( g_break )
	{
    	DBG_PRINT(DbgInfo, ("WAN CHECK 4\n"));
		return ;
	}
#endif


    DBG_PRINT(DbgInfo, ("Apapter Handle in GpWanRx %p\n", pAdapter));
    DBG_PRINT(DbgInfo, ("NdisLink Context in GpWanRx %p\n", pChan));

    spin_lock_irqsave(&(pChan->slock), flags);

    if ((tty = pChan->tty) != NULL)
    {
        u8     *pData;
        u8     *pFlags;
        int     dataLen;

        DBG_RX(DbgInfo, ("###### RxWan %d bytes\n", packetSize));

        pData = pChan->rxData;
        pFlags = pChan->rxFlags;
        dataLen = sizeof(pChan->rxData);

        if ((packetSize + 2) < dataLen)
        {
            u8     *pPkt = (u8 *) pPacket;

            // Add on the Address/Control bytes if required. There is an
            // issue with the GlobeSpan CO card that it transmits these
            // bytes! So, if they are already present, don't add another
            // set. DWG
            if ((pPkt[0] == 0xFF) && (pPkt[1] == 0x03))
            {
                dataLen = packetSize;
                memcpy(pData, pPacket, packetSize);
                memset(pFlags, TTY_NORMAL, dataLen);
            }
            else
            {
                pData[0] = 0xFF; pData[1] = 0x03;

                dataLen = packetSize + 2;
                memcpy(pData + 2, pPacket, packetSize);
                memset(pFlags, TTY_NORMAL, dataLen);
            }

            spin_unlock_irqrestore(&(pChan->slock), flags);

            tty->ldisc.receive_buf(tty, pData, pFlags, dataLen);
            GpWanScheduleEvent(pChan, GSI_WAN_EVENT_WRITE_WAKEUP);
        }
        else
        {
            DBG_ERROR(DbgInfo, ("Packet size too large to accept\n"));
            spin_unlock_irqrestore(&(pChan->slock), flags);
        }
    }
    else
    {
        DBG_WARNING(DbgInfo, ("No channel TTY defined!\n"));
        spin_unlock_irqrestore(&(pChan->slock), flags);
    }

    DBG_LEAVE(DbgInfo);
}

/*+F*************************************************************************
 * Function:
 *   GpWanUpdateFlags
 *
 * Description:
 *   Routine to update the internal channel flags based on the state of
 *   the channel.
 *-F*************************************************************************/
DBG_STATIC void
GpWanUpdateFlags(
    channel_t  *pChan
    )
{
    DBG_FUNCX("GpWanUpdateFlags", DBG_WAN_SYS)

#ifdef	USB_STEP
	if( g_break )
	{
    	DBG_PRINT(DbgInfo, ("WAN CHECK 5\n"));
		return ;
	}
#endif

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pChan", ("(0x%p)", pChan));
	if( pChan )
  	  DBG_PARAMF(DbgInfo, "pTTY", ("(0x%p)", pChan->tty));

    if (C_BAUD(pChan->tty))
    {
        pChan->mcr |= UART_MCR_DTR;
    }
    else
    {
        pChan->mcr &= ~UART_MCR_DTR;
    }

    if (C_CRTSCTS(pChan->tty))
    {
        pChan->flags |= ASYNC_CTS_FLOW;
    }
    else
    {
        pChan->flags &= ~ASYNC_CTS_FLOW;
    }

    if (C_CLOCAL(pChan->tty))
    {
        pChan->flags &= ~ASYNC_CHECK_CD;
    }
    else
    {
        pChan->flags |= ASYNC_CHECK_CD;
    }

    DBG_LEAVE(DbgInfo);
}

/*+F*************************************************************************
 * Function:
 *   GpWanBh
 *
 * Description:
 *   Bottom half handler routine.
 *-F*************************************************************************/
DBG_STATIC void
GpWanBh(
    void
    )
{
    DBG_FUNCX("GpWanBh", DBG_WAN_SYS)

#ifdef	USB_STEP
	if( g_break )
	{
    	DBG_PRINT(DbgInfo, ("WAN CHECK 6\n"));
		return ;
	}
#endif


    DBG_ENTER(DbgInfo);

    run_task_queue(&tq_gsi_wan);

    DBG_LEAVE(DbgInfo);
}

/*+F*************************************************************************
 * Function:
 *   GpWanSoftint
 *
 * Description:
 *   Routine to actually process queued channel events.
 *-F*************************************************************************/
DBG_STATIC void
GpWanSoftint(
    void   *context
    )
{
    DBG_FUNCX("GpWanSoftint", DBG_WAN_SYS)
    channel_t          *pChan = (channel_t *) context;
    struct tty_struct  *tty;

#ifdef	USB_STEP
	if( g_break )
	{
    	DBG_PRINT(DbgInfo, ("WAN CHECK 7\n"));
		return ;
	}
#endif


    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "context", ("(0x%p)", context));

    if ((tty = pChan->tty) != NULL)
    {
        if (test_and_clear_bit(GSI_WAN_EVENT_WRITE_WAKEUP, &(pChan->event)))
        {
            if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
                (tty->ldisc.write_wakeup != NULL))
            {
                (tty->ldisc.write_wakeup)(tty);
            }
            wake_up_interruptible(&(tty->write_wait));
        }
    }

    DBG_LEAVE(DbgInfo);
}

/*+F*************************************************************************
 * Function:
 *   GpWanScheduleEvent
 *
 * Description:
 *   Schedule an event to be processed by the bottom half handler.
 *-F*************************************************************************/
DBG_STATIC void
GpWanScheduleEvent(
    channel_t      *pChan,
    unsigned int    event
    )
{
    DBG_FUNCX("GpWanScheduleEvent", DBG_WAN_SYS)

#ifdef	USB_STEP
	if( g_break )
	{
    	DBG_PRINT(DbgInfo, ("WAN CHECK 8\n"));
		return ;
	}
#endif

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pChan", ("(0x%p)", pChan));
    DBG_PARAMF(DbgInfo, "event", ("0x%08x", event));


    set_bit(event, &(pChan->event));
    queue_task(&(pChan->tqueue), &tq_gsi_wan);
    mark_bh(GSI_WAN_BH);

    DBG_LEAVE(DbgInfo);
}

/*+F*************************************************************************
 * Function:
 *   GpWanChannelStartup
 *
 * Description:
 *   Startup the given channel. This routine allocates the necessary Tx/Rx
 *   resources and makes the TTY ready for operation.
 *   NOTE: The channel spinlock needs to be held by calling routine.
 *-F*************************************************************************/
DBG_STATIC int
GpWanChannelStartup(
    channel_t  *pChan
    )
{
    DBG_FUNCX("GpWanChannelStartup", DBG_WAN_SYS)
    int             status = 0;
    NDIS_STATUS     nStatus;
    int             x;

#ifdef	USB_STEP
	if( g_break )
	{
    	DBG_PRINT(DbgInfo, ("WAN CHECK 9\n"));
		return -ENOMEM;
	}
#endif


    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pChan", ("(0x%p)", pChan));

    pChan->event = 0;

    pChan->txAvailable = GSI_WAN_NUM_NDIS_PACKETS;
    pChan->txHead = 0;
    pChan->txBytes = 0;
    for (x = 0; x < pChan->txAvailable; x++)
    {
        NdisAllocateMemoryWithTag((PVOID *) &(pChan->bufferPool[x]),
            GSI_WAN_NDIS_BUFFER_SIZE, GSI_WAN_BUFFER_TAG);
        if (pChan->bufferPool[x] == NULL)
        {
            DBG_ERROR(DbgInfo, ("NdisAllocateMemoryWithTag failed\n"));
            status = -ENOMEM;
            goto gsi_wan_startup_exit;
        }
    }

    NdisAllocatePacketPool(&nStatus, &(pChan->NdisPacketPoolHandle),
        GSI_WAN_NUM_NDIS_PACKETS, PROTOCOL_RESERVED_SIZE_IN_PACKET);
    if ((nStatus != NDIS_STATUS_SUCCESS) || (pChan->NdisPacketPoolHandle == NULL))
    {
        DBG_ERROR(DbgInfo, ("NdisAllocatePacketPool failed!\n"));
        status = -ENOMEM;
        goto gsi_wan_startup_exit;
    }

    NdisAllocateBufferPool(&nStatus, &(pChan->NdisBufferPoolHandle),
        GSI_WAN_NUM_NDIS_PACKETS);
    if ((nStatus != NDIS_STATUS_SUCCESS) || (pChan->NdisBufferPoolHandle == NULL))
    {
        DBG_ERROR(DbgInfo, ("NdisAllocateBufferPool failed!\n"));
        status = -ENOMEM;
        goto gsi_wan_startup_exit;
    }

    if (!(pChan->flags & ASYNC_INITIALIZED))
    {
        pChan->mcr = (UART_MCR_DTR | UART_MCR_RTS | UART_MCR_OUT2);
        if (pChan->tty)
        {
            clear_bit(TTY_IO_ERROR, &(pChan->tty->flags));
        }

        GpWanUpdateFlags(pChan);

        pChan->flags |= ASYNC_INITIALIZED;
        pChan->msr |= (UART_MSR_DSR | UART_MSR_CTS);
    }

gsi_wan_startup_exit:
    DBG_RETURNF(DbgInfo, ("status %d", status));
    return status;
}

/*+F*************************************************************************
 * Function:
 *   GpWanChannelShutdown
 *
 * Description:
 *   Mark the TTY as down and not ready to receive characters, and release
 *   the Tx/Rx resources.
 *   NOTE: The channel spinlock needs to be held by calling routine.
 *-F*************************************************************************/
DBG_STATIC void
GpWanChannelShutdown(
    channel_t  *pChan
    )
{
    DBG_FUNCX("GpWanChannelShutdown", DBG_WAN_SYS)
    int     x;

#ifdef	USB_STEP
	if( g_break )
	{
    	DBG_PRINT(DbgInfo, ("WAN CHECK a\n"));
		return;
	}
#endif


    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pChan", ("(0x%p)", pChan));

    if (pChan->flags & ASYNC_INITIALIZED)
    {
        pChan->msr &= ~UART_MSR_RI;
        if (!pChan->tty || C_HUPCL(pChan->tty))
        {
            pChan->mcr &= ~(UART_MCR_DTR | UART_MCR_RTS);
        }

        if (pChan->tty)
        {
            set_bit(TTY_IO_ERROR, &(pChan->tty->flags));
        }

        pChan->flags &= ~ASYNC_INITIALIZED;
    }

    for (x = 0; x < pChan->txAvailable; x++)
    {
        if (pChan->bufferPool[x] != NULL)
        {
            NdisFreeMemory(pChan->bufferPool[x], GSI_WAN_NDIS_BUFFER_SIZE, 0);
            pChan->bufferPool[x] = NULL;
        }
    }
    pChan->txAvailable = 0;
    pChan->event = 0;

    if (pChan->NdisBufferPoolHandle != NULL)
    {
        NdisFreeBufferPool(pChan->NdisBufferPoolHandle);
        pChan->NdisBufferPoolHandle = NULL;
    }

    if (pChan->NdisPacketPoolHandle != NULL)
    {
        NdisFreePacketPool(pChan->NdisPacketPoolHandle);
        pChan->NdisPacketPoolHandle = NULL;
    }

    DBG_LEAVE(DbgInfo);
}

/*+F*************************************************************************
 * Function:
 *   GpWanOpen
 *
 * Description:
 *   Entry point for TTY device open request from the application.
 *-F*************************************************************************/
DBG_STATIC int
GpWanOpen(
    struct tty_struct  *tty,
    struct file        *pFile
    )
{
    DBG_FUNCX("GpWanOpen", DBG_WAN_SYS)
    adapter_t  *pAdapter;
    PUSB_PHY	UsbPhy = NULL;
    channel_t  *pChan;
	NTSTATUS	NtStatus;
    int     status = 0;
    int     line;

#ifdef	USB_STEP
	if( g_break )
	{
    	DBG_PRINT(DbgInfo, ("WAN CHECK b\n"));
		return -ENODEV;
	}
#endif

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "tty", ("(0x%p)", tty));
    DBG_PARAMF(DbgInfo, "pFile", ("(0x%p)", pFile));


    line = MINOR(tty->device) - tty->driver.minor_start;
    if ((line < 0) || (line >= GSI_WAN_MAX_CHANNELS))
    {
        DBG_WARNING(DbgInfo, ("Invalid tty line %d!\n", line));
        status = -ENODEV;
        goto gsi_wan_open_exit;
    }

    if ((pAdapter = (adapter_t *) tty->driver.driver_state) == NULL)
    {
        DBG_WARNING(DbgInfo, ("No adapter in tty structure!\n"));
        status = -ENODEV;
        goto gsi_wan_open_exit;
    }

    pChan = &(pAdapter->wan.channel[line]);
    if (pChan != NULL)
    {
	unsigned long   flags;

	spin_lock_irqsave(&(pChan->slock), flags);

	tty->driver_data = pChan;
	tty->low_latency = (pChan->flags & ASYNC_LOW_LATENCY) ? 1 : 0;

	pChan->count++;
	if (pChan->count == 1)
	{
		pAdapter->Call = &pAdapter->LinkCall;
		pAdapter->LinkCall.VPI = pAdapter->Rfc2364Vpi;
		pAdapter->LinkCall.VCI = pAdapter->Rfc2364Vci;
      	if (Rfc2364CreateVc(pAdapter, pAdapter->Call) != NDIS_STATUS_SUCCESS)
      	{       
      		DBG_WARNING(DbgInfo,("Rfc2364CreateVc failed!\n"));
      		status = -ENODEV;
      		spin_unlock_irqrestore(&(pChan->slock), flags);
      		goto gsi_wan_open_exit;
      	}
		pAdapter->Call->NdisLinkContext = &(pAdapter->wan.channel[line]);
      	pAdapter->flags |= _RFC2364_VC_ALLOC;
		pChan->tty = tty;

		status = GpWanChannelStartup(pChan);
		if (status != 0)
		{
			DBG_ERROR(DbgInfo, ("gsi_wan_channel_startup failed!\n"));
			spin_unlock_irqrestore(&(pChan->slock), flags);
			goto gsi_wan_open_exit;
		}

		spin_unlock_irqrestore(&(pChan->slock), flags);

		UsbPhy = ((PUSB_PHY)(pAdapter->UsbPhyHandle));
		/* Submit the URBs */
		if(((pAdapter->flags & _LAN_OPEN) != _LAN_OPEN ) &&
			((pAdapter->flags & _WAN_OPEN) != _WAN_OPEN ) )
		{
			if (UsbPhy->BulkReceivePipeHandle)
			{
				// Start the Bulk Receive pipe
				NtStatus = UsbPhyStartBulkReceive( UsbPhy);

				if (!NT_SUCCESS(NtStatus))
				{
					DBGPRINT( DBG_PRINT_MODULE_RFC2364_ALL, DBG_PRINT_LEVEL_ERR,
						("Start Bulk Receive failed\n"));
					status = -EIO;
					goto gsi_wan_open_exit;
				}
			}

			if (UsbPhy->IsoReceivePipeHandle)
			{
				// Start the Iso Receive pipe
				NtStatus = UsbPhyStartIsoReceive(UsbPhy);
				if (!NT_SUCCESS(NtStatus))
				{
					DBGPRINT( DBG_PRINT_MODULE_RFC2364_ALL, DBG_PRINT_LEVEL_ERR,
							("Start Iso Receive failed\n"));
					if (UsbPhy->BulkReceivePipeHandle)
						UsbPhyStopBulkReceive(UsbPhy);
					status = -EIO;
					goto gsi_wan_open_exit;
				} // if
			}

			// Start the Bulk Transmit pipe
			NtStatus = UsbPhyStartBulkTransmit( UsbPhy);
			if (!NT_SUCCESS(NtStatus))
			{
				DBGPRINT( DBG_PRINT_MODULE_RFC2364_ALL, DBG_PRINT_LEVEL_ERR,
						("Start Bulk Transmit failed\n"));
				if (UsbPhy->BulkReceivePipeHandle)
					UsbPhyStopBulkReceive(UsbPhy);
				if (UsbPhy->IsoReceivePipeHandle)
					UsbPhyStopIsoReceive(UsbPhy);
				status = -EIO;
				goto gsi_wan_open_exit;
			} // if
		}
		
		if (pChan->flags & ASYNC_SPLIT_TERMIOS)
		{
			*tty->termios = (tty->driver.subtype == SERIAL_TYPE_NORMAL) ?
			pChan->normal_termios : pChan->callout_termios;
			GpWanUpdateFlags(pChan);
		}
        }
        else
        {
		spin_unlock_irqrestore(&(pChan->slock), flags);
        }
    }
    DBG_ASSERT(pChan != NULL);

    pAdapter->flags |= _WAN_OPEN;
	MOD_INC_USE_COUNT;
    
gsi_wan_open_exit:

    DBG_RETURNF(DbgInfo, ("status %d", status));
    return status;
}

/*+F*************************************************************************
 * Function:
 *   GpWanClose
 *
 * Description:
 *   Entry point for TTY device close request from the application.
 *-F*************************************************************************/
DBG_STATIC void
GpWanClose(
    struct tty_struct  *tty,
    struct file        *pFile
    )
{
    channel_t	*pChan = NULL;
    adapter_t	*pAdapter = NULL;
    PUSB_PHY	UsbPhy = NULL;
    
    DBG_FUNCX("gsi_wan_close", DBG_WAN_SYS)
#ifdef	USB_STEP
	if( g_break )
	{
    	DBG_PRINT(DbgInfo, ("WAN CHECK c\n"));
		return ;
	}
#endif

    DBG_ENTER(DbgInfo);

    DBG_PARAMF(DbgInfo, "tty", ("(0x%p)", tty));
    DBG_PARAMF(DbgInfo, "ttyname", ("(%s)", tty->driver.name));
    DBG_PARAMF(DbgInfo, "pFile", ("(0x%p)", pFile));

	pAdapter = (adapter_t *) tty->driver.driver_state;

    if ((pChan = (channel_t *) tty->driver_data) != NULL)
    {
        unsigned long       flags;

        spin_lock_irqsave(&(pChan->slock), flags);

        pChan->count--;
        DBG_ASSERT(pChan->count >= 0);
        if (pChan->count == 0)
        {
            pChan->flags |= ASYNC_CLOSING;

            if (pChan->flags & ASYNC_NORMAL_ACTIVE)
            {
                pChan->normal_termios = *tty->termios;
            }

            if (pChan->flags & ASYNC_CALLOUT_ACTIVE)
            {
                pChan->callout_termios = *tty->termios;
            }

            GpWanChannelShutdown(pChan);

            if (tty->driver.flush_buffer)
            {
				DBG_PRINT(DbgInfo, ("Flushing Driver Buffers"));
                tty->driver.flush_buffer(tty);
            }
            if (tty->ldisc.flush_buffer)
            {
				DBG_PRINT(DbgInfo, ("Flushing TTY Buffers"));
                tty->ldisc.flush_buffer(tty);
            }

            pChan->tty = NULL;

            pChan->flags &= ~(ASYNC_NORMAL_ACTIVE | ASYNC_CALLOUT_ACTIVE | ASYNC_CLOSING);

		if ((pAdapter = (adapter_t *) tty->driver.driver_state))
		{
	    	UsbPhy = ((PUSB_PHY)(pAdapter->UsbPhyHandle));
			// abort all the pipes to clear all URBs out of the queue.
			if( (pAdapter->flags & _LAN_OPEN ) != _LAN_OPEN )
			{
				if(!(pAdapter->flags & _USB_BUS_CONNECTED))
					UsbPhyStopInterruptReceive( UsbPhy);
					
				UsbPhyStopBulkTransmit(UsbPhy);

				if (UsbPhy->BulkReceivePipeHandle)
				{
					UsbPhyStopBulkReceive(UsbPhy);
				}

				if (UsbPhy->IsoReceivePipeHandle)
				{
					UsbPhyStopIsoReceive(UsbPhy);
				}
			}
			set_current_state( TASK_INTERRUPTIBLE );
			
			if(!in_interrupt())
				schedule_timeout( HZ * 3 );
		}
      	if (pAdapter->flags & _RFC2364_VC_ALLOC)
      	{
      		DBG_PRINT(DbgInfo, ("**** RFC2364 VC Delete\n"));
      	    Rfc2364DeleteVc(pAdapter, 
      				pAdapter->Rfc2364Vpi, pAdapter->Rfc2364Vci);
      	    pAdapter->flags &= ~_RFC2364_VC_ALLOC;
      	}
		spin_unlock_irqrestore(&(pChan->slock), flags);
        }
        else
        {
            spin_unlock_irqrestore(&(pChan->slock), flags);
        }
		wake_up_interruptible(&(pChan->HangupCheckQ));
    }

    DBG_ASSERT(pChan != NULL);

    pAdapter->flags &= ~_WAN_OPEN;
	MOD_DEC_USE_COUNT;

    DBG_LEAVE(DbgInfo);
}

/*+F*************************************************************************
 * Function:
 *   GpWanGetModemInfo
 *
 * Description:
 *   Return the modem status flags to the application. The flags needed
 *   to be mapped from the kernel version of the flags into the application
 *   versions.
 *-F*************************************************************************/
DBG_STATIC int
GpWanGetModemInfo(
    channel_t      *pChan,
    unsigned int   *pArg
    )
{
    DBG_FUNCX("GpWanGetModemInfo", DBG_WAN_SYS)
    int         status = 0;
    unsigned int    flags;

#ifdef	USB_STEP
	if( g_break )
	{
    	DBG_PRINT(DbgInfo, ("WAN CHECK d\n"));
		return -ENODEV;
	}
#endif

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pChan", ("(0x%p)", pChan));
    DBG_PARAMF(DbgInfo, "pArg", ("(0x%p)", pArg));


    flags =
        ((pChan->mcr & UART_MCR_RTS) ? TIOCM_RTS : 0) |
        ((pChan->mcr & UART_MCR_DTR) ? TIOCM_DTR : 0) |
        ((pChan->msr & UART_MSR_DCD) ? TIOCM_CD  : 0) |
        ((pChan->msr & UART_MSR_RI)  ? TIOCM_RI  : 0) |
        ((pChan->msr & UART_MSR_DSR) ? TIOCM_DSR : 0) |
        ((pChan->msr & UART_MSR_CTS) ? TIOCM_CTS : 0);

#if DBG
    {
    char    flagStr[128];

    *flagStr = '\0';
    if (flags & TIOCM_LE) strcat(flagStr, "LE ");
    if (flags & TIOCM_DTR) strcat(flagStr, "DTR ");
    if (flags & TIOCM_RTS) strcat(flagStr, "RTS ");
    if (flags & TIOCM_ST) strcat(flagStr, "ST ");
    if (flags & TIOCM_SR) strcat(flagStr, "SR ");
    if (flags & TIOCM_CTS) strcat(flagStr, "CTS ");
    if (flags & TIOCM_DSR) strcat(flagStr, "DSR ");
    if (flags & TIOCM_CD) strcat(flagStr, "CD ");
    if (flags & TIOCM_RI) strcat(flagStr, "RI ");
    if (flags & TIOCM_OUT1) strcat(flagStr, "OUT1 ");
    if (flags & TIOCM_OUT2) strcat(flagStr, "OUT2 ");
    if (*flagStr != '\0') flagStr[strlen(flagStr) - 1] = '\0';
    DBG_PRINT(DbgInfo, ("#### Modem Info: %s (0x%08x)\n", flagStr, flags));
    }
#endif /* DBG */

    status = put_user(flags, pArg);

    DBG_RETURNF(DbgInfo, ("status %d", status));
    return status;
}

/*+F*************************************************************************
 * Function:
 *   GpWanSetModemInfo
 *
 * Description:
 *   Interface for the application to modify the modem control information.
 *-F*************************************************************************/
DBG_STATIC int
GpWanSetModemInfo(
    channel_t      *pChan,
    unsigned int    cmd,
    unsigned int   *pArg
    )
{
    DBG_FUNCX("GpWanSetModemInfo", DBG_WAN_SYS)
    int         status = 0;
    unsigned int    flags;

#ifdef	USB_STEP
	if( g_break )
	{
    	DBG_PRINT(DbgInfo, ("WAN CHECK e\n"));
		return -ENODEV;
	}
#endif

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pChan", ("(0x%p)", pChan));
    DBG_PARAMF(DbgInfo, "cmd", ("%d", cmd));
    DBG_PARAMF(DbgInfo, "pArg", ("(0x%p)", pArg));


    status = get_user(flags, pArg);
    if (status == 0)
    {
        switch (cmd)
        {
        case TIOCMBIS:
            DBG_PRINT(DbgInfo, ("#### TIOCMBIS: "));
            if (flags & TIOCM_RTS)
            {
                pChan->mcr |= UART_MCR_RTS;
                DBG_PRINT(DbgInfo, ("RTS "));
            }

            if (flags & TIOCM_DTR)
            {
                pChan->mcr |= UART_MCR_DTR;
                DBG_PRINT(DbgInfo, ("DTR "));
            }
            break;

        case TIOCMBIC:
            DBG_PRINT(DbgInfo, ("#### TIOCMBIC: "));
            if (flags & TIOCM_RTS)
            {
                pChan->mcr &= ~UART_MCR_RTS;
                DBG_PRINT(DbgInfo, ("~RTS "));
            }

            if (flags & TIOCM_DTR)
            {
                pChan->mcr &= ~UART_MCR_DTR;
                DBG_PRINT(DbgInfo, ("~DTR "));
            }
            break;

        case TIOCMSET:
            DBG_PRINT(DbgInfo, ("#### TIOCMSET: "));
            if (flags & TIOCM_RTS)
            {
                pChan->mcr |= UART_MCR_RTS;
                DBG_PRINT(DbgInfo, ("RTS "));
            }
            else
            {
                pChan->mcr &= ~UART_MCR_RTS;
                DBG_PRINT(DbgInfo, ("~RTS "));
            }

            if (flags & TIOCM_DTR)
            {
                pChan->mcr |= UART_MCR_DTR;
                DBG_PRINT(DbgInfo, ("DTR "));
            }
            else
            {
                pChan->mcr &= ~UART_MCR_DTR;
                DBG_PRINT(DbgInfo, ("~DTR "));
            }
            break;

        default:
            DBG_ERROR(DbgInfo, ("Unexpected modem command %d\n", cmd));
            status = -EINVAL;
            break;
        }
        DBG_PRINT(DbgInfo, ("\n"));
    }

    DBG_RETURNF(DbgInfo, ("status %d", status));
    return status;
}

/*+F*************************************************************************
 * Function:
 *   GpWanGetLsrInfo
 *
 * Description:
 *   Return the line status information to the application. The flags need
 *   to be converted from kernel space values to application values.
 *-F*************************************************************************/
DBG_STATIC int
GpWanGetLsrInfo(
    channel_t      *pChan,
    unsigned int   *pArg
    )
{
    DBG_FUNCX("GpWanGetLsrInfo", DBG_WAN_SYS)
    int         status = 0;
    unsigned int    flags;

#ifdef	USB_STEP
	if( g_break )
	{
    	DBG_PRINT(DbgInfo, ("WAN CHECK f\n"));
		return -ENODEV;
	}
#endif

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "pChan", ("(0x%p)", pChan));
    DBG_PARAMF(DbgInfo, "pArg", ("(0x%p)", pArg));


    flags = ((pChan->lsr & UART_LSR_TEMT) ? TIOCSER_TEMT : 0);

#if DBG
    {
    char    flagStr[128];

    *flagStr = '\0';
    if (flags & TIOCSER_TEMT) strcat(flagStr, "TEMT ");
    if (*flagStr != '\0') flagStr[strlen(flagStr) - 1] = '\0';
    DBG_PRINT(DbgInfo, ("#### LSR Info: %s (0x%08x)\n", flagStr, flags));
    }
#endif /* DBG */

    status = put_user(flags, pArg);

    DBG_RETURNF(DbgInfo, ("status %d", status));
    return status;
}

/*+F*************************************************************************
 * Function:
 *   GpWanIoctl
 *
 * Description:
 *   Exported IOCTL interface for the TTY device.
 *-F*************************************************************************/
DBG_STATIC int
GpWanIoctl(
    struct tty_struct  *tty,
    struct file        *pFile,
    unsigned int        cmd,
    unsigned long       arg
    )
{
    DBG_FUNCX("gsi_wan_ioctl", DBG_WAN_SYS)
    adapter_t  *pAdapter;
    channel_t  *pChan;
    int         status = 0;

#ifdef	USB_STEP
	if( g_break )
	{
    	DBG_PRINT(DbgInfo, ("WAN CHECK g\n"));
		return -ENODEV;
	}
#endif

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "tty", ("(0x%p)", tty));
    DBG_PARAMF(DbgInfo, "pFile", ("(0x%p)", pFile));

    DBG_PARAMF(DbgInfo, "cmd", ("%s (0x%04x)", DbgIoctlCmd(cmd), cmd));
    DBG_PARAMF(DbgInfo, "arg", ("0x%08lx", arg));

    if ((pChan = (channel_t *) tty->driver_data) == NULL)
    {
        DBG_WARNING(DbgInfo, ("No channel for operation!\n"));
        status = -ENODEV;
        goto gsi_wan_ioctl_exit;
    }

    pAdapter = pChan->pAdapter;
    DBG_ASSERT(pAdapter != NULL);

    if (!GpatmisIoctl(pAdapter, cmd, (struct gsi_req *) arg, &status))
    {
        switch (cmd)
        {
        case TCSBRK:    /* SVID version: non-zero arg --> no break */
            status = tty_check_change(tty);
            if (status == 0)
            {
                tty_wait_until_sent(tty, 0);
            }
            break;

        case TCSBRKP:   /* support for POSIX tcsendbreak() */
            status = tty_check_change(tty);
            if (status == 0)
            {
                tty_wait_until_sent(tty, 0);
            }
            break;

        case TIOCGSOFTCAR:
            status = put_user((tty->termios->c_cflag & CLOCAL) ? 1 : 0,
                (unsigned long *) arg);
            break;

        case TIOCSSOFTCAR:
            status = get_user(arg, (unsigned long *) arg);
            if (status == 0)
            {
                tty->termios->c_cflag =
                    ((tty->termios->c_cflag & ~CLOCAL) | (arg ? CLOCAL : 0));
            }
            break;

        case TIOCMGET:
            status = GpWanGetModemInfo(pChan, (unsigned int *) arg);
            break;

        case TIOCMBIS:
        case TIOCMBIC:
        case TIOCMSET:
            status = GpWanSetModemInfo(pChan, cmd, (unsigned int *) arg);
            break;

        case TIOCSERGETLSR:
            status = GpWanGetLsrInfo(pChan, (unsigned int *) arg);
            break;

        default:
            DBG_WARNING(DbgInfo, ("Unhandled IOCTL command: %s (0x%x)\n",
                DbgIoctlCmd(cmd), cmd));
            status = -ENOIOCTLCMD;
            break;
        }
    }

gsi_wan_ioctl_exit:
    DBG_RETURNF(DbgInfo, ("status %d", status));
    return status;
}

/*+F*************************************************************************
 * Function:
 *   GpWanWriteRoom
 *
 * Description:
 *   Return the number of additional transmit bytes the device can accept.
 *-F*************************************************************************/
DBG_STATIC int
GpWanWriteRoom(
    struct tty_struct  *tty
    )
{
    DBG_FUNCX("GpWanWriteRoom", DBG_WAN_SYS)
    channel_t  *pChan;
    int         length = 0;

#ifdef	USB_STEP
	if( g_break )
	{
    	DBG_PRINT(DbgInfo, ("WAN CHECK i\n"));
		return -ENODEV;
	}
#endif

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "tty", ("(0x%p)", tty));


    if ((pChan = (channel_t *) tty->driver_data) != NULL)
    {
        length = pChan->txAvailable * GSI_WAN_NDIS_BUFFER_SIZE;
    }
    DBG_ASSERT(pChan != NULL);

    DBG_RETURNF(DbgInfo, ("length %d", length));
    return length;
}

/*+F*************************************************************************
 * Function:
 *   GpWanCharsInBuffer
 *
 * Description:
 *   Return the number of bytes waiting to be transmitted out the adapter.
 *-F*************************************************************************/
DBG_STATIC int
GpWanCharsInBuffer(
    struct tty_struct  *tty
    )
{
    DBG_FUNCX("GpWanCharsInBuffer", DBG_WAN_SYS)
    channel_t  *pChan;
    int         length = 0;

#ifdef	USB_STEP
	if( g_break )
	{
    	DBG_PRINT(DbgInfo, ("WAN CHECK J\n"));
		return -ENODEV;
	}
#endif

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "tty", ("(0x%p)", tty));


    if ((pChan = (channel_t *) tty->driver_data) != NULL)
    {
        length = pChan->txBytes;
    }
    DBG_ASSERT(pChan != NULL);

    DBG_RETURNF(DbgInfo, ("length %d", length));
    return length;
}

/*+F*************************************************************************
 * Function:
 *   GpWanHangup
 *
 * Description:
 *   Exported Hangup routine for the TTY device.
 *-F*************************************************************************/
DBG_STATIC void 
GpWanHangup(struct tty_struct * tty)
{
    DBG_FUNCX("GpWanHangup", DBG_WAN_SYS)

#ifdef	USB_STEP
	if( g_break )
	{
    	DBG_PRINT(DbgInfo, ("WAN CHECK K\n"));
		return; 
	}
#endif

    DBG_ENTER(DbgInfo);
    DBG_PARAMF(DbgInfo, "tty", ("(0x%p)", tty));

    DBG_LEAVE(DbgInfo);
    return ;
}

#ifdef HEXDUMPWAN
void HexDumpWan(unsigned char *pBuffer, unsigned long nSize)
{
    unsigned int nAddr = (unsigned int)pBuffer;
    const int nMaxBPR = 16;				/* Max byes per row of dump */
    int nBPR;
    unsigned long nCount = 0;

	printk("\n---------HEX DUMP BEGIN-------\n");
    while (nCount < nSize)
    {
        nBPR = 0;
        while ((nCount < nSize) && (nBPR < nMaxBPR))
        {
			printk("%02x ",(unsigned char)*(pBuffer+nCount));
            nCount++;
            nBPR++;
        }
        nAddr += nMaxBPR;
        printk("\n");
    }
	printk("\n---------HEX DUMP END-------\n");
    return ;
}
#endif // HEXDUMPWANx
