////////////////////////////////////////////////////////////////////////////////
//
// GlobeSpan Inc. PROPRIETARY INFORMATION
// This software is supplied under the terms of a license agreement
// or non-disclosure agreement with GlobeSpan Inc. and may not be
// copied nor disclosed except in accordance with the terms of that
// agreement.
// Copyright  1998 GlobeSpan Inc. All Rights Reserved.
//
// PVCS Information:
//     $WorkFile$
//     $Revision: 1.1 $
//     $Modtime: 4/26/01 5:16p $
//     $Author: kattre $
//
// ABSTRACT:
//		Contains the bulk data transfer routines
//
////////////////////////////////////////////////////////////////////////////////


#include "Gp.h"
#include "common.h"
#include "debug.h"
#include "atmphy.h"
#include "usbphy.h"
#include "usbcdc.h"


////////////////////////////////////////////////////////////////////////////////
//
// NAME: UsbPhyStartBulkTransmit
//
// ABSTRACT:
//		Starts the bulk transmit pipe
//
// RETURNS:
//	   	NTSTATUS
//
// ASSUMPTIONS:
//		None.
//
////////////////////////////////////////////////////////////////////////////////
NTSTATUS
UsbPhyStartBulkTransmit( PUSB_PHY UsbPhy)
{
	PUSB_PHY_TRANSFER_RESOURCE			TransferResource;
	ULONG								i = 0;
	
	DBGPRINT(
		DBG_PRINT_MODULE_USB_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("UsbPhyStartBulkTransmit\n"));

	
	//
	// Reset the pipe prior to submitting any requests
	//
	UsbPhyResetPipe( UsbPhy,UsbPhy->BulkTransmitPipeHandle);

	//
	// Allocate resources for the Bulk transmit pipe.  Note that for the Bulk 
	// transmit pipe, the data buffer will be given to us by the sender.
	//

	atomic_set(&(UsbPhy->BulkTransmitPending), 0);
	for ( i = 0; i < UsbPhy->BulkTransmitResourceCount; i++) 
	{
		//
		// Allocate the transfer resource
		//
		ALLOCATE_MEMORY( &(TransferResource), 
			sizeof(USB_PHY_TRANSFER_RESOURCE));
		if ( NULL == TransferResource )
			return( STATUS_UNSUCCESSFUL);


		//
		// Allocate the URB
		//
		TransferResource->Urb = (PURB)usb_alloc_urb(0);
		if ( NULL == TransferResource->Urb) 
		{
			UsbPhyStopBulkTransmit( UsbPhy);
			return( STATUS_UNSUCCESSFUL);
		}
		PUSH_ENTRY_LIST( &(UsbPhy->BulkTransmitResourceList),
			(PSINGLE_LIST_NODE)TransferResource);
	}
	return( STATUS_SUCCESS);

} // UsbPhyStartBulkTransmit

////////////////////////////////////////////////////////////////////////////////
//
// NAME: UsbPhyStopBulkTransmit
//
// ABSTRACT:
//		Stops the bulk transmit pipe and frees all resources allocated to it.
//
// RETURNS:
//	   	void
//
// ASSUMPTIONS:
//		Runs at IRQL PASSIVE_LEVEL
//
////////////////////////////////////////////////////////////////////////////////
void
UsbPhyStopBulkTransmit( PUSB_PHY UsbPhy)
{
	PUSB_PHY_TRANSFER_RESOURCE		SendResource;
	KIRQL							Irql;
	
	DBGPRINT(
		DBG_PRINT_MODULE_USB_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("UsbPhyStopBulkTransmit\n"));

	ASSERT( UsbPhy);

	if( UsbPhy->PipeAborting & BULKTX_PIPE_ABORT_PROGRESS)
	{
		DBGPRINT( DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
			("BULK TRANSMIT ABORT ALREADY IN PROGRESS\n"));
		return;
	}

	UsbPhy->PipeAborting |= BULKTX_PIPE_ABORT_PROGRESS;

	KeAcquireSpinLock( &(UsbPhy->BulkTransmitResourceListLock), &Irql);
	SendResource = (PUSB_PHY_TRANSFER_RESOURCE)POP_ENTRY_LIST( 
						&(UsbPhy->BulkTransmitUsedResourceList));
	KeReleaseSpinLock( &(UsbPhy->BulkTransmitResourceListLock),Irql);

	while( SendResource )
	{
		if( SendResource->Urb )
		{
			if( SendResource->InProgress == TRUE )
			{
				SendResource->Urb->transfer_flags &= ~USB_ASYNC_UNLINK;
				usb_unlink_urb(SendResource->Urb);
			}
			usb_free_urb(SendResource->Urb);
		}
		FREE_MEMORY( SendResource, sizeof( USB_PHY_TRANSFER_RESOURCE));
		KeAcquireSpinLock( &(UsbPhy->BulkTransmitResourceListLock), &Irql);
		SendResource = (PUSB_PHY_TRANSFER_RESOURCE)POP_ENTRY_LIST( 
						&(UsbPhy->BulkTransmitUsedResourceList));
		KeReleaseSpinLock( &(UsbPhy->BulkTransmitResourceListLock),Irql);
	}

	KeAcquireSpinLock( &(UsbPhy->BulkTransmitResourceListLock), &Irql);
	SendResource = (PUSB_PHY_TRANSFER_RESOURCE)POP_ENTRY_LIST( 
						&(UsbPhy->BulkTransmitResourceList));
	KeReleaseSpinLock( &(UsbPhy->BulkTransmitResourceListLock),Irql);
	while( SendResource )
	{
		if( SendResource->Urb )
		{
			usb_free_urb(SendResource->Urb);
		}
		FREE_MEMORY( SendResource, sizeof( USB_PHY_TRANSFER_RESOURCE));
		KeAcquireSpinLock( &(UsbPhy->BulkTransmitResourceListLock), &Irql);
		SendResource = (PUSB_PHY_TRANSFER_RESOURCE)POP_ENTRY_LIST( 
						&(UsbPhy->BulkTransmitResourceList));
		KeReleaseSpinLock( &(UsbPhy->BulkTransmitResourceListLock),Irql);
	}
	// Reset Pipe aborting flag
	UsbPhy->PipeAborting &= ~BULKTX_PIPE_ABORT_PROGRESS;
} // UsbPhyStopBulkTransmit

////////////////////////////////////////////////////////////////////////////////
//
// NAME: UsbPhySend
//
// ABSTRACT:
//		Sends the given data buffer over the USB Phy using the bulk transmit
//		pipe
//
// RETURNS:
//	   	void
//
// ASSUMPTIONS:
//		USB Phy has been successfully initialized
//
////////////////////////////////////////////////////////////////////////////////
ATM_PHY_STATUS
UsbPhySend
(
    PUSB_PHY				UsbPhy,
	PXFER_DATA_BUFFER		Buffer,
	ATM_PHY_USER_HANDLE		UserContext
)
{
	PUSB_PHY_TRANSFER_RESOURCE	SendResource;
	KIRQL						Irql;

	//DBGPRINT(
	//	DBG_PRINT_MODULE_USB_ALL,
	//	DBG_PRINT_LEVEL_TRACE,
	//	("UsbPhySend\n"));

#ifdef	USB_HEXDUMP
	HexDump(Buffer->Data, Buffer->Size);
#endif

	// Get a send resource from the free list
	KeAcquireSpinLock( &(UsbPhy->BulkTransmitResourceListLock), &Irql);
	SendResource = (PUSB_PHY_TRANSFER_RESOURCE)POP_ENTRY_LIST( 
						&(UsbPhy->BulkTransmitResourceList));
	KeReleaseSpinLock( &(UsbPhy->BulkTransmitResourceListLock),Irql);

	if( NULL == SendResource )
	{
		//
		// Allocate the transfer resource
		//
		ALLOCATE_MEMORY( &(SendResource), 
			sizeof(USB_PHY_TRANSFER_RESOURCE));
		if ( NULL == SendResource )
			return( STATUS_UNSUCCESSFUL);
		//
		// Allocate the URB
		//
		SendResource->Urb = (PURB)usb_alloc_urb(0);
		if ( NULL == SendResource->Urb) 
		{
			FREE_MEMORY( SendResource, sizeof( USB_PHY_TRANSFER_RESOURCE));
			UsbPhyStopBulkTransmit( UsbPhy);
			return( STATUS_UNSUCCESSFUL);
		}
	}
	//
	// Save the user context and data buffer in our send resource so
	// we can return them when the send is completed
	//		
	SendResource->UserContext 	= UserContext;
	SendResource->Buffer		= Buffer;
	SendResource->UsbPhy		= UsbPhy;
	//
	// Build the URB for the transfer
	//

	FILL_BULK_URB( SendResource->Urb, UsbPhy->AdapterHandle->usbDev,
			UsbPhy->BulkTransmitPipeHandle,
			SendResource->Buffer->Data, SendResource->Buffer->Size, 
			UsbPhySendComplete, SendResource );
	SendResource->Urb->transfer_flags |= USB_QUEUE_BULK;
	if (usb_submit_urb(SendResource->Urb))
	{
		DBGPRINT( DBG_PRINT_MODULE_USB_TX, DBG_PRINT_LEVEL_ERR,
			("Submit Bulk Urb Failed\n"));
		KeAcquireSpinLock( &(UsbPhy->BulkTransmitResourceListLock), &Irql);
		PUSH_ENTRY_LIST( &(UsbPhy->BulkTransmitResourceList),
				(PSINGLE_LIST_NODE)SendResource);
		KeReleaseSpinLock( &(UsbPhy->BulkTransmitResourceListLock),Irql);
		UsbPhyStopBulkTransmit( UsbPhy);
		return( ATM_PHY_STATUS_FAILURE );
	}

	KeAcquireSpinLock( &(UsbPhy->BulkTransmitResourceListLock), &Irql);
	atomic_inc(&(UsbPhy->BulkTransmitPending));
	SendResource->InProgress = TRUE;
	PUSH_ENTRY_LIST( &(UsbPhy->BulkTransmitUsedResourceList),
			(PSINGLE_LIST_NODE)SendResource);
	KeReleaseSpinLock( &(UsbPhy->BulkTransmitResourceListLock),Irql);

    return( ATM_PHY_STATUS_PENDING);
} // UsbPhySend

////////////////////////////////////////////////////////////////////////////////
//
// NAME: UsbPhySendComplete
//
// ABSTRACT:
//		IRP completion routine called when a send request on the bulk out
//		pipe has completed
//
// RETURNS:
//	   	void
//
// ASSUMPTIONS:
//		None.
//
////////////////////////////////////////////////////////////////////////////////
void
UsbPhySendComplete( PURB Urb )
{
	ATM_PHY_STATUS	AtmPhyStatus;
	PUSB_PHY		UsbPhy;
	PUSB_PHY_TRANSFER_RESOURCE	SendResource = NULL;
	PUSB_PHY_TRANSFER_RESOURCE	ListResource = NULL;
	KIRQL			Irql;

	//DBGPRINT(
	//	DBG_PRINT_MODULE_USB_ALL,
	//	DBG_PRINT_LEVEL_TRACE,
	//	("UsbPhySendComplete\n"));


	SendResource = Urb->context;

	UsbPhy = SendResource->UsbPhy;

	SendResource->InProgress = FALSE;

	if(atomic_read(&(UsbPhy->BulkTransmitPending)))
		atomic_dec(&(UsbPhy->BulkTransmitPending));

	if( UsbPhy->PipeAborting & BULKTX_PIPE_ABORT_PROGRESS )
	{
		DBGPRINT( DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
			("BULK TXMIT  ABORT IN PROGRESS : Urb Status 0x%x Pending %d\n",
			Urb->status, atomic_read(&UsbPhy->BulkTransmitPending)));

		// Not putting the resource from used list to free list as it is 
		// anyways going to be freed in StopBulkTransmit

		return;
	}
	if ( IsNotOperational (UsbPhy) ) 
	{
		DBGPRINT( DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
			("Bulk Tx Bus not Operational Urb Status 0x%x Pending : %d\n",
			Urb->status, atomic_read(&UsbPhy->BulkTransmitPending)));
		ListResource = RemoveResourceFromList(
					&(UsbPhy->BulkTransmitUsedResourceList), 
					SendResource,
					&(UsbPhy->BulkTransmitResourceListLock));
		if( ListResource )
		{
			KeAcquireSpinLock( &(UsbPhy->BulkTransmitResourceListLock), &Irql);
			PUSH_ENTRY_LIST(&(UsbPhy->BulkTransmitResourceList), 
				(PSINGLE_LIST_NODE)ListResource);
			KeReleaseSpinLock( &(UsbPhy->BulkTransmitResourceListLock),Irql);
		}
		return;
	}

	//
	// Check the status of the send request
	//
	if( Urb->status ) 
	{
		DBGPRINT( DBG_PRINT_MODULE_USB_TX, DBG_PRINT_LEVEL_ERR,
			("Send completed with error 0x%X\n", Urb->status));

		// If the USBD status indicates that the device is not responding,
		// we will enter the HALTED state (this normally happens when the
		// device has just gone away)
		//
		if ( Urb->status == USB_ST_STALL ) 
		{
			DBGPRINT(DBG_PRINT_MODULE_USB_ALL,DBG_PRINT_LEVEL_ERR,
				("BulkSend: Interface Halted Due to Device Halted Error\n"));

			// Don't process anymore URBs until we are reset
			if( USB_PHY_STATE_RESETTING_PORT != GetUsbState (UsbPhy) )
				SetUsbState (UsbPhy, USB_PHY_STATE_RESETTING_PORT);
		}
		AtmPhyStatus = ATM_PHY_STATUS_FAILURE;
		ListResource = RemoveResourceFromList(
					&(UsbPhy->BulkTransmitUsedResourceList), 
					SendResource,
					&(UsbPhy->BulkTransmitResourceListLock));
		if( ListResource )
		{
			KeAcquireSpinLock( &(UsbPhy->BulkTransmitResourceListLock), &Irql);
			PUSH_ENTRY_LIST(&(UsbPhy->BulkTransmitResourceList), 
				(PSINGLE_LIST_NODE)ListResource);
			KeReleaseSpinLock( &(UsbPhy->BulkTransmitResourceListLock),Irql);
		}
		return;
	}
	else 
	{
		AtmPhyStatus = ATM_PHY_STATUS_SUCCESS;
	}

	// Call the user send complete routine indicating the status of the send request
	(UsbPhy->UserSendComplete)( UsbPhy->UserHandle, SendResource->Buffer,
								SendResource->UserContext, AtmPhyStatus);

	ListResource = RemoveResourceFromList(
				&(UsbPhy->BulkTransmitUsedResourceList), 
				SendResource,
				&(UsbPhy->BulkTransmitResourceListLock));
	if( ListResource )
	{
		KeAcquireSpinLock( &(UsbPhy->BulkTransmitResourceListLock), &Irql);
		PUSH_ENTRY_LIST(&(UsbPhy->BulkTransmitResourceList), 
			(PSINGLE_LIST_NODE)ListResource);
		KeReleaseSpinLock( &(UsbPhy->BulkTransmitResourceListLock),Irql);
	}
    return;

} // UsbPhySendComplete


////////////////////////////////////////////////////////////////////////////////
//
// NAME: UsbPhyStartBulkReceive
//
// ABSTRACT:
//		Starts the bulk receive pipe by initializing and submitting the 
//      transfer request URBs
//
// RETURNS:
//	   	NTSTATUS
//
// ASSUMPTIONS:
//		None.
//
////////////////////////////////////////////////////////////////////////////////
NTSTATUS
UsbPhyStartBulkReceive( PUSB_PHY UsbPhy)
{
	PUSB_PHY_TRANSFER_RESOURCE		TransferResource;
	PURB							Urb;
	PXFER_DATA_BUFFER				Buffer;
	PUCHAR							Data;
	KIRQL							Irql;
	int								i;

	DBGPRINT(
		DBG_PRINT_MODULE_USB_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("UsbPhyStartBulkReceive\n"));

	UsbPhy->BulkReceiveResourceList.Next = NULL;
	UsbPhy->BulkReceiveUsedResourceList.Next = NULL;
	//
	// Reset the pipe prior to submitting any requests
	//
	UsbPhyResetPipe( UsbPhy,UsbPhy->BulkReceivePipeHandle);

	//
	// Allocate resources for the pipe
	//
	//for ( i = 0; i < UsbPhy->BulkReceiveResourceCount; i++) 
	{
		//
		// Allocate the transfer resource
		//
		ALLOCATE_MEMORY( &TransferResource,sizeof(USB_PHY_TRANSFER_RESOURCE));
		if ( NULL == TransferResource) 
		{
			UsbPhyStopBulkReceive( UsbPhy);
			return( STATUS_UNSUCCESSFUL);
		}

		//
		// Allocate the URB
		//
		Urb = usb_alloc_urb(0);
		if ( NULL == Urb) 
		{
			FREE_MEMORY( TransferResource,sizeof(USB_PHY_TRANSFER_RESOURCE));
			UsbPhyStopBulkReceive( UsbPhy);
			return( STATUS_UNSUCCESSFUL);
		}
		TransferResource->Urb = Urb;
		TransferResource->UsbPhy	= UsbPhy;
		//
		// Allocate the data buffer descriptor
		//
		ALLOCATE_MEMORY( &Buffer,sizeof( XFER_DATA_BUFFER));
		if( NULL == Buffer) 
		{
			usb_free_urb( Urb );
			FREE_MEMORY( TransferResource,sizeof(USB_PHY_TRANSFER_RESOURCE));
			UsbPhyStopBulkReceive( UsbPhy);
			return( STATUS_UNSUCCESSFUL);
		} 
		TransferResource->Buffer = Buffer;
		//
		// Allocate data buffer
		//
		ALLOCATE_MEMORY( &Data, UsbPhy->BulkReceiveBufferSize);
		if( NULL == Data) 
		{
			FREE_MEMORY( Buffer,sizeof( XFER_DATA_BUFFER));
			usb_free_urb( Urb );
			FREE_MEMORY( TransferResource,sizeof(USB_PHY_TRANSFER_RESOURCE));
			UsbPhyStopBulkReceive( UsbPhy);
			return( STATUS_UNSUCCESSFUL);
		} 
		TransferResource->Buffer->Data = Data;
		
		//
		// Initialize the transfer resource and submit it to USBD
		//
		UsbPhyInitializeBulkReceiveTransferResource( UsbPhy,TransferResource);

		if( usb_submit_urb( TransferResource->Urb ) )
		{
			DBGPRINT(
				DBG_PRINT_MODULE_USB_ALL,
				DBG_PRINT_LEVEL_ERR,
				("Unable to submit URB\n"));
			FREE_MEMORY( Data, UsbPhy->BulkReceiveBufferSize);
			FREE_MEMORY( Buffer,sizeof( XFER_DATA_BUFFER));
			usb_free_urb( Urb );
			FREE_MEMORY( TransferResource,sizeof(USB_PHY_TRANSFER_RESOURCE));
			UsbPhyStopBulkReceive( UsbPhy);
			//
			// Since our last request failed, we'll stop submitting requests
			// and return errored status
			//
			return( STATUS_UNSUCCESSFUL );
		}

		KeAcquireSpinLock( &(UsbPhy->BulkReceiveResourceListLock), &Irql);
		/* Add to the list */
		PUSH_ENTRY_LIST( &(UsbPhy->BulkReceiveUsedResourceList),
			(PSINGLE_LIST_NODE)TransferResource);
		UsbPhyIncrementIoCount( UsbPhy);
		KeReleaseSpinLock( &(UsbPhy->BulkReceiveResourceListLock),Irql);

	}

	return( STATUS_SUCCESS);

} // UsbPhyStartBulkReceive

////////////////////////////////////////////////////////////////////////////////
//
// NAME: UsbPhyStopBulkReceive
//
// ABSTRACT:
//		Stops the bulk receive pipe and frees all resources allocated to it.
//
// RETURNS:
//	   	void
//
// ASSUMPTIONS:
//		Runs at IRQL PASSIVE_LEVEL
//
////////////////////////////////////////////////////////////////////////////////
void
UsbPhyStopBulkReceive( PUSB_PHY UsbPhy)
{
	PUSB_PHY_TRANSFER_RESOURCE		RecvResource;
	KIRQL								Irql;
	
	DBGPRINT(
		DBG_PRINT_MODULE_USB_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("UsbPhyStopBulkReceive\n"));

	ASSERT( UsbPhy);

	//Set Pipe Aborting Flag
	if( UsbPhy->PipeAborting & BULKRX_PIPE_ABORT_PROGRESS)
	{
	   	DBGPRINT( DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
		   ("BULK RX ABORT ALREADY IN PROGRESS\n"));
		return;
	}

	UsbPhy->PipeAborting |= BULKRX_PIPE_ABORT_PROGRESS;
	
	KeAcquireSpinLock( &(UsbPhy->BulkReceiveResourceListLock), &Irql);
	RecvResource = (PUSB_PHY_TRANSFER_RESOURCE)POP_ENTRY_LIST( 
						&(UsbPhy->BulkReceiveUsedResourceList));
	KeReleaseSpinLock( &(UsbPhy->BulkReceiveResourceListLock),Irql);

	while( RecvResource )
	{
		if( RecvResource->Urb )
		{
			RecvResource->Urb->transfer_flags &= ~USB_ASYNC_UNLINK;
			usb_unlink_urb(RecvResource->Urb);
			usb_free_urb(RecvResource->Urb);
		}
		// Free the data buffer and buffer descriptor
		if ( RecvResource->Buffer) 
		{
			if ( RecvResource->Buffer->Data) 
			{
				FREE_MEMORY( 
					RecvResource->Buffer->Data,
					UsbPhy->BulkReceiveBufferSize);
			}
			FREE_MEMORY( RecvResource->Buffer,sizeof( XFER_DATA_BUFFER));
		}
		
		FREE_MEMORY( RecvResource, sizeof( USB_PHY_TRANSFER_RESOURCE));
		KeAcquireSpinLock( &(UsbPhy->BulkReceiveResourceListLock), &Irql);
		RecvResource = (PUSB_PHY_TRANSFER_RESOURCE)POP_ENTRY_LIST( 
						&(UsbPhy->BulkReceiveUsedResourceList));
		KeReleaseSpinLock( &(UsbPhy->BulkReceiveResourceListLock),Irql);
	}

	KeAcquireSpinLock( &(UsbPhy->BulkReceiveResourceListLock), &Irql);
	RecvResource = (PUSB_PHY_TRANSFER_RESOURCE)POP_ENTRY_LIST( 
						&(UsbPhy->BulkReceiveResourceList));
	KeReleaseSpinLock( &(UsbPhy->BulkReceiveResourceListLock),Irql);
	while( RecvResource )
	{
		if( RecvResource->Urb )
		{
			usb_free_urb(RecvResource->Urb);
		}
		// Free the data buffer and buffer descriptor
		if ( RecvResource->Buffer) 
		{
			if ( RecvResource->Buffer->Data) 
			{
				FREE_MEMORY( 
					RecvResource->Buffer->Data,
					UsbPhy->BulkReceiveBufferSize);
			}
			FREE_MEMORY( RecvResource->Buffer,sizeof( XFER_DATA_BUFFER));
		}
		FREE_MEMORY( RecvResource, sizeof( USB_PHY_TRANSFER_RESOURCE));
		KeAcquireSpinLock( &(UsbPhy->BulkReceiveResourceListLock), &Irql);
		RecvResource = (PUSB_PHY_TRANSFER_RESOURCE)POP_ENTRY_LIST( 
						&(UsbPhy->BulkReceiveResourceList));
		KeReleaseSpinLock( &(UsbPhy->BulkReceiveResourceListLock),Irql);
	}
	UsbPhy->PipeAborting &= ~BULKRX_PIPE_ABORT_PROGRESS;
} // UsbPhyStopBulkReceive

#ifdef	_NTDDK_
////////////////////////////////////////////////////////////////////////////////
//
// NAME: UsbPhyBulkReceiveSubmitUrbs
//
// ABSTRACT:
//		Resubmits all URBs in the free list
//
// RETURNS:
//	   	NTSTATUS
//
// ASSUMPTIONS:
//		None.
//
////////////////////////////////////////////////////////////////////////////////
NTSTATUS
UsbPhyBulkReceiveSubmitUrbs(PUSB_PHY UsbPhy)
{
	PUSB_PHY_TRANSFER_RESOURCE	TransferResource;
	PURB							Urb;

	DBGPRINT(
		DBG_PRINT_MODULE_USB_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("UsbPhyBulkReceiveSubmitUrbs\n"));

	ASSERT( UsbPhy);

	TransferResource = (PUSB_PHY_TRANSFER_RESOURCE)POP_ENTRY_LIST( 
						&(UsbPhy->BulkReceiveResourceList));
	while ( TransferResource) 
	{
		Urb = TransferResource->Urb;

		//
		// Recycle the transfer resource
		//
		UsbPhyInitializeBulkReceiveTransferResource( UsbPhy,TransferResource);

		//
		// submit the request
		//
	   	DBGPRINT(
		   DBG_PRINT_MODULE_USB_ALL,
		   DBG_PRINT_LEVEL_TRACE,
		   ("Calling IoCallDriver\n"));

		if( usb_submit_urb( TransferResource->Urb ) )
		{
			DBGPRINT(
			   DBG_PRINT_MODULE_USB_ALL,
			   DBG_PRINT_LEVEL_ERR,
			   ("Unable to resubmit IRP\n"));
			PUSH_ENTRY_LIST( 
				&(UsbPhy->BulkReceiveResourceList),
				(PSINGLE_LIST_NODE)TransferResource);

		   return( STATUS_UNSUCCESSFUL );
		}

		UsbPhyIncrementIoCount( UsbPhy);

		//
		// Get the the next resource to free
		//
		TransferResource = (PUSB_PHY_TRANSFER_RESOURCE)POP_ENTRY_LIST( 
						&(UsbPhy->BulkReceiveResourceList));
	}

   return (STATUS_SUCCESS);
}   
#endif	//_NTDDK_

////////////////////////////////////////////////////////////////////////////////
//
// NAME: UsbPhyInitializeBulkReceiveTransferResource
//
// ABSTRACT:
//		Initializes the given transfer resource for use with bulk receive
//
// RETURNS:
//	   	void
//
// ASSUMPTIONS:
//		None.
//
////////////////////////////////////////////////////////////////////////////////
void
UsbPhyInitializeBulkReceiveTransferResource
(
	PUSB_PHY					UsbPhy,
	PUSB_PHY_TRANSFER_RESOURCE	TransferResource
)
{
	PURB				Urb;

	DBGPRINT(
		DBG_PRINT_MODULE_USB_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("UsbPhyInitializeBulkReceiveTransferResource\n"));

	ASSERT( UsbPhy);
	ASSERT( TransferResource);

	Urb = TransferResource->Urb;

	ZERO_MEMORY( Urb, sizeof( URB));
	ZERO_MEMORY( TransferResource->Buffer->Data,UsbPhy->BulkReceiveBufferSize);    

	//
	// Initialize the URB
	//
	FILL_BULK_URB( Urb, UsbPhy->AdapterHandle->usbDev, 
		UsbPhy->BulkReceivePipeHandle, TransferResource->Buffer->Data,
		UsbPhy->BulkReceiveBufferSize, UsbPhyBulkReceiveComplete, TransferResource );
	Urb->transfer_flags |= USB_QUEUE_BULK;
} // UsbPhyInitializeBulkReceiveTransferResource

////////////////////////////////////////////////////////////////////////////////
//
// NAME: UsbPhyBulkReceiveComplete
//
// ABSTRACT:
//		Called by USBD when a previously submitted Bulk RX URB has been
//		completed.  If the request was successful, the received data is
//		indicated to the phy user.
//
// RETURNS:
//	   	NTSTATUS
//
// ASSUMPTIONS:
//		None.
//
////////////////////////////////////////////////////////////////////////////////
void
UsbPhyBulkReceiveComplete( PURB Urb)
{
	PUSB_PHY 						UsbPhy;
	PUSB_PHY_TRANSFER_RESOURCE		TransferResource = NULL;
	PUSB_PHY_TRANSFER_RESOURCE		ListResource = NULL;
	PXFER_DATA_BUFFER				Buffer;
	KIRQL							Irql;

	DBGPRINT(
		DBG_PRINT_MODULE_USB_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("UsbPhyBulkReceiveComplete\n"));


	TransferResource	= (PUSB_PHY_TRANSFER_RESOURCE)(Urb->context);
	UsbPhy				= TransferResource->UsbPhy;
	Buffer 				= TransferResource->Buffer;

	UsbPhyDecrementIoCount(UsbPhy);

	if( UsbPhy->PipeAborting & BULKRX_PIPE_ABORT_PROGRESS )
	{
		DBGPRINT( DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
			("BULKRX ABORT IN PROGRESS Urb Status 0x%x Pending Count : %d\n",
			Urb->status, atomic_read(&UsbPhy->PendingIoCount)));

		return;
	}
	//
	// If we have entered the HALTED state, simply put the transfer resource
	// back on the free list
	//
	if ( IsNotOperational (UsbPhy) ) 
	{
		DBGPRINT( DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
			("Bulk Rx Usb bus not operational, Urb status 0x%x Pending : %d\n",
			Urb->status, atomic_read(&UsbPhy->PendingIoCount)));

		ListResource = RemoveResourceFromList(
					&(UsbPhy->BulkReceiveUsedResourceList), 
					TransferResource,
					&(UsbPhy->BulkReceiveResourceListLock));
		if( ListResource )
		{
			KeAcquireSpinLock( &(UsbPhy->BulkReceiveResourceListLock), &Irql);
			PUSH_ENTRY_LIST( &(UsbPhy->BulkReceiveResourceList),
				(PSINGLE_LIST_NODE)TransferResource);
			KeReleaseSpinLock( &(UsbPhy->BulkReceiveResourceListLock),Irql);
		}
		return;
	}

	//
	// Check the  URB status values
	//
	if ( Urb->status )
	{
		DBGPRINT(
			DBG_PRINT_MODULE_USB_ALL,
			DBG_PRINT_LEVEL_ERR,
			("Urb completed with error (0x%X)\n",
			Urb->status));
		//
		// If the USBD status indicates that the device is not responding,
		// we will enter the HALTED state (this normally happens when the
		// device has just gone away)
		//
		if ( Urb->status == USB_ST_STALL ) 
		{
			DBGPRINT(
				DBG_PRINT_MODULE_USB_ALL,
				DBG_PRINT_LEVEL_ERR,
				("Interface Halted Due to Device Halted Error\n"));

			// Don't process anymore URBs until we are reset
			if( USB_PHY_STATE_RESETTING_PORT != GetUsbState (UsbPhy) )
				SetUsbState (UsbPhy, USB_PHY_STATE_RESETTING_PORT);
		}

		ListResource = RemoveResourceFromList(
					&(UsbPhy->BulkReceiveUsedResourceList), 
					TransferResource,
					&(UsbPhy->BulkReceiveResourceListLock));
		if( ListResource )
		{
			KeAcquireSpinLock( &(UsbPhy->BulkReceiveResourceListLock), &Irql);
			PUSH_ENTRY_LIST( &(UsbPhy->BulkReceiveResourceList),
				(PSINGLE_LIST_NODE)TransferResource);
			KeReleaseSpinLock( &(UsbPhy->BulkReceiveResourceListLock),Irql);
		}
		return;
	} 
	else 
	{
		//
		// Indicate the received data to the phy user
		//
		Buffer->Size = Urb->actual_length;
		ASSERT( UsbPhy->UserReceive);
		
#ifdef GSI_USB
		/* The below statement is calling AtmUniAal5Rea... */
		/* Initialize in the UsbPhyBind() */
#ifdef QUASAR   
                if(Buffer->Size) 
#endif//QUASAR
 		(UsbPhy->UserReceive)( UsbPhy->UserHandle,Buffer);
#endif /* GSI_USB */
#ifndef GSI_USB
		(UsbPhy->UserReceive)( UsbPhy->UserHandle,Buffer);
#endif /* !GSI_USB */
	}

	//
	// Recycle the transfer resource
	//
	//UsbPhyInitializeBulkReceiveTransferResource( UsbPhy,TransferResource);
	//
	// submit the request
	//

	TransferResource->Urb->actual_length = 0;
	TransferResource->Urb->dev = UsbPhy->AdapterHandle->usbDev;
	
	DBGPRINT(
		DBG_PRINT_MODULE_USB_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("Calling IoCallDriver\n"));

	if( usb_submit_urb( TransferResource->Urb ) )
	{
		DBGPRINT(
			DBG_PRINT_MODULE_USB_ALL,
			DBG_PRINT_LEVEL_ERR,
			("Unable to resubmit IRP\n"));

		ListResource = RemoveResourceFromList(
					&(UsbPhy->BulkReceiveUsedResourceList), 
					TransferResource,
					&(UsbPhy->BulkReceiveResourceListLock));
		if( ListResource )
		{
			KeAcquireSpinLock( &(UsbPhy->BulkReceiveResourceListLock), &Irql);
			PUSH_ENTRY_LIST( 
				&(UsbPhy->BulkReceiveResourceList),
				(PSINGLE_LIST_NODE)TransferResource);
			KeReleaseSpinLock( &(UsbPhy->BulkReceiveResourceListLock),Irql);
		}
	}
	else
	{
		UsbPhyIncrementIoCount( UsbPhy);
	}

    return;
} // UsbPhyBulkReceiveComplete
