////////////////////////////////////////////////////////////////////////////////
//
// GlobeSpan Inc. PROPRIETARY INFORMATION
// This software is supplied under the terms of a license agreement
// or non-disclosure agreement with GlobeSpan Inc. and may not be
// copied nor disclosed except in accordance with the terms of that
// agreement.
// Copyright  1998 GlobeSpan Inc. All Rights Reserved.
//
// PVCS Information:
//     $WorkFile$
//     $Revision: 1.1 $
//     $Modtime: 7/21/01 10:35p $
//     $Author: kattre $
//
// ABSTRACT:
//		Contains the interrupt data transfer routines
//
////////////////////////////////////////////////////////////////////////////////


#include "Gp.h"
#include "common.h"
#include "debug.h"
#include "atmphy.h"
#include "usbphy.h"
#include "usbcdc.h"

#ifdef GSI_USB
extern GTI_XCVR_INFO GTISysBlock[1];
#endif /* GSI_USB */

void SubmitIntUrb(
	PVOID				SystemSpecific1,
	PUSB_PHY_TRANSFER_RESOURCE		TransferResource,	   
	PVOID				SystemSpecific2,
	PVOID				SystemSpecific3
);
////////////////////////////////////////////////////////////////////////////////
//
// NAME: UsbPhyStartInterruptReceive
//
// ABSTRACT:
//		Starts the interrupt pipe by initializing and submitting the 
//      transfer request URBs
//
// RETURNS:
//	   	NTSTATUS
//
// ASSUMPTIONS:
//		None.
//
////////////////////////////////////////////////////////////////////////////////
NTSTATUS
UsbPhyStartInterruptReceive( PUSB_PHY UsbPhy)
{
	PUSB_PHY_TRANSFER_RESOURCE	TransferResource;
	PURB						Urb;
	PXFER_DATA_BUFFER			Buffer;
	PUCHAR						Data;
    	ULONG 	 					i;
	int							nRet = USB_ST_NOERROR;

	DBGPRINT(
		DBG_PRINT_MODULE_USB_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("UsbPhyStartInterruptReceive\n"));

	//
	// Reset the pipe prior to submitting any requests
	//
	UsbPhyResetPipe( UsbPhy,UsbPhy->InterruptPipeHandle);

	atomic_set(&(UsbPhy->InterruptPendingCount), 0);

	// Initialize the length of the EOC message received
	UsbPhy->eocDatalen = 0;
	//
	// Allocate resources for the pipe
	//
	ALLOCATE_MEMORY( &(UsbPhy->InterruptResource), 
                    sizeof(USB_PHY_TRANSFER_RESOURCE) * 
					UsbPhy->InterruptResourceCount);

	DBGPRINT( DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_ERR,
		("Allocating Int Memory %d %d\n",sizeof(USB_PHY_TRANSFER_RESOURCE),
		sizeof(USB_PHY_TRANSFER_RESOURCE) * UsbPhy->InterruptResourceCount));

	ZERO_MEMORY(UsbPhy->InterruptResource, 
			sizeof(USB_PHY_TRANSFER_RESOURCE) * UsbPhy->InterruptResourceCount);

	for ( i = 0; i < UsbPhy->InterruptResourceCount; i++) 
	{
		TransferResource = (PUSB_PHY_TRANSFER_RESOURCE)
							(((PUCHAR)(UsbPhy->InterruptResource)) + 
							(i * sizeof(USB_PHY_TRANSFER_RESOURCE))); 
		//
		// Allocate the URB
		//
		Urb = usb_alloc_urb(0);
		if ( NULL == Urb) 
		{
			UsbPhyStopInterruptReceive( UsbPhy);
			return( STATUS_UNSUCCESSFUL);
		}
		TransferResource->Urb 		= Urb;
		TransferResource->UsbPhy	= UsbPhy;
		//
		// Allocate the data buffer descriptor
		//
		ALLOCATE_MEMORY( &Buffer,sizeof( XFER_DATA_BUFFER));
		if( NULL == Buffer) 
		{
			UsbPhyStopInterruptReceive( UsbPhy);
			return( STATUS_UNSUCCESSFUL);
		} 
		TransferResource->Buffer = Buffer;
		//
		// Allocate data buffer
		//
		ALLOCATE_MEMORY( &Data, UsbPhy->InterruptBufferSize);
		if( NULL == Data) 
		{
			UsbPhyStopInterruptReceive( UsbPhy);
			return( STATUS_UNSUCCESSFUL);
		} 
		TransferResource->Buffer->Data = Data;
		
		TransferResource->InProgress = FALSE;
	}

	TransferResource = UsbPhy->InterruptResource;
	//
	// Initialize the transfer resource and submit it to USBD
	//
	UsbPhyInitializeInterruptTransferResource( UsbPhy,TransferResource);

	if((nRet = usb_submit_urb( TransferResource->Urb )) != USB_ST_NOERROR )
	{
		DBGPRINT(
			DBG_PRINT_MODULE_USB_ALL,
			DBG_PRINT_LEVEL_ERR,
			("Unable to submit Interrupt URB %d\n",nRet));

		UsbPhyStopInterruptReceive( UsbPhy);
		//
		// Since our last request failed, we'll stop submitting requests
		// and return errored status
		//
		return( STATUS_UNSUCCESSFUL);
	}
	else
	{
		atomic_inc(&(UsbPhy->InterruptPendingCount));
		TransferResource->InProgress = TRUE;
	}

	return( STATUS_SUCCESS);

} // UsbPhyStartInterruptReceive

////////////////////////////////////////////////////////////////////////////////
//
// NAME: UsbPhyStopInterruptReceive
//
// ABSTRACT:
//		Stops the interrupt pipe and frees all resources allocated to it.
//
// RETURNS:
//	   	void
//
// ASSUMPTIONS:
//		Runs at IRQL PASSIVE_LEVEL
//
////////////////////////////////////////////////////////////////////////////////
void
UsbPhyStopInterruptReceive( PUSB_PHY UsbPhy)
{
	PUSB_PHY_TRANSFER_RESOURCE	TransferResource;
    	ULONG 	 					i;

	DBGPRINT(
		DBG_PRINT_MODULE_USB_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("UsbPhyStopInterruptReceive\n"));

	ASSERT( UsbPhy);

	if(UsbPhy->PipeAborting & INT_PIPE_ABORT_PROGRESS)
	{
		DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE, 
			("INT PIPE ABORT ALREADY IN PROGRESS\n"));
		return;
	}

	UsbPhy->PipeAborting |= INT_PIPE_ABORT_PROGRESS;
	//if( atomic_read(&UsbPhy->InterruptPendingCount) == 0 )
	if(UsbPhy->InterruptResource)
	{
		//
		// Free Resources for the pipe
		//
		for ( i = 0; i < UsbPhy->InterruptResourceCount; i++) 
		{
			TransferResource = (PUSB_PHY_TRANSFER_RESOURCE)
								(((PUCHAR)(UsbPhy->InterruptResource)) + 
								(i * sizeof(USB_PHY_TRANSFER_RESOURCE)));
					
			//TransferResource = UsbPhy->InterruptResource + 
			//					(i * sizeof(USB_PHY_TRANSFER_RESOURCE));
			if( TransferResource == NULL )
				goto StopIntExit;
			
			// Unlink the URB as first step
			if( TransferResource->InProgress == TRUE && TransferResource->Urb )
			{
				DBGPRINT( DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
					("Unlinking Interrupt Urb\n"));
				usb_unlink_urb(TransferResource->Urb );
				DBGPRINT( DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
					("Unlinked Interrupt Urb\n"));
				TransferResource->InProgress = FALSE;
				atomic_dec(&(UsbPhy->InterruptPendingCount));
			}
			//
			// Free the data buffer and buffer descriptor
			//
			if ( TransferResource->Buffer) 
			{
				if ( TransferResource->Buffer->Data) 
				{
					FREE_MEMORY( 
						TransferResource->Buffer->Data,
						UsbPhy->InterruptBufferSize);
				}
				FREE_MEMORY( TransferResource->Buffer,sizeof( XFER_DATA_BUFFER));
			}
			//
			// Free the URB
			//	
			if ( TransferResource->Urb) 
			{
				usb_free_urb( TransferResource->Urb );
			}
			//
			// Get the the next resource to free
			//
		}
	
		FREE_MEMORY( UsbPhy->InterruptResource,
			sizeof(USB_PHY_TRANSFER_RESOURCE));
	}
	UsbPhy->InterruptResource = NULL;

StopIntExit:

	UsbPhy->PipeAborting &= ~INT_PIPE_ABORT_PROGRESS;

} // UsbPhyStopInterruptReceive

////////////////////////////////////////////////////////////////////////////////
//
// NAME: UsbPhyInitializeInterruptTransferResource
//
// ABSTRACT:
//		Initializes the given transfer resource for use with bulk receive
//
// RETURNS:
//	   	void
//
// ASSUMPTIONS:
//		None.
//
////////////////////////////////////////////////////////////////////////////////
void
UsbPhyInitializeInterruptTransferResource
(
	PUSB_PHY					UsbPhy,
	PUSB_PHY_TRANSFER_RESOURCE	TransferResource
)
{
	PURB				Urb;

	DBGPRINT( DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
		("UsbPhyInitializeInterruptTransferResource Pipe handle = %x \n",
		UsbPhy->InterruptPipeHandle));

	ASSERT( UsbPhy);
	ASSERT( TransferResource);
	
	if( TransferResource == NULL )
	{
		DBGPRINT( DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
			("NULL HANDLE\n"));
		return;
	}

	Urb = TransferResource->Urb;
	ZERO_MEMORY( Urb, sizeof( URB));
	ZERO_MEMORY( TransferResource->Buffer->Data,UsbPhy->InterruptBufferSize);    
	TransferResource->UsbPhy = UsbPhy;
	//
	// Initialize the URB
	//
	FILL_INT_URB( Urb, UsbPhy->AdapterHandle->usbDev, 
		UsbPhy->InterruptPipeHandle, TransferResource->Buffer->Data,
		UsbPhy->InterruptBufferSize, UsbPhyInterruptComplete, 
		TransferResource, UsbPhy->InterruptInterval);
} // UsbPhyInitializeInterruptTransferResource

#ifdef	_NTDDK_
////////////////////////////////////////////////////////////////////////////////
//
// NAME: UsbPhyInterruptSubmitUrbs
//
// ABSTRACT:
//		Resubmits all URBs in the free list
//
// RETURNS:
//	   	NTSTATUS
//
// ASSUMPTIONS:
//		None.
//
////////////////////////////////////////////////////////////////////////////////
NTSTATUS
UsbPhyInterruptSubmitUrbs(PUSB_PHY UsbPhy)
{
	PUSB_PHY_TRANSFER_RESOURCE		TransferResource;
	PURB							Urb;

	DBGPRINT(
		DBG_PRINT_MODULE_USB_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("UsbPhyInterruptSubmitUrbs\n"));

	ASSERT( UsbPhy);

	TransferResource = (PUSB_PHY_TRANSFER_RESOURCE)
						(UsbPhy->InterruptResourceList.Next); 

	if( TransferResource->InProgress == TRUE )
	{
		DBGPRINT(DBG_PRINT_MODULE_USB_ALL,DBG_PRINT_LEVEL_TRACE,
			("Int URB Already in Progress\n"));
	}
	
	while ( TransferResource) 
	{
	   Urb = TransferResource->Urb;
	   //
	   // Recycle the transfer resource
	   //
	   UsbPhyInitializeInterruptTransferResource( UsbPhy,TransferResource);
       //
       // submit the request
       //
	   DBGPRINT(
		   DBG_PRINT_MODULE_USB_ALL,
		   DBG_PRINT_LEVEL_TRACE,
		   ("Calling IoCallDriver to submit Int URBs\n"));

		if( usb_submit_urb( TransferResource->Urb ) )
		{
		   DBGPRINT(
			   DBG_PRINT_MODULE_USB_ALL,
			   DBG_PRINT_LEVEL_ERR,
			   ("Unable to resubmit URB\n"));
		   PUSH_ENTRY_LIST( 
			   &(UsbPhy->InterruptResourceList),
			   (PSINGLE_LIST_NODE)TransferResource);

		   return (STATUS_UNSUCCESSFUL);
		}

		atomic_inc(&(UsbPhy->InterruptPendingCount));

		//
		// Get the the next resource to free
		//
		TransferResource = (PUSB_PHY_TRANSFER_RESOURCE)POP_ENTRY_LIST( 
								&(UsbPhy->InterruptResourceList));
	}

   return (STATUS_SUCCESS);
}   
#endif	//_NTDDK_

////////////////////////////////////////////////////////////////////////////////
//
// NAME: UsbPhyInterruptComplete
//
// ABSTRACT:
//		Called by USBD when a previously submitted Bulk RX URB has been
//		completed.  If the request was successful, the received data is
//		indicated to the phy user.
//
// RETURNS:
//	   	NTSTATUS
//
// ASSUMPTIONS:
//		None.
//
////////////////////////////////////////////////////////////////////////////////
void
UsbPhyInterruptComplete(PURB Urb)
{
    PUSB_PHY						UsbPhy;
	PUSB_PHY_TRANSFER_RESOURCE		TransferResource;
	PXFER_DATA_BUFFER				Buffer;

	//DBGPRINT(
	//	DBG_PRINT_MODULE_USB_ALL,
	//	DBG_PRINT_LEVEL_TRACE,
	//	("UsbPhyInterruptComplete Urb = %p, Pipe = %x, Flags = %x\n", 
	//	Urb, Urb->pipe, Urb->transfer_flags));


	TransferResource	= (PUSB_PHY_TRANSFER_RESOURCE)(Urb->context);
	UsbPhy				= TransferResource->UsbPhy;
	Buffer				= TransferResource->Buffer;
	Buffer->Data		= Urb->transfer_buffer;

	if(UsbPhy->PipeAborting & INT_PIPE_ABORT_PROGRESS)
	{
		DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE, 
			("INT PIPE ABORT ALREADY IN PROGRESS With Status %d\n",Urb->status));
		return;
	}

	if( (UsbPhy->InterruptInterval == 0) && 
		( atomic_read(&(UsbPhy->InterruptPendingCount))))
	{
		atomic_dec(&(UsbPhy->InterruptPendingCount));
	}

	//
	// If we have entered the HALTED state, simply put the transfer resource
	// back on the free list
	//
    if ( IsNotOperational (UsbPhy) )
	{
		DBGPRINT(
			DBG_PRINT_MODULE_USB_ALL,
			DBG_PRINT_LEVEL_TRACE,
			("Pending IO Count : %d\n",atomic_read(&UsbPhy->InterruptPendingCount)));

#ifdef	_NTDDK_
		KeAcquireSpinLock( &(UsbPhy->InterruptResourceListLock), &Irql);
		PUSH_ENTRY_LIST( 
			&(UsbPhy->InterruptResourceList),
			(PSINGLE_LIST_NODE)TransferResource);
		KeReleaseSpinLock( &(UsbPhy->InterruptResourceListLock),Irql);
#endif	//_NTDDK_
		return;
	}

   //
   // Check the IRP and URB status values
   //
   if (Urb->status)
	{
		DBGPRINT(
			DBG_PRINT_MODULE_USB_ALL,
			DBG_PRINT_LEVEL_ERR,
			("Urb completed with error (0x%X)\n",
			Urb->status));
		//
		// If the USBD status indicates that the device is not responding,
		// we will enter the HALTED state (this normally happens when the
		// device has just gone away)
		//
		if ( Urb->status == USB_ST_STALL ) 
		{
			DBGPRINT(
			   DBG_PRINT_MODULE_USB_ALL,
			   DBG_PRINT_LEVEL_ERR,
			   ("interrupt: Babble Error\n"));

			// Don't process anymore URBs until we are reset
			if( USB_PHY_STATE_RESETTING_PORT != GetUsbState (UsbPhy) )
				SetUsbState (UsbPhy, USB_PHY_STATE_RESETTING_PORT);
		}
		
#ifdef	_NTDDK_
		KeAcquireSpinLock( &(UsbPhy->InterruptResourceListLock), &Irql);
		PUSH_ENTRY_LIST( 
			&(UsbPhy->InterruptResourceList),
			(PSINGLE_LIST_NODE)TransferResource);
		KeReleaseSpinLock( &(UsbPhy->InterruptResourceListLock),Irql);
#endif	//_NTDDK_
		return;
   } 
   else 
   {
		//
		// Process the notification
		//
		Buffer->Size = Urb->actual_length;
		
		UsbPhyProcessNotification( UsbPhy, Buffer);
	}

	if( UsbPhy->InterruptInterval == 0 )
	{
		//
		// Recycle the transfer resource
		//

		// Unlink URB First of all
		usb_unlink_urb(TransferResource->Urb);
		TransferResource->InProgress = FALSE;
		
		// Start a timer to submit another URB
		NdisMInitializeTimer( &UsbPhy->AdapterHandle->IntUrbTimer, 
					UsbPhy->AdapterHandle,
					SubmitIntUrb, TransferResource );
		NdisMSetTimer( &UsbPhy->AdapterHandle->IntUrbTimer, HZ);
	}
    return;

} // UsbPhyInterruptComplete

////////////////////////////////////////////////////////////////////////////////
//
// NAME: UsbPhyProcessNotification
//
// ABSTRACT:
//		Called to process a notification from the device (received over the
//		interrupt pipe)
//
// RETURNS:
//	   	NTSTATUS
//
// ASSUMPTIONS:
//		None.
//
////////////////////////////////////////////////////////////////////////////////
void
UsbPhyProcessNotification(
	PUSB_PHY		UsbPhy,
	PXFER_DATA_BUFFER	Buffer
)
{
	PCDC_NOTIFICATION				Notification;
	PCDC_CONNECTION_SPEED_CHANGE	ConnectionSpeedChange;
#ifdef GSI_USB
	PCDC_DSP_DOWNLOAD_INFO     dspDownloadInfo;
	KIRQL Irql;
#endif /* GSI_USB */

	//DBGPRINT(
	//	DBG_PRINT_MODULE_USB_ALL,
	//	DBG_PRINT_LEVEL_TRACE,
	//	("UsbPhyProcessNotification\n"));

	ASSERT( UsbPhy);
	ASSERT( Buffer);

#ifdef GSI_USB
	// Read the EOC and status information from the usb packet
	if (Buffer->Size == MAX_INTERRUPT_PIPE_SIZE)
	{
		if ( ( ((Buffer->Data)[0]) == 0xF0 ) && (UsbPhy->eocDatalen == 0) )
		{
			COPY_MEMORY( UsbPhy->eocBuffer, Buffer->Data, 
					MAX_INTERRUPT_PIPE_SIZE);
			UsbPhy->eocDatalen = MAX_INTERRUPT_PIPE_SIZE;
			return;
		}
		else if (UsbPhy->eocDatalen == MAX_INTERRUPT_PIPE_SIZE)
		{
			COPY_MEMORY( &(UsbPhy->eocBuffer[MAX_INTERRUPT_PIPE_SIZE]), 
					Buffer->Data, 
					MAX_INTERRUPT_PIPE_SIZE);
			UsbPhy->eocDatalen = 0;
			// EOC buffer starts at byte 5 and ends at byte 40
			DslEocParse( UsbPhy->DslHandle , &(UsbPhy->eocBuffer[2]) );

			// move the status indicator byte (overwrite last EOC byte)
			UsbPhy->eocBuffer[39] = UsbPhy->eocBuffer[1];
			DslStatusParse( UsbPhy->DslHandle, &(UsbPhy->eocBuffer[39]) );
			return;
		}
	}
#endif /* GSI_USB */
	
	if ( Buffer->Size < CDC_MINIMUM_NOTIFICATION_MSG_SIZE ) {
		DBGPRINT(
			DBG_PRINT_MODULE_USB_ALL,
			DBG_PRINT_LEVEL_ERR,
			("Invalid Notification Msg Size (%lu bytes)\n",
			Buffer->Size));
		return;
	}	

	Notification = (PCDC_NOTIFICATION)(Buffer->Data);

    Notification->wValue = le16_to_cpu(Notification->wValue);
    Notification->wIndex = le16_to_cpu(Notification->wIndex);
    Notification->wLength = le16_to_cpu(Notification->wLength);
	
	DBGPRINT( DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
		("CDCNotif ReqType = %x, Notif = %x, Val = %x, Index = %x, Len = %x Buff Size = %lx\n",
		Notification->bmRequestType, Notification->bNotification,
		Notification->wValue, Notification->wIndex, Notification->wLength, Buffer->Size));

	switch ( Notification->bNotification) {

	case CDC_NOTIFICATION_NETWORK_CONNECTION:
		DBGPRINT(
			DBG_PRINT_MODULE_USB_ALL,
			DBG_PRINT_LEVEL_TRACE,
			("NETWORK_CONNECTION (State = %d)\n",
			Notification->wValue));
		//
		// wValue contains the NETWORK_CONNECTION state
		//
		switch( Notification->wValue) {

		case CDC_NETWORK_CONNECTION_DISCONNECT:
			UsbPhyDslDisconnected(UsbPhy);
			break;

		case CDC_NETWORK_CONNECTION_CONNECT:
			UsbPhy->LinkState =   ATM_PHY_LINK_STATE_CONNECTED;
			//
			// Indicate the new link state
			//
			(UsbPhy->UserLinkStatusChange)(
				UsbPhy->UserHandle,
				UsbPhy->LinkState,
				UsbPhy->UpstreamRate,
				UsbPhy->DownstreamRate,
				0,
				UsbPhy->RxCellPadding);
			break;

		default:
			DBGPRINT(
				DBG_PRINT_MODULE_USB_ALL,
				DBG_PRINT_LEVEL_ERR,
				("Invalid NETWORK_CONNECT State (%d)\n",
				Notification->wValue));
			break;
		}
		break;

	case CDC_NOTIFICATION_CONNECTION_SPEED_CHANGE:
		DBGPRINT(
			DBG_PRINT_MODULE_USB_ALL,
			DBG_PRINT_LEVEL_TRACE,
			("CONNECTION_SPEED_CHANGE\n"));

		//
		// Make sure the data buffer is the correct length
		//
		if ( Buffer->Size != CDC_CONNECTION_SPEED_CHANGE_MSG_SIZE) {
			DBGPRINT(
				DBG_PRINT_MODULE_USB_ALL,
				DBG_PRINT_LEVEL_ERR,
				("Invalid CONNECTION_SPEED_CHANGE msg size (%lu bytes)\n",
				Buffer->Size));
			return;
		}

		ConnectionSpeedChange = (PCDC_CONNECTION_SPEED_CHANGE)
									(Notification->Data);

		DBGPRINT(
			DBG_PRINT_MODULE_USB_ALL,
			DBG_PRINT_LEVEL_TRACE,
			("New Connection Speed: (US = %lu) (DS = %lu)\n",
			ConnectionSpeedChange->UpStreamBitRate,
			ConnectionSpeedChange->DownStreamBitRate));

		UsbPhy->UpstreamRate 	= le32_to_cpu(ConnectionSpeedChange->UpStreamBitRate);
		UsbPhy->DownstreamRate	= le32_to_cpu(ConnectionSpeedChange->DownStreamBitRate);

		//
		// Indicate the new link state
		//
		(UsbPhy->UserLinkStatusChange)(
			UsbPhy->UserHandle,
			UsbPhy->LinkState,
			UsbPhy->UpstreamRate,
			UsbPhy->DownstreamRate,
			0,
			UsbPhy->RxCellPadding);

		break;

#ifdef GSI_USB
   case CDC_NOTIFICATION_DSP_INT:

      if (GetDspState(UsbPhy) != DSP_PHY_DISCONNECT)
      {
		UsbPhy->isIpuReady = Notification->wValue;
		UsbPhy->getNextState = Notification->wIndex;
	    //DBGPRINT(
		//  DBG_PRINT_MODULE_USB_ALL,
		//  DBG_PRINT_LEVEL_TRACE,
        // ("\nwValue/isIpuReady = %x\nwIndex/getNextState = %x\n\n",
        //             UsbPhy->isIpuReady, UsbPhy->getNextState));

          if(!UsbPhy->getNextState) //watchdog
          {
           

              SetDspState(UsbPhy, DSP_PHY_RESTART, DSP_INTERRUPT);
          
          }
          else 
          {
  		 SetDspState(UsbPhy, DSP_PHY_EMPTY,DSP_INTERRUPT);	
          }
      }
		
      break;
#endif /* GSI_USB */

	default:
		//DBGPRINT(
		//	DBG_PRINT_MODULE_USB_ALL,
		//	DBG_PRINT_LEVEL_ERR,
		//	("Invalid Notification (0x%X)\n",
		//	Notification->bNotification));
		break;
	}

} // UsbPhyProcessNotification

void SubmitIntUrb(
	PVOID				SystemSpecific1,
	PUSB_PHY_TRANSFER_RESOURCE		TransferResource,	   
	PVOID				SystemSpecific2,
	PVOID				SystemSpecific3
)
{
	PUSB_PHY	UsbPhy;

	if( TransferResource == NULL )
		return;
	
	UsbPhy = TransferResource->UsbPhy;
	DBGPRINT( DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
		("Calling InitializeIntURB TxRes = %p\n", TransferResource));

	UsbPhyInitializeInterruptTransferResource( UsbPhy,TransferResource);
	// submit the request
	DBGPRINT(
		DBG_PRINT_MODULE_USB_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("Resubmitting Interrupt URB\n"));

	if( usb_submit_urb( TransferResource->Urb ) )
	{
		DBGPRINT( DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_ERR,
			("Unable to resubmit URB\n"));
	}
	else
	{
		TransferResource->InProgress = TRUE;
		atomic_inc(&(UsbPhy->InterruptPendingCount));
	}
}


