/************************************************************************
 * FILE NAME   : usbiso.c                                               *
 * DESCRIPTION : Contains the isochronous receive data routines         *
 *                                                                      *
 * COMMENT     :                                                        *
 *                                                                      *
 * (C) Copyright 1999 by GlobeSpan Semiconductor Inc                    *
 ************************************************************************/

#include "Gp.h"
#include "common.h"
#include "debug.h"
#include "atmphy.h"
#include "usbphy.h"
#include "usbcdc.h"

#ifdef DBG_TIMER
void RecordTimestamp(UCHAR id, USHORT numBytes);
#endif

void FreeIsoUrbFromQedPkts(PUSB_PHY UsbPhy);

////////////////////////////////////////////////////////////////////////////////
//
// NAME: UsbPhyStartIsoReceive
//
// ABSTRACT:
//    Starts the Isochronous receive pipe by initializing and submitting the
//      transfer request URBs
//
// RETURNS:
//       NTSTATUS
//
// ASSUMPTIONS:
//    None.
//
////////////////////////////////////////////////////////////////////////////////
NTSTATUS
UsbPhyStartIsoReceive( PUSB_PHY UsbPhy)
{
	DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
		("UsbPhyStartIsoReceive for %lx Packets\n",
		UsbPhy->IsoReceiveResourceCount));

	UsbPhy->TransferBuffer = NULL;
	UsbPhy->TransferResource = NULL;

	UsbPhy->partialCellLen = 0;
	atomic_set(&(UsbPhy->IsoPendingReceive), 0);

	UsbPhy->BufferCount = (ULONG) UsbPhy->IsoReceiveResourceCount;
	UsbPhy->PacketSize = UsbPhy->IsoReceivePipePacketSize;

	UsbPhy->FramesPerBuffer =
		(ULONG)((UsbPhy->IsoReceivePipeTransferSize) / (UsbPhy->IsoReceivePipePacketSize));

	if (UsbPhy->FramesPerBuffer > 10)
		UsbPhy->FramesPerBuffer = 10;

	// Add padding to prevent buffer overrun (some host controllers dont work
	// if we add extra padding per frame so this is an alternative solution)
	UsbPhy->TransferBufferLength = (MAX_ISO_PIPE_PACKET_SIZE - UsbPhy->IsoReceivePipePacketSize) + 
		( UsbPhy->BufferCount * ((UsbPhy->PacketSize * UsbPhy->FramesPerBuffer) + (2 * UsbPhy->ReceiveDataSize)) );

	ALLOCATE_MEMORY( &(UsbPhy->TransferBuffer), UsbPhy->TransferBufferLength );
	if (UsbPhy->TransferBuffer == NULL)
	{
		UsbPhyStopIsoReceive (UsbPhy);
		return( STATUS_UNSUCCESSFUL);
	}

	ALLOCATE_MEMORY( &UsbPhy->TransferResource, 
		sizeof(USB_PHY_TRANSFER_RESOURCE) * UsbPhy->BufferCount);
	if (UsbPhy->TransferResource == NULL)
	{
		UsbPhyStopIsoReceive (UsbPhy);
		return( STATUS_UNSUCCESSFUL);
	}

	ALLOCATE_MEMORY( &UsbPhy->cellPtr, (2 * UsbPhy->ReceiveDataSize));
	if (UsbPhy->cellPtr == NULL)
	{
		UsbPhyStopIsoReceive (UsbPhy);
		return( STATUS_UNSUCCESSFUL);
	}

	return UsbPhyIsoSubmitUrbs(UsbPhy);

} // UsbPhyStartIsoReceive

////////////////////////////////////////////////////////////////////////////////
//
// NAME: UsbPhyIsoSubmitUrbs
//
// ABSTRACT:
//    Allocates and submits a number of URBs to the USBD.
//
// RETURNS:
//    STATUS_SUCCESS or STATUS_UNSUCCESSFUL
//
// ASSUMPTIONS:
//
////////////////////////////////////////////////////////////////////////////////
NTSTATUS
UsbPhyIsoSubmitUrbs(PUSB_PHY UsbPhy)
{
	ULONG                      i, j;
	PUSB_PHY_TRANSFER_RESOURCE TransferResource;
	PURB                       Urb;

	//
	// Reset the pipe prior to submitting any requests
	//
	UsbPhyResetIsoPipe( UsbPhy );
	DBGPRINT( DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
		("UsbPhyIsoSubmitUrbs %lx\n", UsbPhy->BufferCount));

	UsbPhy->partialCellLen = 0;
	UsbPhy->ExpectedIsoPktTag = 0;
	TransferResource = NULL;

	for (i=0; i < UsbPhy->BufferCount; i++)
	{
		TransferResource = &(UsbPhy->TransferResource[i]);
		TransferResource->tag = i;

		//
		// allocate and prepare the URB
		//
		Urb = usb_alloc_urb( UsbPhy->FramesPerBuffer);
		if (Urb == NULL)
		{
			UsbPhyStopIsoReceive (UsbPhy);
			return( STATUS_UNSUCCESSFUL);
		}
		DBGPRINT( DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
			("Allocated Urb %p\n", Urb));
		TransferResource->Urb = Urb;

		// We are reading from the device
		Urb->transfer_flags = 
					USB_DIR_IN | USB_ISO_ASAP;// | USBD_SHORT_TRANSFER_OK; TBD use it or not

		Urb->transfer_buffer_length = UsbPhy->PacketSize * UsbPhy->FramesPerBuffer;
		Urb->complete = UsbPhyIsoReceiveComplete;
		Urb->context = TransferResource;
		Urb->dev = UsbPhy->AdapterHandle->usbDev;
		Urb->pipe = UsbPhy->IsoReceivePipeHandle;
		Urb->transfer_buffer =
			((PUCHAR) (UsbPhy->TransferBuffer)) + 
			//(2 * UsbPhy->ReceiveDataSize) +
			(TransferResource->tag * ((UsbPhy->PacketSize * UsbPhy->FramesPerBuffer) + (2 * UsbPhy->ReceiveDataSize)));

		Urb->number_of_packets = UsbPhy->FramesPerBuffer;

		for (j=0; j< UsbPhy->FramesPerBuffer; j++)
		{
			Urb->iso_frame_desc[j].offset =
						   j * UsbPhy->PacketSize;
			Urb->iso_frame_desc[j].length =
						   UsbPhy->IsoReceivePipePacketSize;
		}

		TransferResource->UsbPhy = UsbPhy;
		TransferResource->Buffer = Urb->transfer_buffer;
	}

	DBGPRINT( DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
		("Total ISO Count %d\n", UsbPhy->BufferCount));

	for (i=0; i < UsbPhy->BufferCount; i++)
	{
		TransferResource = &(UsbPhy->TransferResource[i]);
		DBGPRINT( DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
			("Submitting Urb %p at Res %p Index %d\n",
			TransferResource->Urb, TransferResource, i));
		if( usb_submit_urb( TransferResource->Urb ) )
		{
			DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
				("URB %p at Res %p Failed\n",
				TransferResource->Urb, TransferResource));
			UsbPhyStopIsoReceive (UsbPhy);
			return( STATUS_UNSUCCESSFUL);
		}

		DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
			("URB %p at Res %p submitted successfully\n",
			TransferResource->Urb, TransferResource));
		TransferResource->InProgress = TRUE;
		atomic_inc(&(UsbPhy->IsoPendingReceive));
	}

	return STATUS_SUCCESS;
}   

////////////////////////////////////////////////////////////////////////////////
//
// NAME: UsbPhyStopIsoReceive
//
// ABSTRACT:
//    Stops the isochronous receive pipe.
//
// RETURNS:
//       void
//
// ASSUMPTIONS:
//    Runs at IRQL PASSIVE_LEVEL
//
////////////////////////////////////////////////////////////////////////////////
void
UsbPhyStopIsoReceive( PUSB_PHY UsbPhy)
{
	DBGPRINT(
		DBG_PRINT_MODULE_USB_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("UsbPhyStopIsoReceive\n"));

	ASSERT( UsbPhy);

	if(in_interrupt())
	{
		DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE, 
			("ISO ABORT DEFERRED IN INTERRUPT\n"));
		return;
	}

	if(UsbPhy->PipeAborting & ISORX_PIPE_ABORT_PROGRESS)
	{
		DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE, 
			("ISO ABORT ALREADY IN PROGRESS\n"));
		return;
	}
	UsbPhy->PipeAborting |= ISORX_PIPE_ABORT_PROGRESS;
	FreeIsoUrbFromQedPkts(UsbPhy);

	if (atomic_read(&(UsbPhy->IsoPendingReceive)) == 0)
	{
		DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE, 
			("Freeing the resources for ISO\n"));

		// Free the buffer for the data
		if (UsbPhy->TransferBuffer)
		{
			FREE_MEMORY(UsbPhy->TransferBuffer, UsbPhy->TransferBufferLength);
			UsbPhy->TransferBuffer = NULL;
		}

		// Free the buffer for the TransferResource Descriptor
		if (UsbPhy->TransferResource)
		{
			FREE_MEMORY(UsbPhy->TransferResource, 
				sizeof(USB_PHY_TRANSFER_RESOURCE) * UsbPhy->BufferCount);
			UsbPhy->TransferResource = NULL;
		}

		if (UsbPhy->cellPtr)
		{
			FREE_MEMORY(UsbPhy->cellPtr, (2 * UsbPhy->ReceiveDataSize));
			UsbPhy->cellPtr = NULL;
		}
		UsbPhy->PipeAborting &= ~ISORX_PIPE_ABORT_PROGRESS;
		DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE, 
			("ISO Rx Stop Complete\n"));
	}
	DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE, 
		("Exit ISOStop\n"));
} // UsbPhyStopIsoReceive


////////////////////////////////////////////////////////////////////////////////
//
// NAME: FreeIsoUrbFromQedPkts
//
// ABSTRACT:
//    Frees ISO URB from the save (for out of order URB packets) packets 
//
// RETURNS:
//       void
//
// ASSUMPTIONS:
//    None
//
////////////////////////////////////////////////////////////////////////////////
void FreeIsoUrbFromQedPkts(PUSB_PHY UsbPhy)
{
	USB_ISO_RECEIVE_INFO			*isoRxListPtr;
	ULONG							i=0;
	PUSB_PHY_TRANSFER_RESOURCE		TransferResource;

	if( UsbPhy->TransferResource == NULL )
		return;

	for(i=0; i<MAX_RX_RESOURCE_COUNT; i++)
	{
		//if (UsbPhy->IsoSeqPacket[i])
		{
			//isoRxListPtr = (USB_ISO_RECEIVE_INFO *)UsbPhy->IsoSeqPacket[i];
			//TransferResource =(PUSB_PHY_TRANSFER_RESOURCE)isoRxListPtr->Context;
			FreeIsoResources(UsbPhy, &(UsbPhy->TransferResource[i]));
			//UsbPhy->IsoSeqPacket[i] = NULL;
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// NAME: FreeIsoResources
//
// ABSTRACT:
//    Frees resources for UsbPhyIsoReceiveComplete (Irp, Urb, 
//    and TransferBuffer if IsoPendingReceive is zero).
//
// RETURNS:
//       void
//
// ASSUMPTIONS:
//    None
//
////////////////////////////////////////////////////////////////////////////////
void
FreeIsoResources
( 
	PUSB_PHY UsbPhy,
	PUSB_PHY_TRANSFER_RESOURCE TransferResource
)
{
	// Free the Urb buffer
	if (TransferResource && TransferResource->Urb)
	{
		if( TransferResource->InProgress == TRUE )
		{
			DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE, 
				("Unlinking ISO URB %p %d\n", 
				 TransferResource->Urb, 
				 TransferResource->Urb->status));
			TransferResource->Urb->transfer_flags &= 
				~USB_ASYNC_UNLINK;
			if (TransferResource->Urb->status == (-EINPROGRESS))
			{
	        	if(atomic_read(&(UsbPhy->IsoPendingReceive)))
					atomic_dec(&(UsbPhy->IsoPendingReceive));
			}
			// Unlink URB after decrementing the count as urb->status
			// is reset to ENOENT incase of OHCI after unlinking
			usb_unlink_urb(TransferResource->Urb);
		}
		usb_free_urb(TransferResource->Urb );
		TransferResource->Urb = NULL;
		TransferResource->InProgress = FALSE;
	}
}


////////////////////////////////////////////////////////////////////////////////
//
// NAME: UsbPhyIsoReceiveComplete
//
// ABSTRACT:
//    Called by USBD when a previously submitted Iso RX URB has been
//    completed.  If the request was successful, the received data is
//    indicated to the phy user.
//
// RETURNS:
//       NTSTATUS
//
// ASSUMPTIONS:
//    None.
//
////////////////////////////////////////////////////////////////////////////////
void
UsbPhyIsoReceiveComplete(PURB Urb)
{
	PUSB_PHY                   	UsbPhy;
	PUSB_PHY_TRANSFER_RESOURCE 	TransferResource;
	XFER_DATA_BUFFER           	Buffer;
	ULONG                      	offset;
	ULONG                      	isoPacketLen;
	PUCHAR                     	isoPacketPtr;
	UCHAR                      	morePartial;
	ULONG                      	i;
	ULONG                      	tag;
	USB_ISO_RECEIVE_INFO       	*isoRxListPtr;
	PVOID					  	Context;

	TransferResource= (PUSB_PHY_TRANSFER_RESOURCE)(Urb->context);
	Context			= (PVOID)TransferResource;
	UsbPhy  		= TransferResource->UsbPhy;
	tag				= TransferResource->tag;

	//DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
	//	 ("UsbPhyIsoReceiveComplete\n"));

	#ifdef DBG_TIMER
	{
		USHORT totalBytes = 0;
		for (i=0; i<UsbPhy->FramesPerBuffer; i++)
			totalBytes += Urb->iso_frame_desc[i].length;
		RecordTimestamp(5, totalBytes);
	}
	#endif

	if(atomic_read(&(UsbPhy->IsoPendingReceive)))
		atomic_dec(&(UsbPhy->IsoPendingReceive));

	TransferResource->InProgress = FALSE;

	if( UsbPhy->PipeAborting & ISORX_PIPE_ABORT_PROGRESS )
	{
		DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
			("ISO RX ABORT IN PROGRESS Urb status 0x%x Pending Iso : %d\n",
			Urb->status, atomic_read(&(UsbPhy->IsoPendingReceive))));

		return;
	}

	if ( IsNotOperational (UsbPhy) )
	{
		DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
			("Resetting, Urb status 0x%x Pending Iso IRP Count : %d\n",
			Urb->status, atomic_read(&(UsbPhy->IsoPendingReceive))));
		UsbPhyStopIsoReceive( UsbPhy );
		//KKA FreeIsoUrbFromQedPkts(UsbPhy);
		return;
	}

	else if ( Urb->status == USB_ST_STALL )
	{
		// The device has been halted.  This might indicate that a USB babble
		// condition has been detected.  This happens when the USB driver can't
		// service the USB Host Controller fast enough, and the controller sends
		// out an IN token too much time after the SOF token.  This results in
		// the device sending back the ISO data past its 1 ms frame boundary.
		// The host controller will detect this error and disable the port.  
		// We then need to abort all the pipes, get back all our IRPs, 
		// reset the port, and re-submit the URBs.

		DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_ERR,
			("Device halted, Pending Iso IRP Count : %d\n",
			atomic_read(&(UsbPhy->IsoPendingReceive))));

		// Don't process anymore URBs until we are reset
		if( USB_PHY_STATE_RESETTING_PORT != GetUsbState (UsbPhy) )
			SetUsbState (UsbPhy, USB_PHY_STATE_RESETTING_PORT);

		UsbPhyStopIsoReceive( UsbPhy );
		//KKA FreeIsoUrbFromQedPkts(UsbPhy);
		return;
	}

	else if (Urb->status)
	{
		// There was an error in the returned IRP/URB.  When this happens,
		// we have seen that the ISO pipe has been shut down.  We think this
		// happens because something is holding the PCI bus for too long, and
		// the USB controller can't complete its operation in time.  In this
		// case, we will wait for all the URBs to come back, then we will 
		// reset the iso pipe, and then resubmit the URBs.

		DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_ERR,
			("Irp/Urb Error %x, Pending Iso IRP Count : %d\n",
			Urb->status, atomic_read(&(UsbPhy->IsoPendingReceive))));

		// Don't process anymore URBs until we are reset
		//SetUsbState (UsbPhy, USB_PHY_STATE_RESETTING_PORT);

		UsbPhyStopIsoReceive( UsbPhy );
		return;
	}
	else
	{
		UsbPhy->isoQBuff[tag].Context = Context;
		isoRxListPtr = &(UsbPhy->isoQBuff[tag]);

		if (tag != UsbPhy->ExpectedIsoPktTag)
		{
			DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_ERR,
				("Tag Mismatch\n"))
			UsbPhy->IsoSeqPacket[tag] = (PVOID)&(UsbPhy->isoQBuff[tag]);
			return; //STATUS_MORE_PROCESSING_REQUIRED;
		}

	   do
	   {
			// No errors occurred, so process each Iso packet and call higher layer
			// (ATM) to process cells.  Also, re-initialize and re-submit the IRP/URB

			Context				= isoRxListPtr->Context;
			TransferResource	= (PUSB_PHY_TRANSFER_RESOURCE) Context;
			Urb					= TransferResource->Urb;
			tag					= TransferResource->tag;

			for (i=0; i<UsbPhy->FramesPerBuffer; i++)
			{
				if ( (Urb->iso_frame_desc[i].actual_length ) && 
					 (Urb->iso_frame_desc[i].actual_length <= 
							UsbPhy->IsoReceivePipePacketSize) &&
					 (!(Urb->iso_frame_desc[i].status))) 
				{
					//DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_ERR,
					//	("Processing Buffer 1\n"))
					isoPacketLen = Urb->iso_frame_desc[i].actual_length + 
										UsbPhy->partialCellLen;
					offset = i * UsbPhy->PacketSize;
					isoPacketPtr = ((PUCHAR)(Urb->transfer_buffer)) + offset - UsbPhy->partialCellLen;

					if (UsbPhy->partialCellLen)
						COPY_MEMORY( isoPacketPtr, UsbPhy->cellPtr, UsbPhy->partialCellLen);
#ifdef	USB_HEXDUMP
					if( i == 0 )
					{
						HexDump(isoPacketPtr,isoPacketLen);
					}
#endif	//USB_HEXDUMP
					Buffer.Data = isoPacketPtr;
					Buffer.Size = ((int)(isoPacketLen / UsbPhy->ReceiveDataSize)) * UsbPhy->ReceiveDataSize;
			
					if (Buffer.Size)
					{
						//DBGPRINT(DBG_PRINT_MODULE_USB_ALL, 
						//	DBG_PRINT_LEVEL_ERR, ("Processing Buffer 2\n"))
						morePartial = (UCHAR)(isoPacketLen - Buffer.Size);
	#ifdef WINHIP
						if (GetDspState(UsbPhy) == DSP_PHY_CONNECTED || GetDspState(UsbPhy) == DSP_PHY_WHIP_MODE)
	#else
						if (GetDspState(UsbPhy) == DSP_PHY_CONNECTED )
	#endif
						{
							DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_ERR,
								("Calling User Receive\n"))
							UsbPhy->partialCellLen = (UsbPhy->UserReceive)( UsbPhy->UserHandle, &Buffer);

							if ( (UsbPhy->partialCellLen + morePartial) <= (2 * UsbPhy->ReceiveDataSize) )
								UsbPhy->partialCellLen += morePartial;
							else
								UsbPhy->partialCellLen = 0;

							if (UsbPhy->partialCellLen)
							{
								isoPacketPtr += isoPacketLen - UsbPhy->partialCellLen;
								COPY_MEMORY( UsbPhy->cellPtr, isoPacketPtr, UsbPhy->partialCellLen);
							}
						}
						else
						{
							DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_ERR,
								("Not Calling User Receive Phy not connected\n"))
							UsbPhy->partialCellLen = 0;
						}
					}
					else
					{
						//DBGPRINT(DBG_PRINT_MODULE_USB_ALL, 
						//	DBG_PRINT_LEVEL_ERR, ("Processing Buffer 3\n"))
						UsbPhy->partialCellLen = (UCHAR)(isoPacketLen);

						if (UsbPhy->partialCellLen)
							COPY_MEMORY( UsbPhy->cellPtr, isoPacketPtr, UsbPhy->partialCellLen);
					}
				}
			}

			UsbPhy->IsoSeqPacket[tag] = NULL;
			ZERO_MEMORY(Urb, sizeof(Urb) + UsbPhy->FramesPerBuffer);
			// We are reading from the device
			Urb->transfer_flags = 
						USB_DIR_IN | USB_ISO_ASAP;// | USBD_SHORT_TRANSFER_OK; TBD use it or not

			Urb->transfer_buffer_length = UsbPhy->PacketSize * UsbPhy->FramesPerBuffer;
			Urb->complete = UsbPhyIsoReceiveComplete;
			Urb->context = TransferResource;
			Urb->dev = UsbPhy->AdapterHandle->usbDev;
			Urb->pipe = UsbPhy->IsoReceivePipeHandle;
			Urb->number_of_packets = UsbPhy->FramesPerBuffer;

			for (i=0; i< UsbPhy->FramesPerBuffer; i++)
			{
				Urb->iso_frame_desc[i].offset =
							   i * UsbPhy->PacketSize;
				Urb->iso_frame_desc[i].length =
							   UsbPhy->IsoReceivePipePacketSize;
			}


			//
			// submit the request
			//
			if( usb_submit_urb( Urb ) )
			{
				DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
					("ISO URB RESUBMIT %p at Res %p Failed\n",
					Urb, TransferResource));
				UsbPhyStopIsoReceive (UsbPhy);
				return;
			}
			TransferResource->InProgress = TRUE;
			atomic_inc(&(UsbPhy->IsoPendingReceive));
			UsbPhy->ExpectedIsoPktTag++;
			if (UsbPhy->ExpectedIsoPktTag == UsbPhy->BufferCount)
			   UsbPhy->ExpectedIsoPktTag = 0;
	   } while((isoRxListPtr = ((USB_ISO_RECEIVE_INFO *)UsbPhy->IsoSeqPacket[UsbPhy->ExpectedIsoPktTag])));
	}

	#ifdef DBG_TIMER
	RecordTimestamp(6, 0);
	#endif

	return;

} // UsbPhyIsoReceiveComplete


////////////////////////////////////////////////////////////////////////////////
//
// NAME: UsbPhyResetIsoPipe
//
// ABSTRACT:
//		Resets an Iso pipe and then unstalls it.
//
// RETURNS:
//	   NTSTATUS
//
// ASSUMPTIONS:
//		None.
//
////////////////////////////////////////////////////////////////////////////////
NTSTATUS
UsbPhyResetIsoPipe(	PUSB_PHY UsbPhy)
{
	NTSTATUS ntStatus;

	DBGPRINT(DBG_PRINT_MODULE_USB_ALL, DBG_PRINT_LEVEL_TRACE,
	  ("UsbPhyResetIsoPipe\n"));

	// Reset iso pipe
	ntStatus = UsbPhyResetPipe( UsbPhy, UsbPhy->IsoReceivePipeHandle );
	return(ntStatus);
}


