////////////////////////////////////////////////////////////////////////////////
//
// INTEL CORPORATION PROPRIETARY INFORMATION               
// This software is supplied under the terms of a license agreement 
// or non-disclosure agreement with Intel Corporation and may not be 
// copied nor disclosed except in accordance with the terms of that 
// agreement.                               
// Copyright  1998 Intel Corporation. All Rights Reserved.
//
// PVCS Information:
//     $WorkFile$
//     $Revision: 1.1 $
//     $Modtime: 8/06/01 4:02p $
//     $Author: kattre $
//
// ABSTRACT:
//	   Contains data structures, function declarations, and macros for the
//     CRC module
//
////////////////////////////////////////////////////////////////////////////////
#include "GpSnull.h"

static unsigned int CrcTable[256];


///////////////////////////////////////////////////////////////////////////////
//
// Package to compute 32-bit CRC one byte at a time using
// the high-bit first (Big-Endian) bit ordering convention
//
// Synopsis:
//  GenerateCrcTable() -- generates a 256-word table containing all CRC
//                     remainders for every possible 8-bit byte.  It
//                     must be executed (once) before any CRC updates.
//
//  It is assumed that an unsigned long is at least 32 bits wide and
//  that the predefined type char occupies one 8-bit byte of storage.
//
//  The generator polynomial used for this version of the package is
//  x^32+x^26+x^23+x^22+x^16+x^12+x^11+x^10+x^8+x^7+x^5+x^4+x^2+x^1+x^0 
//  as specified in the Autodin/Ethernet/ADCCP protocol standards. 
//  Other degree 32 polynomials may be substituted by re-defining the
//  symbol POLYNOMIAL below.  Lower degree polynomials must first be  
//  multiplied by an appropriate power of x.  The representation used
//  is that the coefficient of x^0 is stored in the LSB of the 32-bit
//  word and the coefficient of x^31 is stored in the most significant
//  bit.  The CRC is to be appended to the data most significant byte
//  first.  For those protocols in which bytes are transmitted MSB 
//  first and in the same order as they are encountered in the block
//  this convention results in the CRC remainder being transmitted with
//  the coefficient of x^31 first and with that of x^0 last (just as
//  would be done by a hardware shift register mechanization).
//
//  The table lookup technique was adapted from the algorithm described
//  by Avram Perez, Byte-wise CRC Calculations, IEEE Micro 3, 40 (1983).
//
///////////////////////////////////////////////////////////////////////////////
#define POLYNOMIAL 0x04c11db7L


///////////////////////////////////////////////////////////////////////////////
//
//  NAME: CrcGenerateTable
//            
//  ABSTRACT: 
//		Generates a table of CRC remainders for all possible bytes
//
//  RETURNS: 
//		void
//              
//  ASSUMPTIONS: 
//		Must be called prior to calling CrcCalculate   
//
///////////////////////////////////////////////////////////////////////////////
void 
CrcGenerateTable( void)
{ 
   register ULONG i;  
   register UCHAR j;
   register ULONG CrcAccum;
   
   for( i = 0; i < 256; i++ ) { 
      CrcAccum = ( (unsigned long) i << 24 );
      for ( j = 0;  j < 8;  j++ ) { 
         if ( CrcAccum & 0x80000000L ) {
            CrcAccum = ( CrcAccum << 1 ) ^ POLYNOMIAL;
         }
         else {
            CrcAccum = ( CrcAccum << 1 ); 
         }
      }
      CrcTable[i] = CrcAccum; 
   } // for

   return; 

} // CrcGenerateTable

///////////////////////////////////////////////////////////////////////////////
//
//  NAME: CrcCalculate
//            
//  ABSTRACT: 
//		Computes the CRC of 'count' bytes in the given cell payload,
//      using the initialized or previous partial CRC value
//
//  RETURNS: 
//		the new partial CRC
//              
//  ASSUMPTIONS: 
//		Call CrcGenerateTable prior to calling this routine
//
///////////////////////////////////////////////////////////////////////////////
ULONG
CrcCalculate(
   PUCHAR Cell,
   UCHAR  Length,
   ULONG  Crc)
{ 
   register ULONG i, j;
   register ULONG NewCrc;

   NewCrc = Crc;

   for ( j = 0;  j < Length;  j++ )  {
      i = ( (int) ( NewCrc >> 24) ^ Cell[j]) & 0xff;
      NewCrc = ( NewCrc << 8 ) ^ CrcTable[i]; 
   } // for
   
   return( NewCrc); 

} // CrcCalculate
