////////////////////////////////////////////////////////////////////////////////
//
// GlobespanVirata Inc. PROPRIETARY INFORMATION
// This software is supplied under the terms of a license agreement
// or non-disclosure agreement with GlobespanVirata Inc. and may not be
// copied nor disclosed except in accordance with the terms of that
// agreement.
// Copyright  1998-2002 GlobespanVirata Inc. All Rights Reserved.
//
// PVCS Information:
//     $WorkFile$
//     $Revision: 1.1 $
//     $Modtime: 5/15/02 12:18p $
//     $Author: kattre $
//
// ABSTRACT:
//      Contains the implementation for the RFC1483 module.  The RFC1483 module
//      supports four PDU options:
//
//      1. LLC Encapsulated Ethernet
//      2. VC Encapsulated Ethernet (without FCS)
//      3. LLC Encapsulated IP
//      4. VC Encapsulated IP
//
//      The type of PDU option is set during initialization (via the Rfc1483Mode
//      registry parameter).  The default mode is LLC Encapsulated Ethernet.
//      Each of the PDU options is described briefly below.  See RFC1483 for
//      details.
//
//      1. LLC Encapsulated Ethernet
//
//      Ethernet frames are accepted from NDIS and encapsulated as shown below.
//      Note that the PID value of 0x00-07 is used, indicating that the LAN
//      FCS (Frame Check Sequence) is not present.
//
//               +-------------------------------+
//               |       LLC 0xAA-AA-03          |
//               +-------------------------------+
//               |       OUI 0x00-80-C2          |
//               +-------------------------------+
//               |       PID 0x00-07             |
//               +-------------------------------+
//               |       PAD 0x00-00             |
//               +-------------------------------+
//               |    MAC destination address    |
//               +-------------------------------+
//               |                               |
//               |   (remainder of MAC frame)    |
//               |                               |
//               +-------------------------------+
//
//      Once encapsulated, the RFC1483 PDUs are sent using AAL5 on the PVC
//      provisioned during initialization.
//
//      Inbound LLC encapsulated 802.3 frames are accepted from the ATM UNI
//      module.  Note that in the inbound direction, either PID value is
//      allowed.  If the PID value is 0x00-01 (indicating LAN FCS is present),
//      the LAN FCS is simply discarded.  The LLC encapsulation is removed and
//      the Ethernet frame is indicated to NDIS.
//
//      2. VC Encapsulated Ethernet
//
//      Same as (1) above with the exception that the LLC header is ommitted
//      (except for the PAD field).  Thus the PDUs transmitted are as shown
//      below.
//
//               +-------------------------------+
//               |       PAD 0x00-00             |
//               +-------------------------------+
//               |    MAC destination address    |
//               +-------------------------------+
//               |                               |
//               |   (remainder of MAC frame)    |
//               |                               |
//               +-------------------------------+
//
//      Note that the inclusion of the LAN FCS is not supported with VC
//      Encapsulated Ethernet.
//
//      3. LLC Encapsulated IP
//
//      Ethernet frames are accepted from NDIS and the Ethernet header is
//      discarded up to the EtherType.  If the EtherType is not IP, the
//      frame is discarded.  Otherwise, the frame is encapsulated and sent
//      as shown below.
//
//               +-------------------------------+
//               |       LLC 0xAA-AA-03          |
//               +-------------------------------+
//               |       OUI 0x00-00-00          |
//               +-------------------------------+
//               | EtherType 0x08-00 (IP)        |
//               +-------------------------------+
//               |                               |
//               |           IP PDU              |
//               |                               |
//               +-------------------------------+
//
//      In addition, the RFC1483 module will generate responses to outbound
//      Ethernet ARP requests in order to satisfy the IP stack.
//
//      4. VC Encapsulated IP
//
//      Same as (3) above except that the LLC header and Ethertype are ommitted.
//      Thus the PDUs transmitted are as shown below.
//
//               +-------------------------------+
//               |                               |
//               |           IP PDU              |
//               |                               |
//               +-------------------------------+
//
////////////////////////////////////////////////////////////////////////////////
#if defined (LAN_MINIPORT) || defined (PPPOE_MINIPORT) || defined(GSI_LAN)

#ifdef _NTDDK_
    #include "ndis.h"
#else
    #include "GpSnull.h"
    #include "Gp.h"
    #include "atmuni.h"
#endif // linux
    #include "rfc1483.h"
    #include "rfc1577.h"
//------------------------------------------------------------------------------
//
// Local Constant Definitions
//
//------------------------------------------------------------------------------


#ifdef POWERMGMT
NDIS_PNP_CAPABILITIES GsiPnpCapabilities; 
extern NDIS_STATUS PrepareForPwrStateD3 (void *Instance);
extern NDIS_STATUS PrepareForPwrStateD0(void *Instance);
#endif // POWERMGMT

extern u8  PVCCount;

//
// Structures below must be packed on a 1-byte boundary
//
#include "pshpack1.h"
//------------------------------------------------------------------------------
//
// LLC_HEADER
//		Describes the LLC Header used by RFC1483
//
//------------------------------------------------------------------------------
typedef struct tLLC_HEADER {
    UCHAR  	Llc[3];
    UCHAR  	Oui[3];
    UCHAR  	Pid[2];
    UCHAR   Pad[2];
} LLC_HEADER, *PLLC_HEADER;

//------------------------------------------------------------------------------
//
// MAC_HEADER
//		Describes the Ethernet MAC header
//
//------------------------------------------------------------------------------
typedef struct tMAC_HEADER {
    UCHAR       DstAddress[ETH_LENGTH_OF_ADDRESS];
    UCHAR       SrcAddress[ETH_LENGTH_OF_ADDRESS];
    USHORT      EtherType;
} MAC_HEADER, *PMAC_HEADER;

//------------------------------------------------------------------------------
//
// ETH_ARP_PACKET
//		Describes the Ethernet ARP packet
//
//------------------------------------------------------------------------------
typedef struct tETH_ARP_PACKET {
    USHORT	ArpHrd;
    USHORT	ArpPro;
    UCHAR 	ArpHwLen;
    UCHAR 	ArpProtLen;
    USHORT	ArpOpCode;
    UCHAR 	SendHwAddr[ ETH_LENGTH_OF_ADDRESS];
    UCHAR 	SendIpAddr[ IP_ADDRESS_SIZE];
    UCHAR 	DestHwAddr[ ETH_LENGTH_OF_ADDRESS];
    UCHAR 	DestIpAddr[ IP_ADDRESS_SIZE];
} ETH_ARP_PACKET, *PETH_ARP_PACKET;

#include "poppack.h" // restore default packing

//------------------------------------------------------------------------------
//
// Local Variable Declarations
//
//------------------------------------------------------------------------------
static LLC_HEADER Rfc1483LlcEncapBridgedEthNoFcs =
	{ {0xAA,0xAA,0x03},
	  {0x00,0x80,0xC2},
	  {0x00,0x07},
	  {0x00,0x00}
	};

static LLC_HEADER Rfc1483LlcEncapBridgedEthFcs =
	{ {0xAA,0xAA,0x03},
	  {0x00,0x80,0xC2},
	  {0x00,0x01},
	  {0x00,0x00}
	};

static LLC_HEADER Rfc1483LlcEncapRoutedIp =
	{ {0xAA,0xAA,0x03},
	  {0x00,0x00,0x00},
	  {0x08,0x00}
	};

static LLC_HEADER Rfc1483Rfc1577Encap = 
    {
    {0xAA,0xAA,0x03},
	{0x00,0x00,0x00},
	{0x08,0x06}
    };


static UCHAR	Rfc1483VcEncapBridgedEth[ RFC1483_VC_ENCAP_BRIDGED_ETH_SIZE] =
	{ 0x00,0x00 };

#ifdef _NTDDK_
static NDIS_OID SupportedLanOids[] = {
        OID_GEN_SUPPORTED_LIST,
        OID_GEN_HARDWARE_STATUS,
        OID_GEN_MEDIA_SUPPORTED,
        OID_GEN_MEDIA_IN_USE,
        OID_GEN_MAXIMUM_LOOKAHEAD,
        OID_GEN_MAXIMUM_FRAME_SIZE,
        OID_GEN_MAXIMUM_TOTAL_SIZE,
        OID_GEN_MAC_OPTIONS,
        OID_GEN_PROTOCOL_OPTIONS,
        OID_GEN_LINK_SPEED,
        OID_GEN_TRANSMIT_BUFFER_SPACE,
        OID_GEN_RECEIVE_BUFFER_SPACE,
        OID_GEN_TRANSMIT_BLOCK_SIZE,
        OID_GEN_RECEIVE_BLOCK_SIZE,
        OID_GEN_VENDOR_ID,
        OID_GEN_VENDOR_DESCRIPTION,
        OID_GEN_DRIVER_VERSION,
        OID_GEN_CURRENT_PACKET_FILTER,
        OID_GEN_CURRENT_LOOKAHEAD,
        OID_GEN_XMIT_OK,
        OID_GEN_RCV_OK,
        OID_GEN_XMIT_ERROR,
        OID_GEN_RCV_ERROR,
        OID_GEN_RCV_NO_BUFFER,
        OID_GEN_RCV_CRC_ERROR,
        OID_GEN_MEDIA_CONNECT_STATUS,
        OID_GEN_MAXIMUM_SEND_PACKETS,
        OID_GEN_VENDOR_DRIVER_VERSION,
        OID_802_3_PERMANENT_ADDRESS,
        OID_802_3_CURRENT_ADDRESS,
        OID_802_3_MULTICAST_LIST,
        OID_802_3_MAXIMUM_LIST_SIZE,
        OID_802_3_RCV_ERROR_ALIGNMENT,
        OID_802_3_XMIT_ONE_COLLISION,
        OID_802_3_XMIT_MORE_COLLISIONS
#ifdef POWERMGMT		
		,
		OID_PNP_CAPABILITIES,	
		OID_PNP_QUERY_POWER,	
		OID_PNP_SET_POWER,
		OID_PNP_ENABLE_WAKE_UP,
		OID_PNP_ADD_WAKE_UP_PATTERN,
		OID_PNP_REMOVE_WAKE_UP_PATTERN
#endif

   };

// Maintain this list in the same order as the above list
static char *SupportedLanOidNames[] =
{
        "OID_GEN_SUPPORTED_LIST",
        "OID_GEN_HARDWARE_STATUS",
        "OID_GEN_MEDIA_SUPPORTED",
        "OID_GEN_MEDIA_IN_USE",
        "OID_GEN_MAXIMUM_LOOKAHEAD",
        "OID_GEN_MAXIMUM_FRAME_SIZE",
        "OID_GEN_MAXIMUM_TOTAL_SIZE",
        "OID_GEN_MAC_OPTIONS",
        "OID_GEN_PROTOCOL_OPTIONS",
        "OID_GEN_LINK_SPEED",
        "OID_GEN_TRANSMIT_BUFFER_SPACE",
        "OID_GEN_RECEIVE_BUFFER_SPACE",
        "OID_GEN_TRANSMIT_BLOCK_SIZE",
        "OID_GEN_RECEIVE_BLOCK_SIZE",
        "OID_GEN_VENDOR_ID",
        "OID_GEN_VENDOR_DESCRIPTION",
        "OID_GEN_DRIVER_VERSION",
        "OID_GEN_CURRENT_PACKET_FILTER",
        "OID_GEN_CURRENT_LOOKAHEAD",
        "OID_GEN_XMIT_OK",
        "OID_GEN_RCV_OK",
        "OID_GEN_XMIT_ERROR",
        "OID_GEN_RCV_ERROR",
        "OID_GEN_RCV_NO_BUFFER",
        "OID_GEN_RCV_CRC_ERROR",
        "OID_GEN_MEDIA_CONNECT_STATUS",
        "OID_GEN_MAXIMUM_SEND_PACKETS",
        "OID_GEN_VENDOR_DRIVER_VERSION",
        "OID_802_3_PERMANENT_ADDRESS",
        "OID_802_3_CURRENT_ADDRESS",
        "OID_802_3_MULTICAST_LIST",
        "OID_802_3_MAXIMUM_LIST_SIZE",
        "OID_802_3_RCV_ERROR_ALIGNMENT",
        "OID_802_3_XMIT_ONE_COLLISION",
        "OID_802_3_XMIT_MORE_COLLISIONS"
#ifdef POWERMGMT		
		,
		"OID_PNP_CAPABILITIES",	
		"OID_PNP_QUERY_POWER",	
		"OID_PNP_SET_POWER",
		"OID_PNP_ENABLE_WAKE_UP",
		"OID_PNP_ADD_WAKE_UP_PATTERN",
		"OID_PNP_REMOVE_WAKE_UP_PATTERN"
#endif   

   };

#define NUM_LAN_OID_ENTRIES (sizeof(SupportedLanOids) / sizeof(SupportedLanOids[0]))

// Find the string corresponding to the OID value
char *
GsiGetLanOidString(NDIS_OID Oid)
{
    UINT i;

    for (i = 0; i < NUM_LAN_OID_ENTRIES-1; i++)
    {
        if (SupportedLanOids[i] == Oid)
        {
            break;
        }
    }
    return(SupportedLanOidNames[i]);
}
#endif // _NTDDK_

//------------------------------------------------------------------------------
//
// Local Function Declarations
//
//------------------------------------------------------------------------------
static
BOOLEAN
Rfc1483FilterReceivedPacket(
	PATM_USB_INSTANCE	Instance,
	PMAC_HEADER			MacHeader
);


////////////////////////////////////////////////////////////////////////////////
//
// NAME: Rfc1483Initialize
//
// ABSTRACT:
//		Initializes the RFC1483 layer, creating the ATM UNI and provisioning
//		the RFC1483 PVC.
//
// RETURNS:
//		Boolean
//
// ASSUMPTIONS:
//		Runs in the context of MiniportInitialize.
//
////////////////////////////////////////////////////////////////////////////////
BOOLEAN
Rfc1483Initialize( 
	PATM_USB_INSTANCE 		Instance,
	ATM_PHY_BIND 			PhyBind,
	ATM_PHY_DEVICE_HANDLE	PhyDeviceHandle

)
{
	ATM_UNI_STATUS	AtmUniStatus;
	PUCHAR			Encapsulation;
	UCHAR			EncapsulationSize;
	int				index = 0;

	DBGPRINT(
		DBG_PRINT_MODULE_RFC1483_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("Rfc1483Initialize\n"));

	ASSERT( Instance);
	ASSERT( PhyBind);
	ASSERT( PhyDeviceHandle);


	//
	// Set up the VC encapsulation based on the
	// RFC1483 mode
	//
	switch ( Instance->Rfc1483Mode ) {

	case RFC1483_MODE_BRIDGED_ETH_LLC:
		Encapsulation 		= (PUCHAR)&Rfc1483LlcEncapBridgedEthNoFcs;
		EncapsulationSize	= RFC1483_LLC_ENCAP_BRIDGED_ETH_SIZE;

		Instance->Rfc1483EncapsulationSize = EncapsulationSize;
	    Instance->Rfc1483Encapsulation = Encapsulation;
		break;

	case RFC1483_MODE_BRIDGED_ETH_VC:
		Encapsulation		= (PUCHAR)&Rfc1483VcEncapBridgedEth;
		EncapsulationSize	= RFC1483_VC_ENCAP_BRIDGED_ETH_SIZE;

		Instance->Rfc1483EncapsulationSize = EncapsulationSize;
	    Instance->Rfc1483Encapsulation = Encapsulation;
		break;

	case RFC1483_MODE_ROUTED_IP_LLC:
		Encapsulation		= (PUCHAR)&Rfc1483LlcEncapRoutedIp;
		EncapsulationSize	= RFC1483_LLC_ENCAP_ROUTED_IP_SIZE;

		Instance->Rfc1483EncapsulationSize	= EncapsulationSize;
	    Instance->Rfc1483Encapsulation = Encapsulation;
		break;

	case RFC1483_MODE_ROUTED_IP_VC:
		Encapsulation		= NULL;
		EncapsulationSize	= 0;

		Instance->Rfc1483EncapsulationSize = EncapsulationSize;
	    Instance->Rfc1483Encapsulation = Encapsulation;
		break;

	default:
		DBGPRINT(
			DBG_PRINT_MODULE_RFC1483_ALL,
			DBG_PRINT_LEVEL_ERR,
			("Bad value for RFC1483 MODE (%d)\n",
			Instance->Rfc1483Mode));
		return( FALSE);
		break;
	}		

#ifdef PPPOE_MINIPORT
   if (Instance->Rfc2516Enable)
	    Instance->Rfc1483Encapsulation = Encapsulation;
   else
   {
#endif /* PPPOE_MINIPORT */
	    //
	    // Create the ATM UNI
	    //
#if defined(linux)
         if ((Instance->flags & _ATMUNI_ALLOC) != _ATMUNI_ALLOC)
         {
#endif // linux
	    Instance->AtmUniHandle =
	    AtmUniCreate( 
		    Instance,
		    Rfc1483LinkStatusChange,
		    FALSE, 
		    0,	  
		    RFC1483_MAX_SDU_SIZE,
		    Instance->RxBufferCount,
		    Instance->TxBufferCount,
		    Instance->MiniportAdapterHandle,
		    PhyBind, // This is UsbPhyBind
		    PhyDeviceHandle,
		    Instance->UniqueId,
		    &(Instance->UsbPhyHandle));
#if defined(linux)
         }
#endif // linux         
	    if( NULL == Instance->AtmUniHandle) 
        {
			
		    DBGPRINT(
			    DBG_PRINT_MODULE_RFC1483_ALL,
			    DBG_PRINT_LEVEL_ERR,
			    ("AtmUniCreate failed\n"));
		    return( FALSE);
	    } 

       Rfc1483ReadNetworkAddress( Instance, Instance->UniqueId);

	    //
	    // Create and Activate the 1483 PVC
	    //
	/*
	**	Lets run this loop for the number of PVCs as user wants to configure.
	*/
	for( index = 0 ; index < PVCCount ; index++)
	{
	    Instance->Rfc1483VcHandle[index] =
	    AtmUniCreateVc(
		    Instance->AtmUniHandle,
		    NULL,
		    Rfc1483ReceiveData,
		    NULL,
		    Rfc1483SendComplete);


	    if ( NULL == Instance->Rfc1483VcHandle[index]) {
		    DBGPRINT(
			    DBG_PRINT_MODULE_RFC1483_ALL,
			    DBG_PRINT_LEVEL_ERR,
			    ("AtmUniCreateVc failed\n"));
		    return( FALSE);
	    }

		    AtmUniStatus = 
		    AtmUniActivateVc(
			    Instance->Rfc1483VcHandle[index],
			    Instance->Rfc1483Vpi[index],
			    Instance->Rfc1483Vci[index],
			    AAL5_,
		   	    RFC1483_MAX_SDU_SIZE,
			    (PUCHAR)Encapsulation,
			    EncapsulationSize,
			    FALSE,
	            0);
		
		    if ( ATM_UNI_STATUS_SUCCESS != AtmUniStatus) {
			    DBGPRINT(
				    DBG_PRINT_MODULE_RFC1483_ALL,
				    DBG_PRINT_LEVEL_ERR,
				    ("AtmUniActivateVc failed\n"));
			    return( FALSE);
	    	}
	} /* End of for loop */

#ifdef PPPOE_MINIPORT
   }
#endif /* PPPOE_MINIPORT */

	return( TRUE);
} // Rfc1483Initialize

////////////////////////////////////////////////////////////////////////////////
//
// NAME: Rfc1483Shutdown
//
// ABSTRACT:
//		Shuts down the RFC1483 layer, deleting the 1483 PVC and destroying the
//		ATM UNI.
//
// RETURNS:
//		void
//
// ASSUMPTIONS:
//		The RFC1483 layer has been successfully initialized.
//
////////////////////////////////////////////////////////////////////////////////
void
Rfc1483Shutdown( PATM_USB_INSTANCE Instance)
{
	DBGPRINT(
		DBG_PRINT_MODULE_RFC1483_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("Rfc1483Shutdown\n"));


	ASSERT( Instance);

#ifdef PPPOE_MINIPORT
   if (Instance->Rfc1483Handle != NULL)
       FREE_MEMORY( Instance->Rfc1483Handle, sizeof( RFC1483_HANDLE));
#endif /* PPPOE_MINIPORT */

#if !defined(GSI_WAN)
	AtmUniDestroy( Instance->AtmUniHandle);
#endif // GSI_WAN

} // Rfc1483Shutdown

///////////////////////////////////////////////////////////////////////////////
//
//  NAME: Rfc1483Send
//
//  ABSTRACT:
//		Accepts an NDIS packet for transmission.  The packet is passed to
//		the ATM UNI send function. 
//
//  RETURNS: 
//		NDIS_STATUS
//
//  ASSUMPTIONS:
//		None.
//
///////////////////////////////////////////////////////////////////////////////
NDIS_STATUS 
Rfc1483Send(
	PATM_USB_INSTANCE	Instance,
	PNDIS_PACKET  		Packet,
    ATM_UNI_HANDLE      AtmUniVcHandle,
    BOOLEAN             Arp,
	gsi_lan_t			*plan
    )
{

  	ATM_UNI_STATUS	SendStatus;
	PNDIS_BUFFER  	NdisBuffer;
	UINT        	PacketLength;
	UINT        	BufferCount;
	UINT        	PhyBufferCount;
	UINT			BufferLength;
	PMAC_HEADER		MacHeader;

    int Rfc1483Mode = AtmUniGetRfc1483Mode(AtmUniVcHandle);
	ASSERT( Instance);
	ASSERT( Packet);

	DBGPRINT(
		DBG_PRINT_MODULE_RFC1483_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("Rfc1483Send\n"));


   	//
	// If we are operating in routed IP mode, examine the packet to determine
	// if it is an ARP or IP packet (based on the EtherType).  If the packet
	// is an ARP packet, we generate a dummy reply to keep the stack happy.
	// If the packet is an IP packet, we send it with the appropriate 
	// encapsulation.  If the packet is neither an ARP packet nor an IP packet,
	// we don't send the packet.
	//RFC1483_MODE_ROUTED_IP_LLC
    
	if ( ( (Rfc1483Mode == RFC1483_MODE_ROUTED_IP_LLC) ||
		 (Rfc1483Mode == 	RFC1483_MODE_ROUTED_IP_VC) ) && !Arp)
	{
		// 
		// Get packet information 
		//
	
    NdisQueryPacket( 
			Packet,&PhyBufferCount,&BufferCount,&NdisBuffer,&PacketLength);

		ASSERT( BufferCount);
		ASSERT( NdisBuffer);
		//
		// Get the first buffer (the Ethernet header) and examine the 
		// EtherType
		//
		NdisQueryBuffer( NdisBuffer, (PVOID *)&MacHeader, &BufferLength);
		ASSERT( BufferLength >= ETHERNET_HEADER_SIZE);

        
		if ( MacHeader->EtherType != ETHERTYPE_IP ) 
        {
			if ( MacHeader->EtherType == ETHERTYPE_ARP)
            {
				Rfc1483EthernetArpReply( Instance,Packet,plan);
				return( NDIS_STATUS_SUCCESS);
			}
            else {
				DBGPRINT(
					DBG_PRINT_MODULE_RFC1483_TX,
					DBG_PRINT_LEVEL_ERR,
					("Packet has Invalid EtherType (0x%X)\n",
					MacHeader->EtherType));
				return( NDIS_STATUS_SUCCESS);
			}
		}
        else
        {
                SendStatus = AtmUniSend( 
				AtmUniVcHandle,Packet,TRUE,0,0,FALSE,Arp);
		}
      
    }
    
    else
    {
        /*
		**	The Rfc1483VcHandle here needs to be a specific handle and this can be 
		**  mapped by the info from the gsi_lan_t i.e. the lan interface choosen 
		**  by the user.
		*/

		SendStatus = AtmUniSend( 
			plan->Rfc1483VcHandle,Packet,FALSE,0,0,FALSE,Arp);
	}

	if( SendStatus == ATM_UNI_STATUS_FAILURE)
	{
		DBGPRINT(
			DBG_PRINT_MODULE_RFC1483_TX,
			DBG_PRINT_LEVEL_ERR,
			("Transmit Failed on 1483 PVC [AtmUniSend Failed]\n"));
		return( NDIS_STATUS_FAILURE);
	}
	else if ( SendStatus == ATM_UNI_STATUS_PENDING)
	{
		//
		// The transmit will complete asynchronously with a call to 
		// Rfc1483SendComplete
		//
		return( NDIS_STATUS_PENDING);
	}
	else {
		return( NDIS_STATUS_SUCCESS);
	} // if
}

///////////////////////////////////////////////////////////////////////////////
//
//  NAME: Rfc1483SendComplete
//
//  ABSTRACT: 
//		This function is called to complete a previously pended 
//      AAL5 transmit request
//
//  RETURNS: 
//		void
//
//  ASSUMPTIONS:
//		None.
//
///////////////////////////////////////////////////////////////////////////////
void
Rfc1483SendComplete(
    PATM_USB_INSTANCE	Instance,
	ATM_UNI_USER_HANDLE	NotUsed,
    PNDIS_PACKET   		Packet)
{
	ASSERT( Instance);
	ASSERT( Packet);

	PARAM_NOT_USED( NotUsed);

	DBGPRINT(
		DBG_PRINT_MODULE_RFC1483_TX,
		DBG_PRINT_LEVEL_TRACE,
		("Rfc1483SendComplete\n"));

	NdisMSendComplete(
	  Instance->MiniportAdapterHandle,
	  Packet, 
	  NDIS_STATUS_SUCCESS,
	  NULL); /* This should be pointer to the lan ?? */

} // Rfc1483SendComplete

///////////////////////////////////////////////////////////////////////////////
//
//  NAME: Rfc1483EthernetArpReply
//
//  ABSTRACT: 
//		Creates and indicates an Ethernet ARP reply to NDIS
//
//  RETURNS: 
//		void
//
//  ASSUMPTIONS:
//		None.
//
///////////////////////////////////////////////////////////////////////////////
void
Rfc1483EthernetArpReply(
	PATM_USB_INSTANCE	Instance, 
	PNDIS_PACKET		Packet,
	gsi_lan_t			*plan)
{
	PMAC_HEADER	    EthHeader;
	PETH_ARP_PACKET ArpHeader;
	PNDIS_BUFFER    Buffer;
	PNDIS_BUFFER	OldBuffer;
	UINT	        PacketLength;
	UINT			BufferLength;
	UINT          	BufferCount;
	UINT			PhyBufferCount;
	PUCHAR          ArpReplyBuffer;
	PUCHAR          DataBuffer;
	UINT 	        Offset;
	UCHAR           Addr[ IP_ADDRESS_SIZE];
    UINT            SendAddr;
    UCHAR          DestMacAddr[ETH_LENGTH_OF_ADDRESS];
    int             Status;


	DBGPRINT(
		DBG_PRINT_MODULE_RFC1483_TX,
		DBG_PRINT_LEVEL_TRACE,
		("Rfc1483EthernetArpReply\n"));


	ASSERT( Instance);
	ASSERT( Packet);

	//
	// Extract the first buffer from the packet 
	//
	NdisQueryPacket( 
			Packet, 
			&PhyBufferCount, 
			&BufferCount, 
			&Buffer, 
			&PacketLength);

	if ( PacketLength != (sizeof( MAC_HEADER) + sizeof( ETH_ARP_PACKET))) {
		DBGPRINT(
			DBG_PRINT_MODULE_RFC1483_TX,
			DBG_PRINT_LEVEL_ERR,
			("Packet has invalid length (%d)\n",
			PacketLength));
		return;
	} 

    ALLOCATE_MEMORY( &ArpReplyBuffer, PacketLength);
	if(ArpReplyBuffer == NULL) {
		DBGPRINT(
			DBG_PRINT_MODULE_RFC1483_TX,
			DBG_PRINT_LEVEL_ERR,
			("Unable to allocate ARP reply buffer\n"));
		return;
	} 

	// 
	// Copy the contents of the Ethernet ARP packet into our 
	//   Ethernet ARP reply 
	//
	Offset = 0;
	while ( Buffer) {
		NdisQueryBuffer( Buffer, (PVOID *)&DataBuffer, &BufferLength);
		COPY_MEMORY( (ArpReplyBuffer + Offset), DataBuffer, BufferLength);
		Offset += BufferLength;
		OldBuffer = Buffer;       
		NdisGetNextBuffer( OldBuffer, &Buffer);
	} 
	
	EthHeader = (PMAC_HEADER)ArpReplyBuffer;
	ArpHeader = (PETH_ARP_PACKET)(ArpReplyBuffer + sizeof(MAC_HEADER));

	//
	// If this is a gratuitous ARP request (i.e. source and destination IP
	// addresses are the same), don't reply
	//
	if ( COMPARE_MEMORY( ArpHeader->SendIpAddr,ArpHeader->DestIpAddr,
	                     IP_ADDRESS_SIZE) == 0 )
	{
		DBGPRINT(
			DBG_PRINT_MODULE_RFC1483_TX,
			DBG_PRINT_LEVEL_TRACE,
			("Ignoring Gratuitous ARP Request\n"));

		FREE_MEMORY( ArpReplyBuffer,PacketLength);
		return;
	}

	//
	// Build the Ethernet ARP Reply
	//
	ArpHeader->ArpOpCode = ETHERNET_ARP_OPCODE_REPLY;
	
	//
	// Swap the source and destination IP addresses in the ARP packet
	//

    COPY_MEMORY( Addr, ArpHeader->SendIpAddr, IP_ADDRESS_SIZE);
	COPY_MEMORY( ArpHeader->SendIpAddr,ArpHeader->DestIpAddr,IP_ADDRESS_SIZE);
	COPY_MEMORY( ArpHeader->DestIpAddr,Addr,IP_ADDRESS_SIZE);
	//
	// Swap the destination hardware address and set the source hardware
	//   address to the destination IP address in the ARP packet
	//

    
    
    COPY_MEMORY( ArpHeader->DestHwAddr,ArpHeader->SendHwAddr, 
	             ETH_LENGTH_OF_ADDRESS);
	
    
    if(!Instance->Rfc1577Enable)
    {
        COPY_MEMORY( &ArpHeader->SendHwAddr[2],ArpHeader->SendIpAddr, 
				 IP_ADDRESS_SIZE);
	

    }
           
   else
    {
        COPY_MEMORY ((char *) &SendAddr, ArpHeader->SendIpAddr, IP_ADDRESS_SIZE);
#ifndef PPPOE_MINIPORT /* Do not want to include RFC1577 for the PPPoE complie problem */
        DbgPrintAllEntry(Instance->Rfc1577Config.Rfc1577Context);

        Status = Rfc1577ResolveARP (Instance->Rfc1577Config.Rfc1577Context,
                           SendAddr,
                           DestMacAddr);

        if(RFC1577_FAILURE == Status)
        {
            DBGPRINT(DBG_PRINT_MODULE_RFC1483_TX,DBG_PRINT_LEVEL_TRACE,("Rfc1577ResolveARP returns Failture\n"));
            FREE_MEMORY( ArpReplyBuffer,PacketLength);
		    return;
        }

        DBGPRINT(DBG_PRINT_MODULE_RFC1483_TX,DBG_PRINT_LEVEL_TRACE, ("\n Arp resolved for %x", SendAddr));
        COPY_MEMORY( &ArpHeader->SendHwAddr, DestMacAddr, ETH_LENGTH_OF_ADDRESS);
#endif /* PPPOE_MINIPORT */
        
    } 



    COPY_MEMORY( EthHeader->DstAddress,EthHeader->SrcAddress, 
				 ETH_LENGTH_OF_ADDRESS);
	COPY_MEMORY( &EthHeader->SrcAddress[2],ArpHeader->SendIpAddr, 
	             IP_ADDRESS_SIZE);

    
    
    //
	// Indicate the Ethernet ARP reply to NDIS
	//
	NdisMEthIndicateReceive(
		Instance->MiniportAdapterHandle, NULL, 
		(PUCHAR)EthHeader, sizeof( MAC_HEADER), 
		(PUCHAR)ArpHeader, sizeof( ETH_ARP_PACKET), 
		sizeof( ETH_ARP_PACKET),
		plan); /* This should be pointer to lan ??? TODO */

	NdisMEthIndicateReceiveComplete( Instance->MiniportAdapterHandle);

	FREE_MEMORY( ArpReplyBuffer,PacketLength);

} // Rfc1483EthernetArpReply


///////////////////////////////////////////////////////////////////////////////
//
//  NAME: Rfc1483ReceiveData
//
//  ABSTRACT: 
//		Receives inbound RFC1483 PDUs from the ATM UNI layer and requests
//		a callback at Rfc1483IndicateReceived data to indicate the received
//		data to NDIS within Miniport context.  
//
//  RETURNS: 
//		void
//
//  ASSUMPTIONS:
//		None.
//
///////////////////////////////////////////////////////////////////////////////
void
Rfc1483ReceiveData(
    PATM_USB_INSTANCE	Instance,
	ATM_UNI_USER_HANDLE	AtmUniVcHandle,
	PXFER_DATA_BUFFER	Buffer,
	UCHAR				CpcsUu,
	UCHAR				Cpi
)
{
	BOOLEAN Success;


	DBGPRINT(
		DBG_PRINT_MODULE_RFC1483_RX,
		DBG_PRINT_LEVEL_TRACE,
		("Rfc1483ReceiveData\n"));


	ASSERT( Instance);
	ASSERT( Buffer);

	PARAM_NOT_USED( CpcsUu);
	PARAM_NOT_USED( Cpi);

	//
	// In order to indicate the received data to NDIS, we must schedule
	// a receive callback
	//


    Success = AtmUsbScheduleReceiveCallback( 
				Instance,
				Rfc1483IndicateReceivedData,
				Buffer, AtmUniVcHandle);

	//
	// If our request failed, our callback function won't run. 
	// Drop the received PDU
	//
	if (!Success) 
    {
		DBGPRINT(
			DBG_PRINT_MODULE_RFC1483_RX,
			DBG_PRINT_LEVEL_ERR,
			("Dropping RX PDU - Switch to Miniport Failed\n"));

		AtmUniReceiveComplete( AtmUniVcHandle, Buffer);
	}

} // Rfc1483ReceiveData


///////////////////////////////////////////////////////////////////////////////
//
//  NAME: Rfc1483IndicateReceivedData
//
//  ABSTRACT: 
//		Called within a receive callback to indicate data previously received
//		by Rfc1483ReceiveData
//
//  RETURNS: 
//		void
//
//  ASSUMPTIONS:
//		None
//
///////////////////////////////////////////////////////////////////////////////
void
Rfc1483IndicateReceivedData(
	PATM_USB_INSTANCE	    Instance,
    PXFER_DATA_BUFFER		Buffer,
    ATM_UNI_HANDLE          AtmUniVc
    )
{

	PUCHAR		Data;
	PUCHAR		Header = NULL;
	ULONG		Size;
	MAC_HEADER	MacHeader;
	int			ilanindex = 0;
	gsi_lan_t	*plan = NULL;

    BOOLEAN ArpType = FALSE;
int  Rfc1483Mode;
#ifdef PPPOE_MINIPORT
   PRFC1483_INSTANCE   Rfc1483Instance = (PRFC1483_INSTANCE) Instance->Rfc1483Handle;
#endif /* PPPOE_MINIPORT */

	DBGPRINT(
		DBG_PRINT_MODULE_RFC1483_RX,
		DBG_PRINT_LEVEL_TRACE,
		("Rfc1483IndicateReceivedData\n"));


    if ( (Instance->Status & ATM_USB_STATUS_HALTING) || (Instance->Status & ATM_USB_STATUS_SUSPENDING) )
    {
        AtmUniReceiveComplete( AtmUniVc, Buffer);
        return;
    }

    Rfc1483Mode = AtmUniGetRfc1483Mode(AtmUniVc);

	ASSERT( Instance);
	ASSERT( Buffer);

	Data = Buffer->Data;
	Size = Buffer->Size;


	for(ilanindex = 0; ilanindex< PVCCount; ilanindex++)
	{
		if(AtmUniVc == Instance->lan[ilanindex].Rfc1483VcHandle)
		{
			plan = &(Instance->lan[ilanindex]);
			break;
		}
	} /* We have found the lan device */
	
    
    //
	// Process the received data according to the RFC1483 mode
	//

    switch( Rfc1483Mode) 
    {

	case RFC1483_MODE_BRIDGED_ETH_LLC:
		//
		// Accept packets with or without the FCS
		//
		if ( COMPARE_MEMORY( 
				Data, 
				&Rfc1483LlcEncapBridgedEthNoFcs, 
				RFC1483_LLC_ENCAP_BRIDGED_ETH_SIZE) == 0 ) 
		{
			//
			//  Skip over the LLC header
			//
			Data	+= RFC1483_LLC_ENCAP_BRIDGED_ETH_SIZE;
			Size	-= RFC1483_LLC_ENCAP_BRIDGED_ETH_SIZE;
			//
			// Save a pointer to the Ethernet header and advance 
			// the data pointer past it.
			//
			Header = Data;
			Data  += ETHERNET_HEADER_SIZE;
			Size  -= ETHERNET_HEADER_SIZE;

		} 
        else
            if ( COMPARE_MEMORY(
						Buffer->Data,
						&Rfc1483LlcEncapBridgedEthFcs,
						RFC1483_LLC_ENCAP_BRIDGED_ETH_SIZE) == 0 )
		{
			//
			//  Skip over the LLC header
			//
			Data	+= RFC1483_LLC_ENCAP_BRIDGED_ETH_SIZE;
			Size	-= RFC1483_LLC_ENCAP_BRIDGED_ETH_SIZE;
			//
			// Adjust the buffer size to remove the FCS
			//
			Size	 -= ETHERNET_LAN_FCS_SIZE;
			//
			// Save a pointer to the Ethernet header and advance 
			// the data pointer past it.
			//
			Header = Data;
			Data  += ETHERNET_HEADER_SIZE;
			Size  -= ETHERNET_HEADER_SIZE;

		} 
            else
            {
				DBGPRINT(
				DBG_PRINT_MODULE_RFC1483_RX,
				DBG_PRINT_LEVEL_ERR,
				("Dropped unrecognized PDU (Expecting Bridged Eth LLC)\n"));
			AtmUniReceiveComplete( plan->Rfc1483VcHandle,Buffer);
			return;
		}
		break;

	case RFC1483_MODE_BRIDGED_ETH_VC:
		if ( COMPARE_MEMORY( 
				Data, 
				&Rfc1483VcEncapBridgedEth, 
				RFC1483_VC_ENCAP_BRIDGED_ETH_SIZE) == 0 ) 
		{
			//
			//  Skip over the VC Encapsulation padding
			//
			Data	+= RFC1483_VC_ENCAP_BRIDGED_ETH_SIZE;
			Size	-= RFC1483_VC_ENCAP_BRIDGED_ETH_SIZE;
			//
			// Save a pointer to the Ethernet header and advance 
			// the data pointer past it.
			//
			Header = Data;
			Data  += ETHERNET_HEADER_SIZE;
			Size  -= ETHERNET_HEADER_SIZE;
		} else {
			DBGPRINT(
				DBG_PRINT_MODULE_RFC1483_RX,
				DBG_PRINT_LEVEL_ERR,
				("Dropped unrecognized PDU (Expecting Vc Encap Eth)\n"));
			AtmUniReceiveComplete(plan->Rfc1483VcHandle,Buffer);
			return;
		}
		break;

	case RFC1483_MODE_ROUTED_IP_LLC:
    case RFC1483_MODE_ROUTED_IP_VC:
		if ( Instance->Rfc1483Mode == RFC1483_MODE_ROUTED_IP_LLC) 
        {
			if ( COMPARE_MEMORY( 
					Data, 
					&Rfc1483LlcEncapRoutedIp, 
					RFC1483_LLC_ENCAP_ROUTED_IP_SIZE) == 0 ) 
			{
				//
				//  Skip over the LLC header
				//
				Data	+= RFC1483_LLC_ENCAP_ROUTED_IP_SIZE;
				Size	-= RFC1483_LLC_ENCAP_ROUTED_IP_SIZE;
			}
            
            
            	//Check if it is Arp Packet

            else if ( (Size > RFC1483_RFC1577_ENCAP_SIZE) && (COMPARE_MEMORY( Data, 
		                                            &Rfc1483Rfc1577Encap, 
                                                    RFC1483_RFC1577_ENCAP_SIZE ) == 0) )
            { 
           	        DBGPRINT(
	                DBG_PRINT_MODULE_RFC1483_RX,
	                DBG_PRINT_LEVEL_TRACE,
	                ("Arp Reply arrived!!!\n"));
                    ArpType = TRUE;
                    
  			        Data	+= RFC1483_RFC1577_ENCAP_SIZE;
			        Size	-= RFC1483_RFC1577_ENCAP_SIZE;
                    
            }

            
            else
            {
				DBGPRINT(
					DBG_PRINT_MODULE_RFC1483_RX,
					DBG_PRINT_LEVEL_ERR,
					("Dropped unrecognized PDU (Expecting LLC Encap IP)\n"));
				AtmUniReceiveComplete(plan->Rfc1483VcHandle,Buffer);
				return;
			}
		}
		//
		// Set up the dummy Ethernet header used when indicating the IP packet
		//
        if(!ArpType)
        {
            COPY_MEMORY( &(MacHeader.DstAddress), 
	            	 Instance->UniqueId, 
				     ETH_LENGTH_OF_ADDRESS);
		    COPY_MEMORY( &(MacHeader.SrcAddress), 
	                 Instance->UniqueId, 
				     ETH_LENGTH_OF_ADDRESS);
		    MacHeader.EtherType = ETHERTYPE_IP;

		    Header = (PUCHAR)&MacHeader;
        }
		break;

	} //end switch

    
#ifdef PPPOE_MINIPORT
   if (Instance->Rfc2516Enable)
   {
       /* We need to include header */
       Data -= ETHERNET_HEADER_SIZE;
       Size += ETHERNET_HEADER_SIZE;

       Buffer->Data = Data;
       Buffer->Size = Size;
       (Rfc1483Instance->UserReceive) (
               Instance,
               AtmUniVc,
               Buffer,
               FALSE,
               FALSE);
   }
   else
   {
#endif /* PPPOE_MINIPORT */
        //
	    // Pass the received packet through the RX packet filter
	    //
	    if (  (!ArpType) && Rfc1483FilterReceivedPacket( Instance, (PMAC_HEADER)Header) ) 
        {
		    DBGPRINT(
			    DBG_PRINT_MODULE_RFC1483_RX,
			    DBG_PRINT_LEVEL_TRACE,
			    ("Filtered RX Packet DstMac[0] = 0x%X, PF = 0x%X\n",
			    Header[0],
			    Instance->Rfc1483PacketFilter));
		    AtmUniReceiveComplete(plan->Rfc1483VcHandle,Buffer);
		    return;
	    }
	

        Instance->Rfc1577Config.Rfc1483Rfc1577Receive(
            Instance->Rfc1577Config.Rfc1577Context,
	        AtmUniVc, //Instance->Rfc1483VcHandle,
	        Buffer,
            Header,
            (Buffer->Size - Size),
            ArpType
        );
#ifdef PPPOE_MINIPORT
   }
#endif /* PPPOE_MINIPORT */

}


///////////////////////////////////////////////////////////////////////////////
//
//  NAME: Rfc1483FilterReceivedPacket
//
//  ABSTRACT: 
//		Given the Ethernet header for the packet, indicates whether or not
//      the packet should be filtered or indicated
//
//  RETURNS: 
//		BOOOLEAN
//
//  ASSUMPTIONS:
//		None.
//
///////////////////////////////////////////////////////////////////////////////
BOOLEAN
Rfc1483FilterReceivedPacket(
	PATM_USB_INSTANCE	Instance,
	PMAC_HEADER			MacHeader
)
{
	ULONG	Offset;
	ULONG	i;

	DBGPRINT(DBG_PRINT_MODULE_RFC1483_RX,DBG_PRINT_LEVEL_TRACE,("Rfc1483FilterReceivedPacket\n"));

	ASSERT( Instance);
	ASSERT( MacHeader);

	//
	// If we are in promiscuous mode, all packets are indicated
	//
	if ( Instance->Rfc1483PacketFilter & NDIS_PACKET_TYPE_PROMISCUOUS ) {
		return( FALSE);
	}

	//
	// If the packet is a directed packet and we are accepting directed 
	// packets, check the destination address against our station address
	//
	if ( !ETH_IS_MULTICAST( MacHeader->DstAddress) &
	     ( Instance->Rfc1483PacketFilter & NDIS_PACKET_TYPE_DIRECTED))
	{
		//
		// Compare the destination address against our MAC address
		//
		if ( COMPARE_MEMORY( 
				MacHeader->DstAddress,
				Instance->StationAddress,
				ETH_LENGTH_OF_ADDRESS) == 0)
		{
			//
			// Packet is directed to us - don't filter
			//
			return( FALSE);
		} 
		//
		// Packet is not directed to us - filter
		//
		return( TRUE);
	//
	// The packet is either broadcast or multicast.  
	//
	} else {
		//
		// If the packet is a broadcast packet and we are accepting broadcasts,
		// don't filter; otherwise, filter
		//
		if ( ETH_IS_BROADCAST( MacHeader->DstAddress)) {
			if ( (Instance->Rfc1483PacketFilter & NDIS_PACKET_TYPE_BROADCAST)) 		
			{
				return( FALSE);
			}
			return( TRUE);
		//
		// If the packet is a multicast and we are either accepting all 
		// multicasts or accepting multicasts in our multicast list (and
		// the destination address is in our multicast list), don't filter; 
		// otherwise, filter.
		//
		} else {
			if ( Instance->Rfc1483PacketFilter & NDIS_PACKET_TYPE_ALL_MULTICAST)
			{
				return( FALSE);

			} else if ( Instance->Rfc1483PacketFilter & 
			            NDIS_PACKET_TYPE_MULTICAST )
			{
				//
				// Look for the destination address in our multicast filter
				// table
				//
				Offset = 0;
				for ( i = 0; i < Instance->Rfc1483MulticastFilterSize; i++) {
					if ( COMPARE_MEMORY(
							MacHeader->DstAddress,
							&(Instance->Rfc1483MulticastFilter[ Offset]),
							ETH_LENGTH_OF_ADDRESS) == 0)
					{
						return( FALSE);
					}
					Offset += ETH_LENGTH_OF_ADDRESS;
				}
				return( TRUE);
			} else {
				return( TRUE);
			}
		}
	}

} // Rfc1483FilterReceivedPacket

///////////////////////////////////////////////////////////////////////////////
//
//  NAME: Rfc1483TransferData
//
//  ABSTRACT: 
//		The MiniportTransferData function for the RFC1483 driver
//
//  RETURNS: 
//		void
//
//  ASSUMPTIONS:
//		None.
//
///////////////////////////////////////////////////////////////////////////////
NDIS_STATUS 
Rfc1483TransferData(
	PNDIS_PACKET 		Packet,
	PUINT 				BytesTransferred,
	PATM_USB_INSTANCE	Instance,
	PUCHAR 				Buffer,
	UINT 				ByteOffset,
	UINT 				BytesToTransfer
)
{
   UINT       	DestinationBufferCount;
   PNDIS_BUFFER DestinationCurrentBuffer;
   PUCHAR       SourceCurrentAddress;
   PVOID        DestinationVirtualAddress;
   UINT       	DestinationCurrentLength;
   UINT       	PhyBuffCount;
   UINT       	PacketLength;
   ULONG       	LocalBytesCopied = 0;
   ULONG       	AmountToMove;
   ULONG       	Remaining;

	DBGPRINT(
		DBG_PRINT_MODULE_RFC1483_RX,
		DBG_PRINT_LEVEL_TRACE,
		("Rfc1483IndicateReceivedData\n"));

	ASSERT(Instance);
	ASSERT(Packet);
	ASSERT(Buffer);

	//	
	// Take care of boundary condition of zero length copy. 
	//
	*BytesTransferred = 0;
	if (BytesToTransfer == 0) {
		return( NDIS_STATUS_SUCCESS);
	} 

	//	
	// Get the first buffer of the destination. 
	//
	NdisQueryPacket( 
	  Packet,
	  &PhyBuffCount,
	  &DestinationBufferCount,
	  &DestinationCurrentBuffer,
	  &PacketLength);

	//
	// Could have a null packet. 
	//
	if ( DestinationBufferCount == 0) {
		DBGPRINT(
			DBG_PRINT_MODULE_RFC1483_RX,
			DBG_PRINT_LEVEL_ERR,
			("Packet has no buffers\n"));
		return( NDIS_STATUS_FAILURE);
	} 

	NdisQueryBuffer( DestinationCurrentBuffer, &DestinationVirtualAddress,
	  				 &DestinationCurrentLength);

	//
	// Set up the source address
	//
	SourceCurrentAddress = Buffer + ByteOffset;

	while (LocalBytesCopied < BytesToTransfer) {
		//
		// Check to see whether we've exhausted the current destination
		// buffer.  If so, move onto the next one.
		//
		if (DestinationCurrentLength == 0) {
			NdisGetNextBuffer(
				DestinationCurrentBuffer, 
				&DestinationCurrentBuffer);
			if (DestinationCurrentBuffer == NULL) {
				break;
			} 
			NdisQueryBuffer(
				DestinationCurrentBuffer,
				&DestinationVirtualAddress,
				&DestinationCurrentLength);
		} // if

      Remaining 	= BytesToTransfer - LocalBytesCopied;
      AmountToMove 	= DestinationCurrentLength;
      AmountToMove 	= ((Remaining < AmountToMove)? (Remaining):(AmountToMove));

      COPY_MEMORY( DestinationVirtualAddress,SourceCurrentAddress,AmountToMove);

      SourceCurrentAddress      += AmountToMove;
      LocalBytesCopied          += AmountToMove;
      DestinationCurrentLength  -= AmountToMove;
   } 

   *BytesTransferred = LocalBytesCopied;
   
   return( NDIS_STATUS_SUCCESS);

} // Rfc1483TransferData 

////////////////////////////////////////////////////////////////////////////////
//
// NAME: Rfc1483QueryInformation
//
// ABSTRACT:
//		The NDIS 'MiniportQueryInformation' handler for the RFC1483 driver
//
// RETURNS:
//		NDIS_STATUS
//
// ASSUMPTIONS:
//		None.
//
////////////////////////////////////////////////////////////////////////////////
#ifdef  _NTDDK_
NDIS_STATUS
Rfc1483QueryInformation(
	PATM_USB_INSTANCE	Instance,
	NDIS_OID       		Oid,
	PVOID          		InformationBuffer,
	ULONG          		InformationBufferLength,
	PULONG        		BytesWritten,
	PULONG        		BytesNeeded
)
{

    ULONG 			i;
    ULONG 			SupportedOidCount;
    PVOID 			SourceBuffer;
    ULONG 			SourceBufferLength;
    ULONG 			GenericUlong;
    USHORT			GenericUshort;
    UCHAR 			VendorId[4];
    NDIS_STATUS 	Status = NDIS_STATUS_SUCCESS;
	ATM_UNI_STATS	AtmUniStats;

	DBGPRINT(
		DBG_PRINT_MODULE_RFC1483_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("Rfc1483QueryInformation\n"));

	ASSERT( BytesNeeded);
	ASSERT( BytesWritten);
	if( (NULL == BytesNeeded) || (NULL == BytesWritten) ) {
		DBGPRINT(
			DBG_PRINT_MODULE_RFC1483_ALL,
			DBG_PRINT_LEVEL_ERR,
			("Invalid Input Data\n"));
		return NDIS_STATUS_FAILURE;
	} // if
	
	//
	// Initialize the returned values
	//
    *BytesWritten = 0;
    *BytesNeeded = 0;

	//
	// Queries that require a ULONG will use the GenericUlong variable.
	// Otherwise, the SourceBuffer and SourceBufferLength parameters 
	// are set within the case.
	//
    SourceBuffer 		= (PVOID) &GenericUlong;
    SourceBufferLength 	= sizeof (GenericUlong);

	//
	// Process the Query based on the OID value
	//
	switch ( Oid) {

#ifdef POWERMGMT
		case OID_PNP_CAPABILITIES:
			DBGPRINT(DBG_PRINT_MODULE_RFC1483_ALL,DBG_PRINT_LEVEL_ERR,(" OID_PNP_CAPABILITIES \n"));
			GsiPnpCapabilities.WakeUpCapabilities.MinMagicPacketWakeUp = NdisDeviceStateUnspecified;
			GsiPnpCapabilities.WakeUpCapabilities.MinPatternWakeUp = NdisDeviceStateUnspecified;
			GsiPnpCapabilities.WakeUpCapabilities.MinLinkChangeWakeUp = NdisDeviceStateUnspecified;
			SourceBuffer 		= (PUCHAR)&GsiPnpCapabilities;
		    SourceBufferLength 	= sizeof(GsiPnpCapabilities);
            Status = NDIS_STATUS_SUCCESS;
			break;

		case OID_PNP_QUERY_POWER:
			DBGPRINT(DBG_PRINT_MODULE_RFC1483_ALL,DBG_PRINT_LEVEL_ERR,("OID_PNP_QUERY_POWER\n"));
			SourceBufferLength = 0;
			Status = NDIS_STATUS_SUCCESS;
			break;



#endif /* POWERMGMT */


	case OID_GEN_MAC_OPTIONS:
		GenericUlong = 
			(ULONG)( NDIS_MAC_OPTION_COPY_LOOKAHEAD_DATA |
					 NDIS_MAC_OPTION_TRANSFERS_NOT_PEND  |
                     NDIS_MAC_OPTION_NO_LOOPBACK);
           break;
           
	case OID_GEN_SUPPORTED_LIST:
		SourceBuffer 		= SupportedLanOids;
		SourceBufferLength 	= sizeof( SupportedLanOids);
		break;

    case OID_GEN_HARDWARE_STATUS :
       GenericUlong = NdisHardwareStatusReady;
       break;

    case OID_GEN_MEDIA_SUPPORTED:
    case OID_GEN_MEDIA_IN_USE:
		GenericUlong = NdisMedium802_3;
		break;

	case OID_GEN_MAXIMUM_LOOKAHEAD:
	case OID_GEN_CURRENT_LOOKAHEAD:
	case OID_GEN_MAXIMUM_FRAME_SIZE:
        GenericUlong = ETHERNET_MAX_PACKET_SIZE;
        break;

	case OID_GEN_MAXIMUM_TOTAL_SIZE:
	case OID_GEN_TRANSMIT_BLOCK_SIZE:
	case OID_GEN_RECEIVE_BLOCK_SIZE:
		GenericUlong = ETHERNET_MAX_PACKET_SIZE + ETHERNET_HEADER_SIZE;
		break;

    case OID_GEN_XMIT_OK:
    case OID_GEN_RCV_OK:
    case OID_GEN_XMIT_ERROR:
    case OID_GEN_RCV_ERROR:
    case OID_GEN_RCV_NO_BUFFER:
    case OID_GEN_RCV_CRC_ERROR:
		//
		// All of the statistics requests are processed together.
		// The statistics are maintained in the ATM UNI.
		//
		AtmUniGetStats( Instance->AtmUniHandle, &AtmUniStats);
		switch ( Oid) {
	    case OID_GEN_XMIT_OK:
			GenericUlong = AtmUniStats.TxPdusOk;
			break;

	    case OID_GEN_RCV_OK:
			GenericUlong = AtmUniStats.RxPdusOk;
			break;

	    case OID_GEN_XMIT_ERROR:
			GenericUlong = AtmUniStats.TxPdusError;
			break;

	    case OID_GEN_RCV_ERROR:
			GenericUlong = AtmUniStats.RxPdusError;
			break;

	    case OID_GEN_RCV_NO_BUFFER:
			GenericUlong = AtmUniStats.RxPdusNoBuffer;
			break;

    	case OID_GEN_RCV_CRC_ERROR:
			GenericUlong = AtmUniStats.RxCrcError;
			break;
		}
		break;

	case OID_GEN_LINK_SPEED:
		//
		// NOTE:  must return a valid number or netBEUI will crash!
		//
		// The link speed is reported in units of 100 bps,
		// Always report valid number. If not, report the default value
       // to avoid system crash because of the netBEUI.
       //
       if (Instance->DownstreamRate)
           GenericUlong = Instance->DownstreamRate * 10;
       else
		    GenericUlong = 10240;
		break;

    case OID_GEN_TRANSMIT_BUFFER_SPACE:
		GenericUlong = 
			Instance->TxBufferCount * 
			(ETHERNET_MAX_PACKET_SIZE+ETHERNET_HEADER_SIZE);
		break;
        
	case OID_GEN_RECEIVE_BUFFER_SPACE:
		GenericUlong =
			Instance->RxBufferCount * 
			(ETHERNET_MAX_PACKET_SIZE+ETHERNET_HEADER_SIZE);
		break;

    case OID_GEN_VENDOR_ID:
        COPY_MEMORY( VendorId,Instance->UniqueId, 3);
        VendorId[3] = 0x0;
		SourceBuffer = VendorId;
		SourceBufferLength = sizeof(VendorId);
		break;

	case OID_GEN_VENDOR_DESCRIPTION:
        SourceBuffer 		= Global_AtmUsbVendorDescriptor.Buffer;
        SourceBufferLength 	= Global_AtmUsbVendorDescriptor.MaximumLength;
		break;

    case OID_GEN_DRIVER_VERSION:
		GenericUshort = ATM_USB_DRIVER_VERSION;
		SourceBuffer = &GenericUshort;
		SourceBufferLength = sizeof(USHORT);
		break;

	case OID_GEN_MEDIA_CONNECT_STATUS:
		if ( ATM_PHY_LINK_STATE_CONNECTED == Instance->LinkState) {
			GenericUlong = NdisMediaStateConnected;
		} else {
			GenericUlong = NdisMediaStateDisconnected;
		}
        break;

	case OID_GEN_MAXIMUM_SEND_PACKETS:
		//
		// Since we don't support the MiniportSendPackets interface, we'll
		//   return one as described in the NDIS documentation.
		//
		GenericUlong = 1;
		break;

	case OID_GEN_VENDOR_DRIVER_VERSION:
		//
		// Plug the build version information in here if it's available
		//
		// Least significant word contains the minor version
		// Most significant word contains the major version
		//
		GenericUlong = 0x00010000;
		break;
	
	case OID_GEN_CURRENT_PACKET_FILTER:
		GenericUlong = Instance->Rfc1483PacketFilter;
		break;

	case OID_802_3_PERMANENT_ADDRESS:
		SourceBuffer 		= Instance->UniqueId;
		SourceBufferLength 	= ETH_LENGTH_OF_ADDRESS;
		break;
   
	case OID_802_3_CURRENT_ADDRESS:
		SourceBuffer 		= Instance->StationAddress;
		SourceBufferLength 	= ETH_LENGTH_OF_ADDRESS;
		break;

	case OID_802_3_MAXIMUM_LIST_SIZE:
		GenericUlong = RFC1483_MAX_MULTICAST_ADDRESSES;
		break;

    case OID_802_3_RCV_ERROR_ALIGNMENT:
    case OID_802_3_XMIT_ONE_COLLISION:
    case OID_802_3_XMIT_MORE_COLLISIONS:
		//
		// These statistics have no meaning since we aren't really Ethernet
		//
		GenericUlong = 0;
		break;

    default:

		DBGPRINT(
			DBG_PRINT_MODULE_RFC1483_ALL,
			DBG_PRINT_LEVEL_ERR,
			("Unsupported OID 0x%X\n",
			Oid));

		Status = NDIS_STATUS_NOT_SUPPORTED;
		break;
    } // switch
	
	if ( Status == NDIS_STATUS_SUCCESS) {
	    if (SourceBufferLength > InformationBufferLength) {
	       *BytesNeeded = SourceBufferLength;
	       Status = NDIS_STATUS_BUFFER_TOO_SHORT;
	    } else {
		    COPY_MEMORY( InformationBuffer,SourceBuffer, SourceBufferLength);
		    *BytesWritten = SourceBufferLength;
		} 
	}
    return( Status);

} // Rfc1483QueryInformation 

////////////////////////////////////////////////////////////////////////////////
//
// NAME: Rfc1483SetInformation
//
// ABSTRACT:
//		The NDIS 'MiniportSetInformation' handler for the RFC1483 driver
//
// RETURNS:
//		NDIS_STATUS
//
// ASSUMPTIONS:
//		None.
//
////////////////////////////////////////////////////////////////////////////////
NDIS_STATUS
Rfc1483SetInformation (
	PATM_USB_INSTANCE 	Instance,
	NDIS_OID 			Oid,
	PVOID 				InformationBuffer,
	ULONG 				InformationBufferLength,
	PULONG 				BytesRead,
	PULONG 				BytesNeeded
)
{
    NDIS_STATUS         Status = NDIS_STATUS_SUCCESS;
    ULONG               PacketFilter;

#ifdef POWERMGMT
   NDIS_DEVICE_POWER_STATE powstate;
#endif // POWERMGMT

	DBGPRINT(
		DBG_PRINT_MODULE_RFC1483_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("Rfc1483SetInformation\n"));

	//
	// Initialize the returned values
	//
    *BytesRead = 0;
    *BytesNeeded = 0;

    switch (Oid) {
#ifdef POWERMGMT
		case OID_PNP_SET_POWER:
			powstate = (NDIS_DEVICE_POWER_STATE) 
			*((PNDIS_DEVICE_POWER_STATE)(InformationBuffer));

			DBGPRINT(DBG_PRINT_MODULE_RFC1483_ALL,DBG_PRINT_LEVEL_ERR,("OID_PNP_SET_POWER\n"));
			DBGPRINT(DBG_PRINT_MODULE_RFC1483_ALL,DBG_PRINT_LEVEL_ERR,("power state %d\n", powstate));

		if (powstate == NdisDeviceStateD3)  // suspend
		{
			DBGPRINT(DBG_PRINT_MODULE_RFC1483_ALL,DBG_PRINT_LEVEL_ERR,("Power State Suspend\n"));
			Status = PrepareForPwrStateD3 (Instance);
        }

		else if (powstate == NdisDeviceStateD0) //resume
		{
			DBGPRINT(DBG_PRINT_MODULE_RFC1483_ALL,DBG_PRINT_LEVEL_ERR,("Power State Resume\n"));
			Status = PrepareForPwrStateD0 (Instance);
        }
		else
		{
			DBGPRINT(DBG_PRINT_MODULE_RFC1483_ALL,DBG_PRINT_LEVEL_ERR,("Power State not Specified\n"));		
		}

		break;

	case OID_PNP_ENABLE_WAKE_UP:
		DBGPRINT(DBG_PRINT_MODULE_RFC1483_ALL,DBG_PRINT_LEVEL_ERR,("Should not receive OID_PNP_ENABLE_WAKE_UP\n"));
		Status = NDIS_STATUS_NOT_SUPPORTED;
		break;

	case OID_PNP_ADD_WAKE_UP_PATTERN:
		DBGPRINT(DBG_PRINT_MODULE_RFC1483_ALL,DBG_PRINT_LEVEL_ERR,("Should not receive OID_PNP_ADD_WAKE_UP_PATTERN\n"));
		Status = NDIS_STATUS_NOT_SUPPORTED;
		break;

	case OID_PNP_REMOVE_WAKE_UP_PATTERN:
		
		DBGPRINT(DBG_PRINT_MODULE_RFC1483_ALL,DBG_PRINT_LEVEL_ERR,("Should not receive OID_PNP_REMOVE_WAKE_UP_PATTERN\n"));
		Status = NDIS_STATUS_NOT_SUPPORTED;
		break;

#endif // POWERMGMT



    case OID_802_3_MULTICAST_LIST:
		//
        // The data must be a multiple of the Ethernet address size.
		//
        if (InformationBufferLength % ETH_LENGTH_OF_ADDRESS == 0) {
            *BytesRead = InformationBufferLength;
			//
			// Copy the multicast address list to our multicast filter
			//	( maintained within the instance data)
			//
            COPY_MEMORY( Instance->Rfc1483MulticastFilter,
                         InformationBuffer,
                         InformationBufferLength);

            Instance->Rfc1483MulticastFilterSize = 
            	(USHORT)(InformationBufferLength / ETH_LENGTH_OF_ADDRESS);

            Status = NDIS_STATUS_SUCCESS;
        }
        else
        {
            Status = NDIS_STATUS_INVALID_LENGTH;
        }
        break;

    case OID_GEN_CURRENT_PACKET_FILTER:
        if (InformationBufferLength == 4) {
			PacketFilter = *((PULONG)InformationBuffer);
            *BytesRead = 4;

            // Verify bits, any filter with an unsupported bit will be failed
            if (PacketFilter & 
				~(NDIS_PACKET_TYPE_DIRECTED |
				 NDIS_PACKET_TYPE_MULTICAST |
				 NDIS_PACKET_TYPE_BROADCAST |
				 NDIS_PACKET_TYPE_PROMISCUOUS |
				 NDIS_PACKET_TYPE_ALL_MULTICAST))
            {
				DBGPRINT(
					DBG_PRINT_MODULE_RFC1483_ALL,
					DBG_PRINT_LEVEL_ERR,
					("Unsupported Filter 0x%X\n",
					PacketFilter));
                Status = NDIS_STATUS_NOT_SUPPORTED;
            } else {
				//
				// Save the new packet filter
				//
				Instance->Rfc1483PacketFilter = PacketFilter;
			} 
        } else {
            Status = NDIS_STATUS_INVALID_LENGTH;
        }
        break;

    case OID_GEN_CURRENT_LOOKAHEAD:
        if (InformationBufferLength == 4)           // Verify the Length
        {
            *BytesRead = 4;
            Status = NDIS_STATUS_SUCCESS;
        }
        else
        {
            Status = NDIS_STATUS_INVALID_LENGTH;
        }
        break;


    default:

		DBGPRINT(
			DBG_PRINT_MODULE_RFC1483_ALL,
			DBG_PRINT_LEVEL_ERR,
			("Unsupported OID 0x%X\n",
			Oid));

        Status = NDIS_STATUS_NOT_SUPPORTED;
        break;
    }

    return Status;

} // Rfc1483SetInformation

#endif //_NTDDK_

////////////////////////////////////////////////////////////////////////////////
//
// NAME: Rfc1483LinkStatusChange
//
// ABSTRACT:
//		Called by ATM UNI to indicate that the physical layer link status
//		has changed ( and to report the new values )
//
// RETURNS:
//		void
//
// ASSUMPTIONS:
//		None
//
////////////////////////////////////////////////////////////////////////////////
void
Rfc1483LinkStatusChange(
	PATM_USB_INSTANCE	Instance,
	ATM_PHY_LINK_STATE	LinkState,
	ULONG				UpstreamRate,
	ULONG				DownstreamRate,
	USHORT              COType
)
{
	NDIS_STATUS	NdisStatus;
#ifdef PPPOE_MINIPORT
   PRFC1483_INSTANCE   Rfc1483Instance = (PRFC1483_INSTANCE) Instance->Rfc1483Handle;
#endif /* PPPOE_MINIPORT */

	DBGPRINT(
		DBG_PRINT_MODULE_RFC1483_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("Rfc1483LinkStatusChange\n"));

	//
	// If the link state has changed, indicate the new link state to NDIS
	//
	if ( LinkState != Instance->LinkState) {
		if ( ATM_PHY_LINK_STATE_CONNECTED == LinkState) {
			
			NdisStatus = NDIS_STATUS_MEDIA_CONNECT;
		} else {
			NdisStatus = NDIS_STATUS_MEDIA_DISCONNECT;
		}
#ifdef PPPOE_MINIPORT
       if (Instance->Rfc2516Enable)
       {
           if (Rfc1483Instance->UserLinkStatusChange)
               (Rfc1483Instance->UserLinkStatusChange)(Instance,LinkState,UpstreamRate,DownstreamRate,COType);
       }
       else
       {
#endif /* PPPOE_MINIPORT */

#ifdef ANNEXB
           Rfc1483OpenVc(Instance, LinkState, COType);
#endif // ANNEXB

		    NdisMIndicateStatus(
			    Instance->MiniportAdapterHandle,
			    NdisStatus,
			    NULL,
			    0);
		    NdisMIndicateStatusComplete( Instance->MiniportAdapterHandle);
#ifdef PPPOE_MINIPORT
       }
#endif /* PPPOE_MINIPORT */
	}

	//
	// Save the new link status
	//
	Instance->LinkState 		= LinkState;
	Instance->UpstreamRate		= UpstreamRate;
	Instance->DownstreamRate	= DownstreamRate;


} // Rfc1483LinkStatusChange


void Rfc1483ReceiveComplete(
       	gsi_lan_t			*plan,
        PXFER_DATA_BUFFER	Buffer)
{
    	DBGPRINT(
		DBG_PRINT_MODULE_RFC1483_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("Rfc1483ReceiveComplete\n"));

ASSERT(Instance);
ASSERT(Buffer);
AtmUniReceiveComplete(plan->Rfc1483VcHandle, Buffer);
}


BOOLEAN Rfc1483GetEncapsulation(
                            PATM_USB_INSTANCE	Instance,
                            RFC1483_MODE Rfc1483Mode,
                            PUCHAR *Encapsulation,
                            PUCHAR EncapsulationSize
                            )
{

	switch ( Rfc1483Mode )
    {

	case RFC1483_MODE_BRIDGED_ETH_LLC:
		*Encapsulation 		= (PUCHAR)&Rfc1483LlcEncapBridgedEthNoFcs;
		*EncapsulationSize	= RFC1483_LLC_ENCAP_BRIDGED_ETH_SIZE;

		break;

	case RFC1483_MODE_BRIDGED_ETH_VC:
		
        *Encapsulation		= (PUCHAR)&Rfc1483VcEncapBridgedEth;
		*EncapsulationSize	= RFC1483_VC_ENCAP_BRIDGED_ETH_SIZE;

		break;

	case RFC1483_MODE_ROUTED_IP_LLC:
		*Encapsulation		= (PUCHAR)&Rfc1483LlcEncapRoutedIp;
		*EncapsulationSize	= RFC1483_LLC_ENCAP_ROUTED_IP_SIZE;

		break;

	case RFC1483_MODE_ROUTED_IP_VC:
		*Encapsulation		= NULL;
		*EncapsulationSize	= 0;

		break;

	
    case RFC1483_MODE_RFC1577_ENCAP:
        *Encapsulation = (PUCHAR)&Rfc1483Rfc1577Encap;
        *EncapsulationSize = RFC1483_RFC1577_ENCAP_SIZE;
        break;
    
    default:
		DBGPRINT(
			DBG_PRINT_MODULE_RFC1483_ALL,
			DBG_PRINT_LEVEL_ERR,
			("Bad value for RFC1483 MODE (%d)\n",
			Rfc1483Mode));
		return  FALSE;
		break;
	}//end switch		


return TRUE;
}

////////////////////////////////////////////////////////////////////////////////
//
// NAME: Rfc1483ReadNetworkAddress
//
// ABSTRACT:
//	    This routine will check the registry for a Network address override.
//     If there is none, it will attempt the read the Network address from
//     the serial eeprom. If fails to read it from the serial eeprom then
//     it will default to the broadcast address.
//
// RETURNS:
//		Boolean
//
// ASSUMPTIONS:
//
////////////////////////////////////////////////////////////////////////////////
void Rfc1483ReadNetworkAddress(
                   PATM_USB_INSTANCE   Instance,
                   PUCHAR              PermanentAddress)
{
#ifdef _NTDDK_   
NDIS_STATUS    Status;
NDIS_HANDLE    ConfigHandle = NULL;
UINT           NewNetworkAddressLength=0;
PUCHAR         NewNetworkAddress = NULL;
#endif // _NTDDK_

   /* Intialize the StationAddress */
   COPY_MEMORY( Instance->StationAddress, PermanentAddress, ETH_LENGTH_OF_ADDRESS);

#ifdef  _NTDDK_
   //
   // Open the configuration section of the registry to read the Network Address
   //
   NdisOpenConfiguration( &Status, &ConfigHandle, Instance->WrapperConfigurationContext );
   if ((Status != NDIS_STATUS_SUCCESS) && (ConfigHandle != NULL))
       return;

   //
   // See if user has defined new MAC Address
   //
   NdisReadNetworkAddress(
               (PNDIS_STATUS) &Status,
               (PVOID*) &NewNetworkAddress,
               &NewNetworkAddressLength,
               ConfigHandle);
   if ((Status == NDIS_STATUS_SUCCESS) &&
       (NewNetworkAddressLength != 0)  &&
       (NewNetworkAddress != NULL))
   {
       // Copy it to current address
       COPY_MEMORY( Instance->StationAddress, 
               NewNetworkAddress,
               NewNetworkAddressLength);
   }

   //
   // Close the configuration handle
   //
	if (ConfigHandle != NULL)
		NdisCloseConfiguration( ConfigHandle );
#endif //_NTDDK_
   return;
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//
// NAME: Rfc1483Create
//
// ABSTRACT:
//
// RETURNS:
//
// ASSUMPTIONS:
//		None
//
////////////////////////////////////////////////////////////////////////////////
RFC1483_HANDLE
Rfc1483Create(
       PATM_USB_INSTANCE               Instance,
       RFC1483_USER_LINK_STATUS_CHANGE UserLinkStatusChange,
       BOOLEAN                         IndicatePackets,
       ULONG                           MediaSpecificInformationSize,
       USHORT                          MaxSduSize,
       USHORT                          RxBufferCount,
       USHORT                          TxBufferCount,
       NDIS_HANDLE                     MiniportAdapterHandle,
       ATM_PHY_BIND                    PhyBind,
       ATM_PHY_DEVICE_HANDLE           PhyDeviceHandle,
       PUCHAR                          PhyUniqueId,
       PVOID                           UsbPhyHandle)
{
PRFC1483_INSTANCE  Rfc1483Instance;

   DBGPRINT(DBG_PRINT_MODULE_RFC1483_ALL, DBG_PRINT_LEVEL_TRACE,
               ("Rfc1483Create\n"));

   /* No need to check for the Feature flag because this function will be call
   ** from the higher lasyer (RFC2516) only.
   */
   
   /* Allocate memory for the RFC instance */   
   ALLOCATE_MEMORY( &Rfc1483Instance, sizeof( RFC1483_INSTANCE));
   if( NULL == Rfc1483Instance)
       return( NULL);

   Rfc1483Instance->UserHandle = Instance;
   Rfc1483Instance->UserLinkStatusChange = UserLinkStatusChange;
   Rfc1483Instance->MiniportAdapterHandle = MiniportAdapterHandle;

   Instance->Rfc1483Handle = Rfc1483Instance;
   
   //
   // Create the ATM UNI
   //
   Rfc1483Instance->AtmUniHandle = AtmUniCreate( 
	            Instance,
	            UserLinkStatusChange,
	            IndicatePackets,
	            MediaSpecificInformationSize,
	            MaxSduSize,
	            RxBufferCount,
	            TxBufferCount,
	            MiniportAdapterHandle,
	            PhyBind,
	            PhyDeviceHandle,
	            PhyUniqueId,
	            UsbPhyHandle);

   if( NULL == Rfc1483Instance->AtmUniHandle) 
   {
	    DBGPRINT(DBG_PRINT_MODULE_RFC1483_ALL, DBG_PRINT_LEVEL_ERR,
	                            ("AtmUniCreate failed\n"));
       FREE_MEMORY( Rfc1483Instance, sizeof( RFC1483_INSTANCE));
	    return( NULL);
   }
   
   Instance->AtmUniHandle = Rfc1483Instance->AtmUniHandle;
   
   return ( Rfc1483Instance);
}

////////////////////////////////////////////////////////////////////////////////
//
// NAME: Rfc1483CreateVc
//
// ABSTRACT:
//
// RETURNS:
//
// ASSUMPTIONS:
//		None
//
////////////////////////////////////////////////////////////////////////////////
RFC1483_VC_HANDLE
Rfc1483CreateVc(
   RFC1483_HANDLE              Rfc1483Handle,
  	RFC1483_USER_RECEIVE        UserReceive,
   RFC1483_USER_SEND_COMPLETE  UserSendComplete,
   int							lanindex)
{
PATM_USB_INSTANCE  Instance;
PRFC1483_INSTANCE  Rfc1483Instance = (PRFC1483_INSTANCE) Rfc1483Handle;

   DBGPRINT(DBG_PRINT_MODULE_RFC1483_ALL, DBG_PRINT_LEVEL_TRACE,
               ("Rfc1483CreateVc\n"));

   Instance = Rfc1483Instance->UserHandle,
   Rfc1483Instance->UserReceive = UserReceive;
   Rfc1483Instance->UserSendComplete = UserSendComplete;
   
	Rfc1483Instance->Rfc1483VcHandle = AtmUniCreateVc(
		            Instance->AtmUniHandle,
		            NULL,
		            Rfc1483ReceiveData,
		            NULL,
		            Rfc1483SendComplete);

	if ( NULL == Rfc1483Instance->Rfc1483VcHandle)
   {
       DBGPRINT(DBG_PRINT_MODULE_RFC1483_ALL, DBG_PRINT_LEVEL_ERR,
                       ("AtmUniCreateVc failed\n"));
		return( NULL);
	}
   Instance->Rfc1483VcHandle[lanindex] = Rfc1483Instance->Rfc1483VcHandle;
   
   return (Rfc1483Instance->Rfc1483VcHandle);
}

////////////////////////////////////////////////////////////////////////////////
//
// NAME: Rfc1483ActivateVc
//
// ABSTRACT:
//
// RETURNS:
//
// ASSUMPTIONS:
//		None
//
////////////////////////////////////////////////////////////////////////////////
RFC_STATUS
Rfc1483ActivateVc(
   PATM_USB_INSTANCE   Instance,
   RFC1483_VC_HANDLE   Rfc1483VcHandle,
   UCHAR               Vpi,
   USHORT              Vci,
   ATM_UNI_AAL         Aal,
   USHORT              MaxSduSize,
   PUCHAR              VcEncapsulation,
   USHORT              VcEncapsulationLength,
   BOOLEAN             ReserveResources)
{
RFC_STATUS RfcStatus;

   DBGPRINT(DBG_PRINT_MODULE_RFC1483_ALL, DBG_PRINT_LEVEL_TRACE,
                       ("Rfc1483ActivateVc\n"));

	RfcStatus = 
	AtmUniActivateVc(
		Rfc1483VcHandle,
		Vpi,
		Vci,
		Aal,
	   	MaxSduSize,
		(PUCHAR)Instance->Rfc1483Encapsulation,
		Instance->Rfc1483EncapsulationSize,
		ReserveResources,
        0);
	if ( RFC_STATUS_SUCCESS != RfcStatus)
   {
       DBGPRINT(DBG_PRINT_MODULE_RFC1483_ALL, DBG_PRINT_LEVEL_ERR,
                       ("AtmUniActivateVc failed\n"));
		return( RfcStatus);
	}
   
   return (RFC_STATUS_SUCCESS);
}

////////////////////////////////////////////////////////////////////////////////
//
// NAME: Rfc1483DeactivateVc
//
// ABSTRACT:
//
// RETURNS:
//
// ASSUMPTIONS:
//		None
//
////////////////////////////////////////////////////////////////////////////////
void Rfc1483DeactivateVc(RFC1483_VC_HANDLE Rfc1483VcHandle)
{
   DBGPRINT(DBG_PRINT_MODULE_RFC1483_ALL, DBG_PRINT_LEVEL_TRACE,
               ("Rfc1483DeactivateVc\n"));
   AtmUniDeactivateVc(Rfc1483VcHandle);
}

////////////////////////////////////////////////////////////////////////////////
//
// NAME: Rfc1483DeleteVc
//
// ABSTRACT:
//
// RETURNS:
//
// ASSUMPTIONS:
//		None
//
////////////////////////////////////////////////////////////////////////////////
void Rfc1483DeleteVc(RFC1483_VC_HANDLE Rfc1483VcHandle)
{
   DBGPRINT(DBG_PRINT_MODULE_RFC1483_ALL, DBG_PRINT_LEVEL_TRACE,
               ("Rfc1483DeleteVc\n"));
   AtmUniDeleteVc(Rfc1483VcHandle);
}

#ifdef ANNEXB
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
void Rfc1483OpenVc(
	PATM_USB_INSTANCE	Instance,
	ATM_PHY_LINK_STATE	LinkState,
   USHORT              COType)
{
ATM_UNI_STATUS	AtmUniStatus;

   /* If there is no change, do nothing */
	if ( LinkState == Instance->LinkState)
       return;

   if (ATM_PHY_LINK_STATE_CONNECTED == LinkState)
   {
       //Rfc1483LookupConnectionInfo();
       /* Select the VPI/VCI based on the DSLAM connection */
       switch(COType)
       {
           case 0x12:  //ADI
               Instance->Rfc1483Vpi = (UCHAR)Instance->ADIVpi;
               Instance->Rfc1483Vci = Instance->ADIVci;
               break;

           case 0x13:
           case 0x14:
               Instance->Rfc1483Vpi = (UCHAR)Instance->TIVpi;
               Instance->Rfc1483Vci = Instance->TIVci;
               break;

           default:
               Instance->Rfc1483Vpi = (UCHAR)Instance->DefaultVpi;
               Instance->Rfc1483Vci = Instance->DefaultVci;
               break;
       }
/* FIXME: Need to create another handle inorder activate another PVC(192/32) */
       /* Disconnect the Vc Handle */
/* Will cause less problems:
1) Unpluggin the USB cable after the DSL cable is removed
2) No need to confiure Rfc1577 routines (I do not understand it)
3)
*/
       // FIXME:  might want to just create and activate all PVCs
	   AtmUniDeactivateVc(Instance->Rfc1483VcHandle);

       /* Activate the VC */
	    AtmUniStatus = AtmUniActivateVc(
	            Instance->Rfc1483VcHandle,
	            Instance->Rfc1483Vpi,
	            Instance->Rfc1483Vci,
	            AAL5_,
	            RFC1483_MAX_SDU_SIZE,
	            (PUCHAR)Instance->Rfc1483Encapsulation,
	            Instance->Rfc1483EncapsulationSize,
	            FALSE,
                0);
	
       if ( ATM_UNI_STATUS_SUCCESS != AtmUniStatus)
       {
		        DBGPRINT(DBG_PRINT_MODULE_RFC1483_ALL, DBG_PRINT_LEVEL_ERR,
			        ("AtmUniActivateVc failed\n"));
		        return;
	    }
   }	
}
#endif // ANNEXB

#endif // GSI_LAN
