// ABSTRACT:
//		Contains the implementation for the RFC2364 module.  The RFC2364 module
//		supports four PDU options:
//
//		1. LLC Encapsulated PPP
//		2. VC Encapsulated  PPP
//		
//		The type of Encapsulation is set during initialization 
//      The default mode is VC Encapsulated  PPP (hard coded by now)
//		Each of the Encapsulation options is described briefly below. 
//      See RFC2364 for details.
//
//		1. LLC Encapsulated PPP
//
//
//			      +-------------------------------+
//               |       LLC 0xFE-FE-03          |
//               +-------------------------------+
//               |       NLPID = PPP 0xCF        |
//               +-------------------------------+
//               |                               |
//               |         PPP payload           |
//               |                               |
//               +-------------------------------+
//
//		Once encapsulated, the RFC1483 PDUs are sent using AAL5 on the PVC 
//		provisioned during initialization.

//		2. VC Encapsulated PPP
//
//
//               +-------------------------------+
//               |                               |
//               |         PPP payload           |
//               |                               |
//               +-------------------------------+
//
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
//
// Local Constant Definitions
//
//------------------------------------------------------------------------------
#if defined(WAN_MINIPORT) || defined(GSI_WAN)

#include "rfc2364.h"
#include "atmuni.h"
#ifdef _NTDDK_
#include "version.h"
#endif //_NTDDK_

#define    RFC2364_MAX_SDU_SIZE   2048
#define    MAX_CONNECTIONS        1
#define    MAX_PACKET_POOL        1
//#define    PLINEPTR_VENDOR_ID             "NTS"
#ifdef _NTDDK_
#define    PLINEPTR_VENDOR_DESCRPTION     GUSB_DESCRIPTION
#endif // _NTDDK_

#ifdef POWERMGMT
NDIS_PNP_CAPABILITIES GsiPnpCapabilities; 
extern NDIS_STATUS PrepareForPwrStateD3 (void *Instance);
extern NDIS_STATUS PrepareForPwrStateD0(void *Instance);
#endif // POWERMGMT


#ifdef _NTDDK_
extern void LinkInitialize(
	       IN PATM_USB_INSTANCE Instance,
	       IN PSTRING           AddressList
	       );
extern void TapiShutdown(PATM_USB_INSTANCE Instance);
#endif // _NTDDK_

#include "pshpack1.h"
typedef struct _LLC_HEADER_OSI 
{
    UCHAR  	Llc[3];
    UCHAR      Nlpid;
} LLC_HEADER_OSI, *PLLC_HEADER_OSI;
#include "poppack.h" // restore default packing

//------------------------------------------------------------------------------
//
// Local Variable Declarations
//
//------------------------------------------------------------------------------
static LLC_HEADER_OSI Rfc2364LlcEncap = {{0xFE,0xFE,0x03},0xCF};

/************************************************************************
*                          DATA DECLARATIONS
*************************************************************************
*/
#ifdef _NTDDK_
// This list includes all OIDS which the driver should see from NDIS 
static const NDIS_OID SupportedOidArray[] =
{
   OID_GEN_SUPPORTED_LIST,
   OID_GEN_HARDWARE_STATUS,
   OID_GEN_MEDIA_SUPPORTED,
   OID_GEN_MEDIA_IN_USE,
   OID_GEN_MAXIMUM_LOOKAHEAD,
   OID_GEN_MAXIMUM_FRAME_SIZE,
   OID_GEN_MAXIMUM_TOTAL_SIZE,
   OID_GEN_MAC_OPTIONS,
   OID_GEN_LINK_SPEED,
   OID_GEN_TRANSMIT_BUFFER_SPACE,
   OID_GEN_RECEIVE_BUFFER_SPACE,
   OID_GEN_TRANSMIT_BLOCK_SIZE,
   OID_GEN_RECEIVE_BLOCK_SIZE,
   OID_GEN_VENDOR_ID,
   OID_GEN_VENDOR_DESCRIPTION,
   OID_GEN_DRIVER_VERSION,
   OID_GEN_CURRENT_LOOKAHEAD,
   OID_WAN_MEDIUM_SUBTYPE,
   OID_WAN_GET_INFO,
   OID_WAN_PERMANENT_ADDRESS,
   OID_WAN_CURRENT_ADDRESS,
   OID_WAN_GET_LINK_INFO,
   OID_WAN_SET_LINK_INFO,
   OID_WAN_GET_COMP_INFO,
   OID_WAN_SET_COMP_INFO,
#ifdef POWERMGMT
    OID_PNP_CAPABILITIES,
	OID_PNP_QUERY_POWER,
    OID_PNP_ENABLE_WAKE_UP,      
    OID_PNP_WAKE_UP_PATTERN_LIST,
    OID_PNP_SET_POWER,
    OID_PNP_ENABLE_WAKE_UP,
    OID_PNP_ADD_WAKE_UP_PATTERN,
    OID_PNP_REMOVE_WAKE_UP_PATTERN,
#endif /* POWERMGMT */
   0
};

#if DBG
// Maintain this list in the same order as the above list
static char *SupportedOidNames[] =
{
   "OID_GEN_SUPPORTED_LIST",
   "OID_GEN_HARDWARE_STATUS",
   "OID_GEN_MEDIA_SUPPORTED",
   "OID_GEN_MEDIA_IN_USE",
   "OID_GEN_MAXIMUM_LOOKAHEAD",
   "OID_GEN_MAXIMUM_FRAME_SIZE",
   "OID_GEN_MAXIMUM_TOTAL_SIZE",
   "OID_GEN_MAC_OPTIONS",
   "OID_GEN_LINK_SPEED",
   "OID_GEN_TRANSMIT_BUFFER_SPACE",
   "OID_GEN_RECEIVE_BUFFER_SPACE",
   "OID_GEN_TRANSMIT_BLOCK_SIZE",
   "OID_GEN_RECEIVE_BLOCK_SIZE",
   "OID_GEN_VENDOR_ID",
   "OID_GEN_VENDOR_DESCRIPTION",
   "OID_GEN_DRIVER_VERSION",
   "OID_GEN_CURRENT_LOOKAHEAD",
   "OID_WAN_MEDIUM_SUBTYPE",
   "OID_WAN_GET_INFO",
   "OID_WAN_PERMANENT_ADDRESS",
   "OID_WAN_CURRENT_ADDRESS",
   "OID_WAN_GET_LINK_INFO",
   "OID_WAN_SET_LINK_INFO",
   "OID_WAN_GET_COMP_INFO",
   "OID_WAN_SET_COMP_INFO",
   #ifdef POWERMGMT
    "OID_PNP_CAPABILITIES",
	"OID_PNP_QUERY_POWER",
    "OID_PNP_ENABLE_WAKE_UP",      
    "OID_PNP_WAKE_UP_PATTERN_LIST",
    "OID_PNP_SET_POWER",
    "OID_PNP_ENABLE_WAKE_UP",
    "OID_PNP_ADD_WAKE_UP_PATTERN",
    "OID_PNP_REMOVE_WAKE_UP_PATTERN",
#endif /* POWERMGMT */
   "OID_UNKNOWN"
};

#define NUM_OID_ENTRIES (sizeof(SupportedOidArray) / sizeof(SupportedOidArray[0]))

// Find the string corresponding to the OID value
char * NsGetOidString(NDIS_OID Oid)
{
UINT i;

   for (i = 0; i < NUM_OID_ENTRIES-1; i++)
   {
      if (SupportedOidArray[i] == Oid)
         break;
   }
   return(SupportedOidNames[i]);
}
#endif // DBG
#endif // _NTDDK_

// Returned from an OID_WAN_PERMANENT_ADDRESS GsiQueryInformation request.
// The WAN wrapper wants the miniport to return a unique address for this
// adapter.  This is used as an ethernet address presented to the protocols.
// The least significant bit of the first byte must not be a 1, or it could
// be interpreted as an ethernet multicast address.  If the vendor has an
// assigned ethernet vendor code (the first 3 bytes), they should be used
// to assure that the address does not conflict with another vendor's address.
// The last digit is replaced during the call with the adapter instance number.
static UCHAR GsiWanAddress[4] = {0};
//static UCHAR GsiWanAddress[4] = {'G','s','i','0'};

// Returned from an OID_GEN_VENDOR_ID GsiQueryInformation request.
// Again, the vendor's assigned ethernet vendor code should be used if possible.

/*
* Since, the vendorID number needs to be unique, we will read the MAC address
* for the GsiVendorID
*/
static UCHAR GsiVendorID[4] = {0};
//static UCHAR GsiVendorID[4] = PLINEPTR_VENDOR_ID;

// Returned from an OID_GEN_VENDOR_DESCRIPTION GsiQueryInformation request.
// This is an arbitrary string which may be used by upper layers to present
// a user friendly description of the adapter.
#ifdef _NTDDK_
static UCHAR GsiVendorDescription[] = PLINEPTR_VENDOR_DESCRPTION;
#endif // _NTDDK_

////////////////////////////////////////////////////////////////////////////////
//
// NAME: Rfc2364Initialize
//
// ABSTRACT:
//		Initializes the RFC2364 layer,
//
// RETURNS:
//		Boolean
//
// ASSUMPTIONS:
//		Runs in the context of MiniportInitialize.
//
////////////////////////////////////////////////////////////////////////////////
BOOLEAN
Rfc2364Initialize( PATM_USB_INSTANCE       Instance ,
                  	ATM_PHY_BIND 			PhyBind,
        	        ATM_PHY_DEVICE_HANDLE	PhyDeviceHandle)
{
	PUCHAR			Encapsulation;
	UCHAR			EncapsulationSize;
   NDIS_STATUS     Status;  
#ifdef _NTDDK_   
   ANSI_STRING     AddressList;

	NDIS_PHYSICAL_ADDRESS HighestAcceptableAddress = NDIS_PHYSICAL_ADDRESS_CONST(-1,-1);
#endif // _NTDDK_

   DBGPRINT(
		DBG_PRINT_MODULE_RFC2364_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("Rfc2364Initialize\n"));

ASSERT(Instance);
ASSERT(PhyBind);
ASSERT(PhyDeviceHandle);

   
 	switch ( Instance->Rfc2364Mode ) 
   {
       case RFC2364_MODE_LLC:
		    Encapsulation 		= (PUCHAR)&Rfc2364LlcEncap;
		    EncapsulationSize	= RFC2364_LLC_ENCAP_SIZE;
		    break;

	    case RFC2364_MODE_VC:
           Encapsulation		= NULL;
           EncapsulationSize	= RFC2364_VC_ENCAP_SIZE; 
           break;

       default:
		    DBGPRINT(DBG_PRINT_MODULE_RFC2364_ALL,DBG_PRINT_LEVEL_TRACE,
				("Bad value for RFC2364 MODE (%d)\n", Instance->Rfc2364Mode));
		    return( FALSE);
	}		

   Instance->Encapsulation = Encapsulation;
   Instance->EncapsulationSize = EncapsulationSize;

#ifdef PPPOE_MINIPORT
   if (Instance->Rfc2516Enable)
   {
       /* Intialize everything first */
       if ( !(Rfc2516Initialize(Instance, PhyBind, PhyDeviceHandle)) )
       {
           DBGPRINT(DBG_PRINT_MODULE_RFC2364_ALL, DBG_PRINT_LEVEL_ERR,
               ("Rfc2516Intialize failed\n"));
           return( FALSE);
	    }

       Instance->Rfc2516Handle = Rfc2516Create(
		                                    Instance,
		                                    Rfc2364LinkStatusChange,
		                                    FALSE, 
		                                    0,	  
		                                    RFC2364_MAX_SDU_SIZE,
		                                    Instance->RxBufferCount,
		                                    Instance->TxBufferCount,
		                                    Instance->MiniportAdapterHandle,
		                                    PhyBind,
		                                    PhyDeviceHandle,
		                                    Instance->UniqueId,
		                                    &(Instance->UsbPhyHandle));
   }
   else
   {
#endif /* PPPOE_MINIPORT */
#if defined(linux)
         if ((Instance->flags & _ATMUNI_ALLOC) != _ATMUNI_ALLOC)
         {
#endif // linux

       Instance->AtmUniHandle =
	    AtmUniCreate( 
		    Instance,
		    Rfc2364LinkStatusChange,
		    FALSE, 
		    0,	  
		    RFC2364_MAX_SDU_SIZE,
		    Instance->RxBufferCount,
		    Instance->TxBufferCount,
		    Instance->MiniportAdapterHandle,
		    PhyBind,
		    PhyDeviceHandle,
		    Instance->UniqueId,
		    &(Instance->UsbPhyHandle));
#if defined(linux)
         }
#endif // linux         

	    if( NULL == Instance->AtmUniHandle) 
       {
           DBGPRINT(
			    DBG_PRINT_MODULE_RFC2364_ALL,
			    DBG_PRINT_LEVEL_ERR,
			    ("AtmUniCreate failed\n"));
		    return( FALSE);
	    } 
#ifdef PPPOE_MINIPORT
   }
#endif /* PPPOE_MINIPORT */

   /* Initialize the address and vendor ID */
	GsiWanAddress[0] = GsiVendorID[0] = Instance->UniqueId[0];
	GsiWanAddress[1] = GsiVendorID[1] = Instance->UniqueId[1];
	GsiWanAddress[2] = GsiVendorID[2] = Instance->UniqueId[2];
	GsiWanAddress[3] = GsiVendorID[3] = 0;


   /* Allocate PacketPool queue */
   NdisAllocatePacketPool (
       &Status,
       &Instance->NdisPktPoolHandle,
       MAX_PACKET_POOL,
       64);
   if (Status != NDIS_STATUS_SUCCESS)
		return(FALSE);

   NdisAllocateBufferPool (
       &Status,
       &Instance->NdisBuffPoolHandle,
       MAX_PACKET_POOL);
   if (Status != NDIS_STATUS_SUCCESS)
		return(FALSE);

   NdisAllocatePacket (&Status, &Instance->NdisPkt, Instance->NdisPktPoolHandle);
   if (Status != NDIS_STATUS_SUCCESS)
		return( FALSE); 

   Instance->transmitFrameCounter = 0;

#ifdef _NTDDK_
   // Initialize the WAN information structure to match the capabilities of
   // the adapter.
   Instance->WanInfo.MaxFrameSize   = Instance->MaxFrameSize;
   Instance->WanInfo.MaxTransmit    = PLINEPTR_MAX_TRANSMITS;
       
   // Since we copy the packets to/from adapter RAM, no padding information is
   // needed in the WAN packets we get.  We can just use adapter RAM as needed.
   Instance->WanInfo.HeaderPadding  = 0;
   Instance->WanInfo.TailPadding    = 0;
       
   // This value indicates how many point to point connections are allowed
   // per WAN link.  Currently, the WAN wrapper only supports 1 line per link.
   Instance->WanInfo.Endpoints      = 1;
       
   // Found that for some reason, this memoryFlag set to MEMORY_CONTIGUOUS does
   // not work well when unpluging the USB device or unloading the PCI device.
	Instance->WanInfo.MemoryFlags    = 0;

   Instance->WanInfo.HighestAcceptableAddress = HighestAcceptableAddress;
        
   // We only support point to point framing, and we don't need to see the
   // address or control fields.  The TAPI_PROVIDER bit is set to indicate
   // that we can accept the TAPI OID requests.
   Instance->WanInfo.FramingBits = PPP_FRAMING | TAPI_PROVIDER;
       
   // This value is ignored by this driver, but its default behavior is such
   // that all these control bytes would appear to be handled transparently.
   Instance->WanInfo.DesiredACCM    = 0;

   // Get AddressList for this adapter.
   strcpy(Instance->DeviceName, GUSB_DEVICE_NAME);
	
	LinkInitialize(Instance, &AddressList);

#endif // _NTDDK_

   return TRUE;
}

////////////////////////////////////////////////////////////////////////////////
//
// NAME: Rfc2364Encapsulate
//
// ABSTRACT:
//		Put rfc 2364 Header,
//
// RETURNS:
//		Encapsulation Size, in Case of Error 0xFFFF 
//
// ASSUMPTIONS:
//		Runs in the context of MiniportInitialize.
//
////////////////////////////////////////////////////////////////////////////////
USHORT Rfc2364Encapsulate(PATM_USB_INSTANCE Instance , PUCHAR pData,   ULONG Size )
{

	UCHAR          *savBuffPtr;
	PUCHAR			Encapsulation = NULL;
	USHORT			EncapsulationSize;

#ifdef _NTDDK_
   NDIS_PHYSICAL_ADDRESS   MaxAddress = NDIS_PHYSICAL_ADDRESS_CONST (-1, -1);
#endif // _NTDDK_

   DBGPRINT(
		DBG_PRINT_MODULE_RFC2364_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("Rfc2364Encapsulate\n"));

#ifdef _NTDDK_
   NdisAllocateMemory((PVOID *)&savBuffPtr, Size, 0, MaxAddress);
#elif defined(linux)
    ALLOCATE_MEMORY((PVOID *)&savBuffPtr, Size);
    if (savBuffPtr)
    {
#endif // linux
   NdisMoveMemory(savBuffPtr, pData, Size);

   switch( Instance->Rfc2364Mode ) 
   {
	    case RFC2364_MODE_LLC:
           Encapsulation = (PUCHAR)&Rfc2364LlcEncap;
           EncapsulationSize = RFC2364_LLC_ENCAP_SIZE;
           break;

       case RFC2364_MODE_VC:
           EncapsulationSize = RFC2364_VC_ENCAP_SIZE;
           Encapsulation = NULL;
           break;

       default:
           DBGPRINT(DBG_PRINT_MODULE_RFC2364_ALL,DBG_PRINT_LEVEL_TRACE,
			   ("Invalid Rfc2364 Mode Value = %d\n" ,Instance->Rfc2364Mode  ));
           return 0xFFFF; //(-1)
    }//end switch   
	
	NdisMoveMemory(pData + EncapsulationSize,
                      savBuffPtr,
                      Size);
   if(Encapsulation && EncapsulationSize)
       NdisMoveMemory(pData, Encapsulation, EncapsulationSize);
#ifdef _NTDDK_       
   NdisFreeMemory ((PVOID *)savBuffPtr, Size, 0);
#elif defined(linux)   
   FREE_MEMORY(savBuffPtr, Size);
   }
#endif // linux        
   return EncapsulationSize;
}


void
Rfc2364LinkStatusChange(
	PATM_USB_INSTANCE	Instance,
	ATM_PHY_LINK_STATE	LinkState,
	ULONG				UpstreamRate,
	ULONG				DownstreamRate,
	USHORT              COType
)
{
NDIS_STATUS    NdisStatus;

#ifdef _NTDDK_
/* The Call is hard-coded to 0 because we only support 1 connection */
PPLINEPTR_CALL  Call = &(Instance->WanLink.CallsOnLink[0]);
#endif //_NTDDK_

   DBGPRINT(
       DBG_PRINT_MODULE_RFC2364_ALL,
       DBG_PRINT_LEVEL_TRACE,
       ("Rfc2364LinkStatusChange\n"));

   //
   // If the link state has changed, indicate the new link state to NDIS
   //
   if ( LinkState != Instance->LinkState)
   {
       if ( ATM_PHY_LINK_STATE_CONNECTED == LinkState)
			NdisStatus = NDIS_STATUS_MEDIA_CONNECT;
       else
			NdisStatus = NDIS_STATUS_MEDIA_DISCONNECT;
           
		NdisMIndicateStatus(
			Instance->MiniportAdapterHandle,
			NdisStatus,
			NULL,
			0);
		NdisMIndicateStatusComplete( Instance->MiniportAdapterHandle);

#ifdef _NTDDK_
       if ( ATM_PHY_LINK_STATE_CONNECTED != LinkState)
       {
           /* Just updating the LinkState ahead of time */
	        Instance->LinkState = LinkState;
           TapiCheckCallState(Call);
       }
#endif // _NTDDK_       
	}

	//
	// Save the new link status
	//
	Instance->LinkState 		= LinkState;
	Instance->UpstreamRate		= UpstreamRate;
	Instance->DownstreamRate	= DownstreamRate;
   
} // Rfc2364LinkStatusChange

NDIS_STATUS 
Rfc2364CreateVc(
	PATM_USB_INSTANCE	Instance,
   PATM_USB_INSTANCE_CALL  Call
)
{
	ATM_UNI_HANDLE	AtmUniVcHandle;
	ATM_UNI_STATUS	AtmUniStatus;
#ifdef PPPOE_MINIPORT   
	NDIS_STATUS     Status;
#endif /* PPPOE_MINIPORT */   
    UCHAR 		VPI = (UCHAR)(Call->VPI);
	USHORT 		VCI = (USHORT)(Call->VCI);
	PATM_UNI	AtmUni = Instance->AtmUniHandle;

   ASSERT(Instance);

	DBGPRINT(
		DBG_PRINT_MODULE_RFC2364_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("Rfc2364CreateVc\n"));

    if( Instance->Rfc2364Connections > MAX_CONNECTIONS ) 
    {
		DBGPRINT(
			DBG_PRINT_MODULE_RFC2364_ALL,
			DBG_PRINT_LEVEL_ERR,
			("ERROR Number of rfc2364 Connections = %d\n", Instance->Rfc2364Connections));

        return NDIS_STATUS_FAILURE;
    }

#ifdef ANNEXB
	/* Select the VPI/VCI based on the DSLAM connection */
	switch(AtmUni->COType)
	{
	   case 0x12:  //ADI
		   VPI = (UCHAR)Instance->ADIVpi;
		   VCI = Instance->ADIVci;
		   break;

	   case 0x13:
	   case 0x14:
		   VPI = (UCHAR)Instance->TIVpi;
		   VCI = Instance->TIVci;
		   break;
	}
#endif // ANNEXB

   /* Check if the VPI and VCI values are out-of-range */
   if ((VPI > 0xff) || (VCI > 0x0000ffff))
   {
       DBGPRINT(
           DBG_PRINT_MODULE_RFC2364_ALL,
           DBG_PRINT_LEVEL_ERR,
           ("ERROR VPI = %d VCI = %d\n", VPI , VCI ));

       return (NDIS_STATUS_TAPI_INVALCALLPARAMS);
	}

#ifdef PPPOE_MINIPORT
   /* Feature Flag indicates, if we need
   ** to include Encapsulation (or RFCs).
   */
   if (Instance->Rfc2516Enable)
   {
       Instance->Rfc2364VcHandle = Rfc2516CreateVc(
		                Instance->Rfc2516Handle,
		                Rfc2364ReceiveData,
		                Rfc2364SendComplete );
	    if (NULL == Instance->Rfc2364VcHandle) 
       {
           DBGPRINT(DBG_PRINT_MODULE_RFC2364_ALL, DBG_PRINT_LEVEL_ERR,
                           ("Rfc2516CreateVc failed\n"));
           return(NDIS_STATUS_RESOURCES);
       }
       AtmUniVcHandle = Instance->Rfc2364VcHandle;
       
	    Status = Rfc2516ActivateVc(
                    Instance,
                    Instance->Rfc2364VcHandle,
		            VPI,
		            VCI,
		            AAL5_,
	   	            RFC2364_MAX_SDU_SIZE,
		            Instance->Encapsulation,
		            Instance->EncapsulationSize,
		            FALSE);
	    if ( ATM_UNI_STATUS_SUCCESS != Status)
       {
           DBGPRINT(DBG_PRINT_MODULE_RFC2364_ALL, DBG_PRINT_LEVEL_ERR,
			    ("Rfc2516ActivateVc failed\n"));
		    return( NDIS_STATUS_FAILURE );
	    }
	}
   else
   {
#endif /* PPPOE_MINIPORT */
	    AtmUniVcHandle =
	    AtmUniCreateVc(
		    Instance->AtmUniHandle,
		    NULL,
		    Rfc2364ReceiveData,
		    NULL,
		    Rfc2364SendComplete );

	    if ( NULL == AtmUniVcHandle) 
       {
		    DBGPRINT(
			    DBG_PRINT_MODULE_RFC2364_ALL,
			    DBG_PRINT_LEVEL_ERR,
			    ("AtmUniCreateVc failed\n"));
		    return( NDIS_STATUS_RESOURCES);
	    } 

	    AtmUniStatus = 
	    AtmUniActivateVc(
		    AtmUniVcHandle,
		    VPI,
		    VCI,
		    AAL5_,
	   	    RFC2364_MAX_SDU_SIZE,
		    Instance->Encapsulation,
		    Instance->EncapsulationSize,
		    FALSE,
            0);
	
	    if ( ATM_UNI_STATUS_SUCCESS != AtmUniStatus)
       {
		    DBGPRINT(
			    DBG_PRINT_MODULE_RFC2364_ALL,
			    DBG_PRINT_LEVEL_ERR,
			    ("AtmUniActivateVc failed\n"));
		    return( NDIS_STATUS_FAILURE );
	    }
#ifdef PPPOE_MINIPORT
   }
#endif /* PPPOE_MINIPORT */

   Instance->Rfc2364Vpi = VPI;
	Instance->Rfc2364Vci = VCI;
	Instance->Rfc2364VcHandle = AtmUniVcHandle;
   Instance->Call = Call;
   Call->NdisLinkContext = NULL;	// prevents indication before TAPI link is up
   Instance->Rfc2364Connections++;

   return( NDIS_STATUS_SUCCESS);
} // AtmMiniportCreateVc

NDIS_STATUS Rfc2364DeleteVc(
              	PATM_USB_INSTANCE	Instance,
            	UINT 		VPI,
	            UINT 		VCI
               )
{
   DBGPRINT(
       DBG_PRINT_MODULE_RFC2364_ALL,
       DBG_PRINT_LEVEL_TRACE,
       ("Rfc2364DeleteVc\n"));

   if( (VPI == Instance->Rfc2364Vpi) && (Instance->Rfc2364Vci == VCI) )
   {
       /* Sanity Check */
       if (Instance->Rfc2364Connections <= 0)
       {
           DBGPRINT(DBG_PRINT_MODULE_RFC2364_ALL, DBG_PRINT_LEVEL_ERR,
               ("Rfc2364DeleteVc: No Connection Opened\n"));
           return( NDIS_STATUS_FAILURE );
       }

       Instance->Rfc2364Connections--;
#ifdef PPPOE_MINIPORT
       if (Instance->Rfc2516Enable)
       {
           Rfc2516DeactivateVc(Instance->Rfc2364VcHandle);
           Rfc2516DeleteVc(Instance->Rfc2364VcHandle);
       }
       else
       {
#endif /* PPPOE_MINIPORT */
           AtmUniDeactivateVc(Instance->Rfc2364VcHandle );
           AtmUniDeleteVc(	Instance->Rfc2364VcHandle);
#ifdef PPPOE_MINIPORT
       }
#endif /* PPPOE_MINIPORT */
   }
   else
   {
       DBGPRINT(
           DBG_PRINT_MODULE_RFC2364_ALL,
           DBG_PRINT_LEVEL_ERR,
           (" Rfc2364DeleteVc failed\n"));
       return( NDIS_STATUS_FAILURE );
   }
   return( NDIS_STATUS_SUCCESS);
}

////////////////////////////////////////////////////////////////////////////////
//
// NAME: Rfc1483Shutdown
//
// ABSTRACT:
//		Shuts down the RFC1483 layer, deleting the 1483 PVC and destroying the
//		ATM UNI.
//
// RETURNS:
//		void
//
// ASSUMPTIONS:
//		The RFC1483 layer has been successfully initialized.
//
////////////////////////////////////////////////////////////////////////////////
void Rfc2364Shutdown( PATM_USB_INSTANCE Instance)
{
	DBGPRINT(
		DBG_PRINT_MODULE_RFC2364_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("Rfc2364Shutdown\n"));

	ASSERT( Instance);

   /* Shutdown the TAPI, incase, it is not closed */
#ifdef _NTDDK_
   TapiShutdown(Instance);
#endif // _NTDDK_

	NdisFreePacket ( Instance->NdisPkt);
	NdisFreePacketPool ( Instance->NdisPktPoolHandle);
	NdisFreeBufferPool ( Instance->NdisBuffPoolHandle);

#ifdef PPPOE_MINIPORT
   if (Instance->Rfc2516Enable)
       Rfc2516Shutdown(Instance->Rfc2516Handle);
   else
#endif /* PPPOE_MINIPORT */
#if !defined(GSI_LAN)
	    AtmUniDestroy( Instance->AtmUniHandle);
#endif //GSI_LAN
} // Rfc1483Shutdown

/************************************************************************
* Function     : GsiWanSend
* Description  : GsiWanSend instructs a WAN driver to transmit a packet through the
*              : adapter onto the medium.
*
*              : Ownership of both the packet descriptor and the packet data is transferred
*              : to the WAN driver until the request is completed, either synchronously or
*              : asynchronously.  If the WAN driver returns a status other than
*              : NDIS_STATUS_PENDING, then the request is complete, and ownership of the
*              : packet immediately reverts to the protocol.  If the WAN driver returns
*              : NDIS_STATUS_PENDING, then the WAN driver must later indicate completion
*              : of the request by calling Ndis(M)WanSendComplete.
*
*              : The packet passed in Ndis(M)WanSend will contain simple HDLC PPP framing.
*
*              : NOTE: The MacReservedx section as well as the WanPacketQueue section of
*              :       the NDIS_WAN_PACKET is fully available for use by the WAN driver.
*
*              : Interrupts are in any state during this routine.
*
* Input        : MacBindingHandle _ The handle to be passed to NdisMWanSendComplete().
*              : NdisLinkHandle _ The Miniport link handle passed to NDIS_LINE_UP
*              : WanPacket _ A pointer to the NDIS_WAN_PACKET strucutre.  The structure
*              :             contains a pointer to a contiguous buffer with guaranteed
*              :             padding at the beginning and end.  The driver may manipulate
*              :             the buffer in any way.
*
*              : The available header padding is simply CurrentBuffer-StartBuffer.
*              : The available tail padding is EndBuffer-(CurrentBuffer+CurrentLength).
*
* Output       : None
* Return       : NDIS_STATUS_SUCCESS
*              : NDIS_STATUS_FAILURE
************************************************************************/
#ifdef _NTDDK_
NDIS_STATUS Rfc2364Send(
		       IN NDIS_HANDLE         MacBindingHandle,
		       IN NDIS_HANDLE         CallContext,
		       IN PNDIS_WAN_PACKET    NdisWanPkt
		       )
#elif defined(linux)
NDIS_STATUS Rfc2364Send(
		       IN NDIS_HANDLE         MacBindingHandle,
		       IN NDIS_HANDLE         CallContext,
		       IN PNDIS_PACKET       NdisPkt
		       )
#endif // linux
{
#ifdef _NTDDK_
PATM_USB_INSTANCE_CALL	Call;
PATM_USB_INSTANCE_LINK	Link;
NDIS_STATUS				Status;
ULONG					BytesToSend; 
ULONG					i = 0;
#endif // _NTDDK_
PATM_USB_INSTANCE		Instance;
#ifdef _NTDDK_
PUCHAR					buffDataPtr;
PNDIS_BUFFER			NdisBuff;
#endif // _NTDDK_
ATM_UNI_STATUS			AtmUniSendStatus;

/* Check:
** 1) Is the physical layer connected
** 2) Is we have the TAPI connection opened
** 3) Sanity check (like length)
*/

	DBGPRINT(
		DBG_PRINT_MODULE_RFC2364_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("Rfc2364Send\n"));

#if defined(linux)
   Instance = (PATM_USB_INSTANCE)CallContext;
#endif // linux

#ifdef _NTDDK_
   BytesToSend = NdisWanPkt->CurrentLength;
   Call = (PATM_USB_INSTANCE_CALL) CallContext;
   Instance = Call->Link->Instance;
    
   if(Call->CallState != LINECALLSTATE_CONNECTED)
       return (NDIS_STATUS_FAILURE); // Return if line is not connected.
   else
   {
       buffDataPtr = (PUCHAR) NdisWanPkt->CurrentBuffer ;
       /* Check if the Address and Control Field (ACF) is present:
       ** We should alway remove the ACF, if far-end has not negotiated (ACFC)
       ** If the ACF Compression (ACFC) is set by the far-end, we do not remove it.
       */
       if ((buffDataPtr[0] == 0xFF) && (buffDataPtr[1] == 0x03))
       {
           buffDataPtr  = buffDataPtr + 2;
           BytesToSend -= 2;
       }

       // Save these values for use when the transmit completes.
       NdisWanPkt->MacReserved1 = (PVOID) Call;
       NdisWanPkt->MacReserved2 = (PVOID) MacBindingHandle;

       /* Convert the WAN_PACKET buffer into NDIS_PACKET before giving to
       ** SAR layer (ATMUNI).
       */

       //  First, attach the buffer to the packet if available
       NdisAllocateBuffer (
           &Status,
           &NdisBuff,
           Instance->NdisBuffPoolHandle,
           buffDataPtr,
           BytesToSend);

       if (Status != NDIS_STATUS_SUCCESS)
           return (NDIS_STATUS_FAILURE);

	   NdisReinitializePacket (Instance->NdisPkt);
       NdisChainBufferAtFront (Instance->NdisPkt, NdisBuff);
       (Instance->NdisPkt)->Reserved[0] = ( ULONG_PTR )NdisWanPkt;
       Instance->NdisWanPkt = (PVOID)NdisWanPkt;
#endif // _NTDDK_

#ifdef PPPOE_MINIPORT
       if (Instance->Rfc2516Enable)
       {
           AtmUniSendStatus = Rfc2516Send (
               Instance->Rfc2364VcHandle,
               Instance->NdisPkt,
               FALSE,
               0,
               0,
               FALSE,
		       FALSE);
       }
       else
       {
#endif /* PPPOE_MINIPORT */
           AtmUniSendStatus = AtmUniSend (
               Instance->Rfc2364VcHandle ,
#ifdef   _NTDDK_
               Instance->NdisPkt,
#elif defined(linux)
               NdisPkt,
#endif // linux               
               FALSE,
               0,
               0,
               FALSE,
		       FALSE);
#ifdef PPPOE_MINIPORT
       }
#endif /* PPPOE_MINIPORT */

       if (AtmUniSendStatus == ATM_UNI_STATUS_FAILURE)
       {
#ifdef   _NTDDK_         
           NdisUnchainBufferAtFront (Instance->NdisPkt, &NdisBuff);
           if (NdisBuff)
               NdisFreeBuffer (NdisBuff);
#endif // _NTDDK_           
           return(NDIS_STATUS_FAILURE);
       }
       else  if (AtmUniSendStatus == ATM_UNI_STATUS_PENDING)
           return(NDIS_STATUS_PENDING);
       else
       {
#ifdef _NTDDK_         
			NdisUnchainBufferAtFront (Instance->NdisPkt, &NdisBuff);
			if (NdisBuff)
               NdisFreeBuffer (NdisBuff);
#endif   // _NTDDK_
            return(NDIS_STATUS_SUCCESS);
       }
#ifdef _NTDDK_       
   }
#endif   // _NTDDK_
   return (NDIS_STATUS_SUCCESS);
}
    
/************************************************************************
* Function     : GsiWanSendComplete
* Description  : This routine is called by GsiHandleInterrupt() to handle a transmit
*              : complete interrupt.  We enter here with interrupts disabled at the adapter,
*              : but the Adapter->Lock is held.  We walk the TransmitBusyList to find all
*              : the transmits that have completed.
*
*              : NOTE: The Adapter->Lock must be held before calling this routine.
*
* Input        : Adapter _ A pointer ot our adapter information structure.
*
* Output       : None
* Return       : None
************************************************************************/
void Rfc2364SendComplete(
           PATM_USB_INSTANCE   Instance,
           ATM_UNI_USER_HANDLE NotUsed,
           PNDIS_PACKET        NdisPkt)
{
#ifdef _NTDDK_   
PNDIS_WAN_PACKET           NdisWanPkt = NULL;
PNDIS_BUFFER               NdisBuff = NULL;
PATM_USB_INSTANCE_CALL     Call;
PATM_USB_INSTANCE_LINK     Link;

	DBGPRINT(
		DBG_PRINT_MODULE_RFC2364_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("Rfc2364SendComplete\n"));

   NdisWanPkt =  (PNDIS_WAN_PACKET)Instance->NdisWanPkt;
   Call   = (PATM_USB_INSTANCE_CALL) NdisWanPkt->MacReserved1;
   Link = (PATM_USB_INSTANCE_LINK) NdisWanPkt->MacReserved2;

   Instance->transmitFrameCounter++;
   /* Let the higher layer know, we are done with the NdisWanPkt */
   NdisMWanSendComplete(Instance->MiniportAdapterHandle,
                        NdisWanPkt,
                        NDIS_STATUS_SUCCESS);

   NdisUnchainBufferAtFront (Instance->NdisPkt, &NdisBuff);
   if (NdisBuff)
	   NdisFreeBuffer (NdisBuff);
#elif defined(linux)
   /* Let the higher layer know, we are done with the NdisPkt */
   NdisMWanSendComplete(Instance->MiniportAdapterHandle,
                        NdisPkt,
                        NDIS_STATUS_SUCCESS, NULL); /* TODO */
#endif // linux
}
    
/************************************************************************
* Function     : GsiReceiveData
* Description  : 
*
* Input        : 
*
* Output       : None
* Return       : None
************************************************************************/
void Rfc2364ReceiveData (
   PATM_USB_INSTANCE   Instance,
   ATM_UNI_USER_HANDLE AtmUniVcHandle,
   PXFER_DATA_BUFFER   Buffer,
   UCHAR               CpcsUu,
   UCHAR               Cpi
)
{
BOOLEAN     Success;
PATM_USB_INSTANCE_CALL     Call = Instance->Call;
#ifdef PPPOE_MINIPORT
NDIS_STATUS    Status;
PUCHAR         DataPtr;
ULONG          DataLen;

   DataPtr = Buffer->Data;
   DataLen = Buffer->Size;
#endif /* PPPOE_MINIPORT */

	DBGPRINT(
		DBG_PRINT_MODULE_RFC2364_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("Rfc2364ReceiveData\n"));

    if (NULL != Call->NdisLinkContext)		// prevents indication before TAPI link is up
    {
#ifdef PPPOE_MINIPORT
       if (Instance->Rfc2516Enable)
       {
           NdisMWanIndicateReceive (
                       &Status,
                       Instance->MiniportAdapterHandle,
                       Call->NdisLinkContext,
                       (PUCHAR)DataPtr,
                       DataLen);
           // Indicate receive complete if we had any that were accepted.
           if (NDIS_STATUS_SUCCESS == Status)
           {
               NdisMWanIndicateReceiveComplete(Instance->MiniportAdapterHandle,Call->NdisLinkContext);
           }
           Rfc2364ReceiveNotify (Instance, Buffer);
      }
       else
       {
#endif /* PPPOE_MINIPORT */
		// In order to indicate the received data to NDIS, we must schedule
		// a receive callback
		//
		Success = AtmUsbScheduleReceiveCallback( 
					Instance,
					Rfc2364IndicateReceivedData,
					Buffer, AtmUniVcHandle);

		// If our request failed, our callback function won't run. 
		// Drop the received PDU
		//
		if (!Success) 
		{
           Rfc2364ReceiveNotify (Instance, Buffer);
		}
#ifdef PPPOE_MINIPORT
       }
#endif /* PPPOE_MINIPORT */
    }
    else
    {
		DBGPRINT(
			DBG_PRINT_MODULE_RFC2364_ALL,
			DBG_PRINT_LEVEL_TRACE,
			("LinkContext == NULL  "));
        
       Rfc2364ReceiveNotify (Instance, Buffer);
	}
}


///////////////////////////////////////////////////////////////////////////////
//
//  NAME: Rfc2364IndicateReceivedData
//
//  ABSTRACT: 
//		Called within a receive callback to indicate data previously received
//		by Rfc2364ReceiveData
//
//  RETURNS: 
//		void
//
//  ASSUMPTIONS:
//		None
//
///////////////////////////////////////////////////////////////////////////////
void
Rfc2364IndicateReceivedData(
	PATM_USB_INSTANCE	    Instance,
    PXFER_DATA_BUFFER		Buffer,
    ATM_UNI_HANDLE          AtmUniVc
    )
{
	NDIS_STATUS Status;
	PUCHAR      DataPtr;
	ULONG       DataLen;
	PATM_USB_INSTANCE_CALL     Call = Instance->Call;
    ULONG       EncapsulationSize;


    if ( (Instance->Status & ATM_USB_STATUS_HALTING) || (Instance->Status & ATM_USB_STATUS_SUSPENDING) )
    {
        Rfc2364ReceiveNotify (Instance, Buffer);
        return;
    }

	DataPtr = Buffer->Data;
	DataLen = Buffer->Size;

#ifdef HEXDUMPWAN
	HexDumpWan(DataPtr, DataLen);
#endif
	switch ( Instance->Rfc2364Mode ) 
	{
       case RFC2364_MODE_LLC:
           if (COMPARE_MEMORY (DataPtr, 
                           Instance->Encapsulation, 
                           Instance->EncapsulationSize) == 0)
		    {
               EncapsulationSize = Instance->EncapsulationSize;
           }
           else
           {
               DBGPRINT(
                   DBG_PRINT_MODULE_RFC2364_ALL,
                   DBG_PRINT_LEVEL_ERR,
                   ("Encapsulation Error\n"));

               Rfc2364ReceiveNotify (Instance, Buffer);
               return;
           }
           break;

       case RFC2364_MODE_VC:
           EncapsulationSize = Instance->EncapsulationSize;
           break;
       
	    default:
			DBGPRINT(
				DBG_PRINT_MODULE_RFC2364_ALL,
				DBG_PRINT_LEVEL_ERR,
				("Encapsulation Error Mode = %d\n", Instance->Rfc2364Mode ));

           Rfc2364ReceiveNotify (Instance, Buffer);
			return;
	}//end switch

   DataPtr    += EncapsulationSize;
   DataLen    -= EncapsulationSize;

   /* FIXME:
   ** This portion of the code is never tested because lack of proper equipment.
   ** Does the LLC encapsulation comes before the Address and Control Field (ACF)?
   ** I do not think so but others think, it is LLC encapsulation and than ACF in
   ** the current implementation. If I am correct, we need to move this code below.
   ** It is not tested.
   */
   if ((DataPtr[0] == 0xFF) && (DataPtr[1] == 0x03))
   {
       DataPtr  = DataPtr + 2;
	    DataLen -= 2;
   }

	NdisMWanIndicateReceive (
		   &Status,
		   Instance->MiniportAdapterHandle,
		   Call->NdisLinkContext,
		   (PUCHAR)DataPtr,
		   DataLen);

	// Indicate receive complete if we had any that were accepted.
	if (NDIS_STATUS_SUCCESS == Status)
	{
		NdisMWanIndicateReceiveComplete (
			Instance->MiniportAdapterHandle,
			Call->NdisLinkContext);
	}
	else
	{
		DBGPRINT(
			DBG_PRINT_MODULE_RFC2364_ALL,
			DBG_PRINT_LEVEL_ERR,
			("NdisWan rejected the packet\n"));
	}

	// Return the buffer
	//
    Rfc2364ReceiveNotify (Instance, Buffer);
}

#ifdef   _NTDDK_
/************************************************************************
* Function     : GsiQueryInformation
* Description  : This routine initializes the Link structures within the Adapter
*              : The MiniportQueryInformation request allows the inspection of the
*              : Miniport driver's capabilities and current status.
*
*              : If the Miniport does not complete the call immediately (by returning
*              : NDIS_STATUS_PENDING), it must call NdisMQueryInformationComplete to
*              : complete the call.  The Miniport controls the buffers pointed to by
*              : InformationBuffer, BytesWritten, and BytesNeeded until the request
*              : completes.
*
*              : No other requests will be submitted to the Miniport driver until
*              : this request has been completed.
*
*              : Note that the wrapper will intercept all queries of the following OIDs:
*              :     OID_GEN_CURRENT_PACKET_FILTER,
*              :     OID_GEN_PROTOCOL_OPTIONS,
*              :     OID_802_5_CURRENT_FUNCTIONAL,
*              :     OID_802_3_MULTICAST_LIST,
*              :     OID_FDDI_LONG_MULTICAST_LIST,
*              :     OID_FDDI_SHORT_MULTICAST_LIST.
*
*              : Interrupts are in any state during this call.
*
* Input        : Adapter _ A pointer ot our adapter information structure.
*
*              : Oid _ The OID.  (See section 7.4 of the NDIS 3.0 specification for a
*              :       complete description of OIDs.)
*
*              : InformationBuffer _ The buffer that will receive the information.
*              :                     (See section 7.4 of the NDIS 3.0 specification
*              :                     for a description of the length required for each
*              :                     OID.)
*
*              : InformationBufferLength _ The length in bytes of InformationBuffer.
*
*              : BytesWritten _ Returns the number of bytes written into
*              :                InformationBuffer.
*
*              : BytesNeeded _ This parameter returns the number of additional bytes
*              :               needed to satisfy the OID.
*
* Output       : None
* Return       : NDIS_STATUS_INVALID_DATA
*              : NDIS_STATUS_INVALID_LENGTH
*              : NDIS_STATUS_INVALID_OID
*              : NDIS_STATUS_NOT_ACCEPTED
*              : NDIS_STATUS_NOT_SUPPORTED
*              : NDIS_STATUS_PENDING
*              : NDIS_STATUS_RESOURCES
*              : NDIS_STATUS_SUCCESS
************************************************************************/
NDIS_STATUS Rfc2364QueryInformation(
                           IN PATM_USB_INSTANCE Instance,
                           IN NDIS_OID Oid,
                           IN PVOID InformationBuffer,
                           IN ULONG InformationBufferLength,
                           OUT PULONG BytesWritten,
                           OUT PULONG BytesNeeded
                           )
{
// Holds the status result returned by this function.
NDIS_STATUS Status = NDIS_STATUS_NOT_SUPPORTED;

// Pointer to driver data to be copied back to caller's InformationBuffer
PVOID SourceBuffer;
ULONG SourceBufferLength;  // Number of bytes to be copied from driver.

// Most return values are long integers, so this is used to hold the
// return value of a constant or computed result.
ULONG GenericUlong = 0;
USHORT GenericUshort; // Like above, only short.
PPLINEPTR_LINK Link;
PPLINEPTR_CALL Call;
// if it's the comp info OID we'll need this
NDIS_WAN_GET_COMP_INFO WanCompInfo;

	DBGPRINT(
		DBG_PRINT_MODULE_NDIS_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("GsiQueryInformation\n"));

   // If this is a TAPI OID, pass it on over.
   if ((Oid & 0xFFFFFF00L) == (OID_TAPI_ACCEPT & 0xFFFFFF00L))
   {
      Status = NsTapiQueryInformation(Instance,
                           Oid, InformationBuffer, InformationBufferLength,
                           BytesWritten, BytesNeeded);
      return (Status);
   }


   // Initialize these once, since this is the majority of cases.
   SourceBuffer = &GenericUlong;
   SourceBufferLength = sizeof(ULONG);

   // Determine which OID is being requested and do the right thing.
   // Refer to section 7.4 of the NDIS 3.0 specification for a complete 
   // description of OIDs and their return values.
   switch (Oid)
   {
#ifdef POWERMGMT
		case OID_PNP_CAPABILITIES:
		    GsiPnpCapabilities.WakeUpCapabilities.MinMagicPacketWakeUp = NdisDeviceStateUnspecified;
			GsiPnpCapabilities.WakeUpCapabilities.MinPatternWakeUp = NdisDeviceStateUnspecified;
			GsiPnpCapabilities.WakeUpCapabilities.MinLinkChangeWakeUp = NdisDeviceStateUnspecified;

			SourceBuffer = (PUCHAR)&GsiPnpCapabilities;
			SourceBufferLength = sizeof(GsiPnpCapabilities);
			Status = NDIS_STATUS_SUCCESS;
           break;

       case OID_PNP_QUERY_POWER:
			// tell NDIS that we are ready to set NIC to low power state.
			SourceBufferLength = 0;
			Status = NDIS_STATUS_SUCCESS;
			break;
#endif   

       case OID_GEN_SUPPORTED_LIST:
		    SourceBuffer = (PVOID)(SupportedOidArray);
           SourceBufferLength = sizeof(SupportedOidArray);
           break;

       case OID_GEN_HARDWARE_STATUS:
		    GenericUlong = Instance->NeedReset ? NdisHardwareStatusNotReady :
                                                 NdisHardwareStatusReady;
           break;

       case OID_GEN_MEDIA_SUPPORTED:
       case OID_GEN_MEDIA_IN_USE:
		    GenericUlong = NdisMediumWan;
           break;

       case OID_GEN_MAXIMUM_LOOKAHEAD:
       case OID_GEN_CURRENT_LOOKAHEAD:
           /* Normally, we would report lower number because of the MAC header
           ** BUT we do not have MAC header for this product
           */
		    GenericUlong = Instance->MaxFrameSize;
           break;
         
       case OID_GEN_LINK_SPEED:
		    GenericUlong = PLINEPTR_LINK_SPEED;
           break;

       case OID_GEN_TRANSMIT_BUFFER_SPACE:
		    GenericUlong = Instance->TxBufferCount * Instance->MaxFrameSize;
           break;

       case OID_GEN_RECEIVE_BUFFER_SPACE:
		    GenericUlong = Instance->RxBufferCount * Instance->MaxFrameSize;
           break;

       case OID_GEN_RECEIVE_BLOCK_SIZE:
       case OID_GEN_MAXIMUM_FRAME_SIZE:
       case OID_GEN_TRANSMIT_BLOCK_SIZE:
       case OID_GEN_MAXIMUM_TOTAL_SIZE:
		    GenericUlong = Instance->MaxFrameSize;
           break;

       case OID_GEN_VENDOR_ID:
		    SourceBuffer = GsiVendorID;
           SourceBufferLength = sizeof(GsiVendorID);
           break;

       case OID_GEN_VENDOR_DESCRIPTION:
		    SourceBuffer = GsiVendorDescription;
           SourceBufferLength = strlen(GsiVendorDescription) + 1;
           break;

       case OID_GEN_MAC_OPTIONS:
		    GenericUlong = NDIS_MAC_OPTION_COPY_LOOKAHEAD_DATA |
                           NDIS_MAC_OPTION_NO_LOOPBACK |
                           NDIS_MAC_OPTION_TRANSFERS_NOT_PEND;
           break;

       case OID_GEN_DRIVER_VERSION:
       	GenericUshort = ATM_USB_DRIVER_VERSION;	  
		    SourceBuffer = &GenericUshort;
           SourceBufferLength = sizeof(USHORT);
           break;

       case OID_WAN_MEDIUM_SUBTYPE:
		    GenericUlong = NdisWanMediumIsdn;
           break;

       case OID_WAN_GET_INFO:
		    SourceBuffer = &Instance->WanInfo;
           SourceBufferLength = sizeof(NDIS_WAN_INFO);
           break;

       case OID_WAN_PERMANENT_ADDRESS:
       case OID_WAN_CURRENT_ADDRESS:
		    GsiWanAddress[3] = Instance->InstanceNumber + '0';
           SourceBuffer = GsiWanAddress;
           SourceBufferLength = sizeof(GsiWanAddress);
           break;

       case OID_WAN_GET_COMP_INFO:
		    // The first field in the info buffer is a MiniportLinkContext
           // which is really a pointer to an entry in our link information.
           // If this aint so, bail out...
           Call =(PPLINEPTR_CALL)(((PNDIS_WAN_GET_COMP_INFO)InformationBuffer)->NdisLinkHandle);
           Link = Call->Link;
           if (!IS_VALID_LINK(Instance, Link))
           {
                SourceBufferLength = 0;
                Status = NDIS_STATUS_INVALID_DATA;
                break;
           }
   
           SourceBuffer = &WanCompInfo;
           NdisMoveMemory(SourceBuffer,InformationBuffer,InformationBufferLength);

           WanCompInfo.SendCapabilities.CompType = COMPTYPE_NONE;   
           WanCompInfo.SendCapabilities.CompLength = 0;   
           WanCompInfo.RecvCapabilities.CompType = COMPTYPE_NONE;   
           WanCompInfo.RecvCapabilities.CompLength = 0;   

           SourceBufferLength = InformationBufferLength;
           break;
      
       case OID_WAN_GET_LINK_INFO:
 		    // NdisLinkHandle is set to be the address of the call in the LINE_UP message
           Call = (PPLINEPTR_CALL) (((PNDIS_WAN_SET_LINK_INFO)InformationBuffer)->NdisLinkHandle);
           Link = Call->Link;
           if (!IS_VALID_LINK(Instance, Link))
           {
                SourceBufferLength = 0;
                Status = NDIS_STATUS_INVALID_DATA;
                break;
           }

           Link->WanLinkInfo.NdisLinkHandle = Call;   
                
           SourceBuffer = &(Link->WanLinkInfo);
           SourceBufferLength = sizeof(NDIS_WAN_GET_LINK_INFO);
           break;

       default: // Unknown OID
	        Status = NDIS_STATUS_INVALID_OID;
           SourceBufferLength = 0;
           break;
   }

   // Copy the data into the caller's buffer if there is enough room.
   // Otherwise, report the error and and the needed length.
   if (SourceBufferLength > InformationBufferLength)
   {
      *BytesNeeded = SourceBufferLength;
      Status = NDIS_STATUS_INVALID_LENGTH;
   }
   else if (SourceBufferLength)
   {
      NdisMoveMemory(InformationBuffer, SourceBuffer, SourceBufferLength);
      *BytesNeeded = *BytesWritten = SourceBufferLength;
      Status = NDIS_STATUS_SUCCESS;
   }
   else
      *BytesNeeded = *BytesWritten = 0;

	DBGPRINT(
		DBG_PRINT_MODULE_NDIS_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("GsiQueryInformation return\n"));

   return (Status);
}

/************************************************************************
* Function     : GsiSetInformation
* Description  : The MiniportSetInformation request allows for control of the Miniport
*              : by changing information maintained by the Miniport driver.
*
*              : Any of the settable NDIS Global Oids may be used. (see section 7.4 of
*              : the NDIS 3.0 specification for a complete description of the NDIS Oids.)
*
*              : If the Miniport does not complete the call immediately (by returning
*              : NDIS_STATUS_PENDING), it must call NdisMSetInformationComplete to
*              : complete the call.  The Miniport controls the buffers pointed to by
*              : InformationBuffer, BytesRead, and BytesNeeded until the request completes.
*
*              : Interrupts are in any state during the call, and no other requests will
*              : be submitted to the Miniport until this request is completed.
*
* Input        : Adapter _ A pointer ot our adapter information structure.
*
*              : Oid _ The OID.  (See section 7.4 of the NDIS 3.0 specification for
*              :       a complete description of OIDs.)
*
*              : InformationBuffer _ The buffer that will receive the information.
*              :                     (See section 7.4 of the NDIS 3.0 specification for
*              :                     a description of the length required for each OID.)
*
*              : InformationBufferLength _ The length in bytes of InformationBuffer.
*
*              : BytesRead_ Returns the number of bytes read from InformationBuffer.
*
*              : BytesNeeded _ This parameter returns the number of additional bytes
*              :               expected to satisfy the OID.
*
* Output       : None
* Return       : NDIS_STATUS_INVALID_DATA
*              : NDIS_STATUS_INVALID_LENGTH
*              : NDIS_STATUS_INVALID_OID
*              : NDIS_STATUS_NOT_ACCEPTED
*              : NDIS_STATUS_NOT_SUPPORTED
*              : NDIS_STATUS_PENDING
*              : NDIS_STATUS_RESOURCES
*              : NDIS_STATUS_SUCCESS
************************************************************************/
NDIS_STATUS Rfc2364SetInformation(
                           IN PATM_USB_INSTANCE Instance,
                           IN NDIS_OID Oid,
                           IN PVOID InformationBuffer,
                           IN ULONG InformationBufferLength,
                           OUT PULONG BytesRead,
                           OUT PULONG BytesNeeded
                           )
{
// Holds the status result returned by this function.
NDIS_STATUS Status;
PPLINEPTR_LINK Link;
PPLINEPTR_CALL Call;

#ifdef POWERMGMT
NDIS_DEVICE_POWER_STATE powstate;
#endif // POWERMGMT

	DBGPRINT(
		DBG_PRINT_MODULE_NDIS_ALL,
		DBG_PRINT_LEVEL_TRACE,
		("GsiSetInformation\n"));

   // If this is a TAPI OID, pass it on over.
   if ((Oid & 0xFFFFFF00L) == (OID_TAPI_ACCEPT & 0xFFFFFF00L))
   {
      Status = NsTapiSetInformation(Instance, Oid,
                              InformationBuffer, InformationBufferLength,
                              BytesRead, BytesNeeded);
      return (Status);
   }


   // Assume no extra bytes are needed.
   ASSERT(BytesRead && BytesNeeded);
   *BytesRead = 0;
   *BytesNeeded = 0;

   // Determine which OID is being requested and do the right thing.
   switch (Oid)
   {
#ifdef POWERMGMT
       case OID_PNP_SET_POWER:

		    powstate = (NDIS_DEVICE_POWER_STATE)*((PNDIS_DEVICE_POWER_STATE)(InformationBuffer));
           
  		if (powstate == NdisDeviceStateD3)  // suspend
		{
			DBGPRINT(DBG_PRINT_MODULE_RFC2364_ALL,DBG_PRINT_LEVEL_TRACE,("Power State Suspend\n"));
			Status = PrepareForPwrStateD3 (Instance);
        }

		else if (powstate == NdisDeviceStateD0) //resume
		{
			DBGPRINT(DBG_PRINT_MODULE_RFC2364_ALL,DBG_PRINT_LEVEL_TRACE,("Power State Resume\n"));
			Status = PrepareForPwrStateD0 (Instance);
        }
		else
		{
			DBGPRINT(DBG_PRINT_MODULE_RFC2364_ALL,DBG_PRINT_LEVEL_TRACE,("Power State not Specified\n"));		
		}

            
            break;

	    case OID_PNP_ENABLE_WAKE_UP:
		    Status = NDIS_STATUS_NOT_SUPPORTED;
		    break;

	    case OID_PNP_ADD_WAKE_UP_PATTERN:
		    Status = NDIS_STATUS_NOT_SUPPORTED;
		    break;

	    case OID_PNP_REMOVE_WAKE_UP_PATTERN:
		    Status = NDIS_STATUS_NOT_SUPPORTED;
		    break;
#endif // POWERMGMT

       case OID_GEN_CURRENT_LOOKAHEAD:
           // WAN drivers always indicate the entire packet regardless of the
           // lookahead size.  So this request should be politely ignored.
           ASSERT(InformationBufferLength == sizeof(ULONG));
           *BytesNeeded = *BytesRead = sizeof(ULONG);
           Status = NDIS_STATUS_SUCCESS;
           break;

       case OID_WAN_SET_COMP_INFO:
           // We don't really need to do anything here
           Status = NDIS_STATUS_SUCCESS;
           break;

       case OID_WAN_SET_LINK_INFO:
           if (InformationBufferLength == sizeof(NDIS_WAN_SET_LINK_INFO))
           {
                // The first field in the info buffer is a MiniportLinkContext
                // which is really a pointer to an entry in our WanLinkArray.
                // If this aint so, bail out...
                Call =(PPLINEPTR_CALL)(((PNDIS_WAN_SET_LINK_INFO)InformationBuffer)->NdisLinkHandle);
                if (Call == NULL)
                {
                   Status = NDIS_STATUS_INVALID_DATA;
                   break;
                }

                Link = Call->Link;

                // Copy the data into our WanLinkInfo structure.
                NdisMoveMemory(&(Link->WanLinkInfo),InformationBuffer,InformationBufferLength);
                *BytesRead = sizeof(NDIS_WAN_SET_LINK_INFO);
                Status = NDIS_STATUS_SUCCESS;
           }
           else
           {
                Status = NDIS_STATUS_INVALID_LENGTH;
           }
           *BytesNeeded = sizeof(NDIS_WAN_SET_LINK_INFO);
           break;

       default: // Unknown OID
           Status = NDIS_STATUS_INVALID_OID;
           break;
   }

   return (Status);
}

#endif   // _NTDDK_

/************************************************************************
* Function     : Rfc2364ReceiveNotify
* Description  : 
*
* Input        : 
*
* Output       : None
* Return       : 
************************************************************************/
void Rfc2364ReceiveNotify(
           PATM_USB_INSTANCE  Instance,
           PXFER_DATA_BUFFER   Buffer)
{

#ifdef PPPOE_MINIPORT
   if (Instance->Rfc2516Enable)
       Rfc2516ReceiveComplete(Instance->Rfc2364VcHandle, Buffer);
   else
#endif /* PPPOE_MINIPORT */
       AtmUniReceiveComplete (Instance->Rfc2364VcHandle, Buffer);
}

#endif //WAN_MINIPORT
